// Copyright YALI AMIT, January 15, 2001.

#ifndef CARTTREE_H
#define CARTTREE_H

// __________________________________________________________________________
// Include files

#include <iomanip>
#include "histgram.h"
#include "tree.h"
#include "landmark.h"

extern int point1x, point1y, point2x, point2y, point3x, point3y;
extern int cartstyle;
extern int carthistsize;
// __________________________________________________________________________

class carthist : public histogram
{
 private:

 public:
  carthist() : histogram(carthistsize) { }

}; // class carthist

// __________________________________________________________________________

class cartdata
{ 
 private:

 public:
  int pixindex;
  double pixthresh;
  lgrid datalist;
  carthist hist;

  cartdata() : pixindex(-1), pixthresh(0) { }
};

class carttree;

typedef treeiterator<cartdata> cartiter;

class carttreestats
{
 private:
  int total;
  int correct;
  int wtdepth;
  double wtent;

 public:
  carttreestats() : total(0), correct(0), wtdepth(0), wtent(0) { }
  void operator+=(const carttreestats& d) { total += d.total; 
    correct += d.correct; wtdepth += d.wtdepth; wtent += d.wtent;
  }
  void dump() const;

  friend class carttree;
};

class carttree : public tree<cartdata>
{
 private:

 public:
  carttreestats *stats;
  int carttype;
  int cartsprd;
  int cartsx;
  int cartsy;
  igrid points;

  carttree() : stats(0), carttype(cartstyle) { points.init(6);
                                               points[0]=point1x;
                                               points[1]=point1y;
					       points[2]=point2x;
                                               points[3]=point2y;
					       points[4]=point3x;
                                               points[5]=point3y;
                                               }
  ~carttree() { delete stats; }

  void calc_stats(const cartiter& diter);

  void dump(ostream& os = cout);
  void dumpnodes(const cartiter& t, ostream& os);
  void dumpnode(const cartiter& t, ostream& os);

  void write(ostream& ofs);
  void write(const simpstrg& treename);
  void writenodes(const cartiter& t, ofstream& ofs);

  int read(const simpstrg& treename);
  void readnodes(const cartiter& t, ifstream& ifs);
  void updatehists(const cartiter& t);

  void deepen(const cartiter& diter, dgrids& cartims, ucgridss& lands);
  void drop(dgrids& cartims, ucgridss& reglandss, dgrid& agg) const;
  void drop(dgrid& cartimage, ucgrids& reglands, dgrid& agg) const;
  void dropdata(dgrid& currim, ucgrids& reglands, cartiter& diter) const;
  void grow(const lgrid& datalist, dgrids& cartims, ucgridss& lands);
  void calcsummary(const cartiter& t);
  void dumpsummary(ostream& os);

  void drop(dgrid& traingrid, dgrid& targetgrid,
	    const cartiter& t, int numdates, int& yescount, int& nocount);
  void clearhists(const cartiter& curr);
};

typedef vector<carttree> carttrees;

// __________________________________________________________________________
// carttree.h
#endif
