// Copyright YALI AMIT, January 15, 2001.

#include <iostream>
#include <cstdio>
#include "timer.h"
#include "includes.h"
#include "image.h"
#include "trainstat.h"
#include "detection.h"
#include "ltatree.h"
#include "viewgrid.h"
#include "viewport.h"
#include "globals.h"
#include "wave.h"
#include "pixel.h"
#include "detfuncs.h"

double Norm(dgrid a);
double Angle(dgrid a);
dgrid pixratio(Pixel a, Pixel b);
void initialize_curve(igrid &dm, Pixels &cf, viewport &vp);
void dynamic(Pixels &cm,int bsize,image &im,int dis, detection &det);
void makepoint(image &im, int dis, landmarks& edgelands);
void Psi(Pixel p0, Pixel p1,double h,Pixel  m0,Pixel  m1, double lp, double  lpb, double  ap , double apb, dgrid &vals);
void update_graphics(viewport &vp, Pixelss& S, igrid &I, dgrids &H, int k);

// The eight nearest neighbor directions. Each segment of curve
// can have one of these 8 directions.
int snn[8][2] = { {0, 1}, {1, 1}, {1, 0}, {1,-1}, {0, -1}, {-1, -1}, {-1, 0}, {-1, 1}};

double po = .95; // Probability of finding feature on curve
double pb = .05; // Probability of finding feature off curve
double lp, lpb, ap, apb;

// Run the dynamic programming with initialization from user.
void init_dynamic(image &im)
{
  igrid m; 

  // Size of the `S' regions for the state space of each point on the curve.
  int bsize=dyna_regionsize;
  // Distance orthogonal to curve for comparisons in evaluating presence
  //  of feature
  int dis=dyna_dis;

  im.blur_downscale(ratio,sd);

  image tempim=im;
  tempim.normalize(imrange*numcolors,(1-imrange)*numcolors-1);
  
  
  viewport vp;
  vp.add_grid(tempim);
  // Either choose an initial curve by pointing with the mouse  
  cout << tempfile << endl;
  if (tempfile=="")
    {
      vp.add_string(0,0,"CHOOSE A CURVE");
      vp.view();
      vp.get_marked_points(m);
      m.write("dyna.temp");
    }
  // Or use a curve stored in file `tempfile'
  else
    {
      if (!m.read(tempfile))
	{
	  cerr << " Can't find the template file " << tempfile << endl;
	  exit(0);
	}
    }
  
  Pixels cf(m.dimx());
  // Get two point from user which determine a scale, translation and rotation
  // of the model curve. This becomes the initial curve used to determine the
  // regions.  
  initialize_curve(m, cf, vp);
  detection det;

  // Start dynamic programming
  dynamic(cf, bsize, im, dis, det);
  if (detfile!="")
    {
      ofstream ofs(detfile);
      det.write(ofs);
    }
}

// Get 2 points from user to initialize curve.
void initialize_curve(igrid &dm, Pixels &cf, viewport &vp)
{
 
  Pixels cm(cf.dim());
  for (int i=0; i<dm.dimx(); i++)
    cm(i)= Pixel(dm(i,0),dm(i,1)); 
  
  // Get first and last points of initial curve
  igrid z;
  if (point1x < 0 || point1y < 0 || point2x < 0 || point2y < 0)
    {
      vp.view();
      vp.get_marked_points(z);
      if (z.dimx()!=2)
	{
	  cout << " I didn't get two points " << endl;
	  exit(0);
	}
    }
  else
    {
      z.init(2,2);
      z(0,0)=point1x; z(0,1)=point1y; z(1,0)=point2x; z(1,1)=point2y;
      vp.add_point(z(0,0),z(0,1),basecol,5);
      vp.add_point(z(1,0),z(1,1),basecol,5);
      if (dum1!=13 && dum1!=1)
	vp.view();
    }
  Pixels cz(2);
  cz(0) = Pixel(z(0,0),z(0,1));
  cz(1) = Pixel(z(1,0),z(1,1));
  // transform initial curve to new endpoints
  
  dgrid cfac = pixratio(cz(1)-cz(0),cm(dm.dimx()-1)-cm(0));
  
  for (int i=0; i<cm.dim(); i++)
    {
      cf(i)[0] = cz(0)[0] + (int)rint((cm(i)-cm(0))[0]*cfac(0)
				      - (cm(i)-cm(0))[1]*cfac(1));
      cf(i)[1] = cz(0)[1] + (int)rint((cm(i)-cm(0))[0]*cfac(1) 
				      + (cm(i)-cm(0))[1]*cfac(0));
      
    }

}


// Initialize the algorithm with parameters from a detection.
// An affine map taking `ltatree1, ltatree2, ltatree3' to
// det.tripts.
void init_dynamic(image &im, detection &det)
{
  int lshiftx = (ltatree1x + ltatree2x + ltatree3x)/3;
  int lshifty = (ltatree1y + ltatree2y + ltatree3y)/3;
  partialtrans pt;
  pt.ptr(Pixel(ltatree1x-lshiftx, ltatree1y-lshifty), 
	 Pixel(ltatree2x-lshiftx, ltatree2y-lshifty),
	 Pixel(ltatree3x-lshiftx, ltatree3y-lshifty));
  
  
  affinetrans at(pt, Pixel(det.tripts[0], det.tripts[1]),
		 Pixel(det.tripts[2], det.tripts[3]),
		 Pixel(det.tripts[4], det.tripts[5]));
  
  igrid dm;
  
  if (!dm.read(tempfile))
    {
      cerr << "Couldn't find template file " << tempfile << endl;
      exit(0);
    }
  
  
  dm.dump();
  double cx=0, cy=0;
  for (int i=0; i<dm.dimx(); i++)
    {
      cx+=dm(i,0);
      cy+=dm(i,1);
    }
  cx = cx/dm.dimx();
  cy = cy/dm.dimx();
  
  for (int i=0; i<dm.dimx(); i++)
    {
      double d0 = at.coefs[0][0]*(dm(i,0)-cx)*post+
	at.coefs[0][1]*(dm(i,1)-cy)*post+at.coefs[0][2] + shiftx;
      double d1 = at.coefs[1][0]*(dm(i,0)-cx)*post+
	at.coefs[1][1]*(dm(i,1)-cy)*post+at.coefs[1][2] + shifty;
      dm(i,0)=int(rint(d0));
      dm(i,1)=int(rint(d1));
      
    }  
  
  int bsize=dyna_regionsize;
  int dis=dyna_dis;
  Pixels cm(dm.dimx());
  for (int i=0; i<dm.dimx(); i++)
    cm(i)= Pixel(dm(i,0),dm(i,1)); 

  // Start dynamic programming
  dynamic(cm,bsize,im,dis, det);
}



// Dynamic progamming. Structure det.locs gets OVERWRITTEN with final
// curve.
void dynamic(Pixels &cf, int bsize,image &im, int dis, detection &det)
{
  // cf - model curve
  // bsize - size of area around initial points
  // im - image
  // dis - distance for definition of point processes
  
  // nearest neighbors of the origin
  viewport vp;
  image tempim=im;
  tempim.normalize(imrange*numcolors,(1-imrange)*numcolors-1);
  vp.add_grid(tempim);
  viewport vp1(tempim);

  // Probability p_o for finding the local feature of the correct
  // angle on curve, p_b the probability of finding any other local feature.
  po=inside;
  pb=outside;
  lp = log(po);
  lpb = log(pb);
  ap = log(1-po);
  apb = log(1-pb);
  
  // cf - model configuration
  int length=cf.dim();
  
  
  // Show initial curve.
  if (dum1==13)
    {
      for (int i=0; i<length;i++)
	{
	  if (i>0)
	    vp1.add_line(cf(i-1)[0],cf(i-1)[1],cf(i)[0],cf(i)[1],basecol+1);
	  vp1.add_point(cf(i)[0],cf(i)[1],10,basecol+1);
	}
      vp1.view();
    }
  
  //Get eight point processes.
  landmarks edgelands;
  makepoint(im, dis, edgelands);
  
  
  //Initialize array for recording results
  int maxnum = (2*bsize+1)*(2*bsize+1)+5;
  
  // H(k) arrays keeps track of optimal choice and cost for
  // each candidate for step k.
  dgrids H(length);
  // Regions defining state space for each step.
  Pixelss S(length);
  igrid p(length);
  p.clear(0);
  
  //  cout << " maxnum " << maxnum << endl;
  for (int i=0; i<length; i++)
    {
      S(i).init(maxnum);
      H(i).init(maxnum,5);
      H(i).clear(-10000);
    }
  // Define `S' regions for search. `length' regions for each
  // point along curve.
  for (int x=0; x<im.dimx(); x++)
    for (int y=0; y<im.dimy(); y++)
      {	
	for (int i=0; i<length; i++)	
	  {			
	    if ( fabs(x - cf(i)[0]) < bsize && fabs(y-cf(i)[1]) < bsize)
	      {
		S(i)(p(i)) = Pixel(x,y);
		p(i) += 1;
              }
	  }							
      }	
  
  igrid I(length);
  dgrid V(length);
  // Start dynamic programming iterations
  for (int k=length-2; k>=0; k--)
    {
      cout << " K " << k << " " << p(k) << endl;
      timer timer1;
      timer1.start("Step");
      for (int s=0; s<p(k); s+=2)
	{
	  
	  dgrid oldval(4), val(4);
	  oldval(0) = -10000;
	  oldval(1) = 0;
	  int oldind = -1;
	  // Find best point in S(k+1) for point `s' in S(k).
	  for (int t=0; t<p(k+1); t+=2)
	    {
	      double h=0;
	      if (k < length-2)
		h = H(k+1)(t,1);
	      // Evaluate cost for point `t' in S(k+1), need to add in
	      // cost assigned to `t' passed in `h'. cost is returned 
	      //in `val(0)'. Other entries in val are the prior component
	      // and the two components of the data term. Not used.
	      Psi(S(k)(s), S(k+1)(t), h, cf(k), cf(k+1), 
	      	                     lp, lpb, ap, apb, val);
	      
	      if (val(0) > oldval(0))
		{
		  oldval=val;
		  oldind = t;
		}
	    }
	  // Nothing was found
	  if (oldind==-1)
            return;

	  // Update information in H array for point `s'.
	  H(k)(s,0) = oldind;
	  H(k)(s,1) = oldval(0);
          H(k)(s,2) = oldval(1);
          H(k)(s,3) = oldval(2);
          H(k)(s,4) = oldval(3);
	  
	}
      
      int indk=-1;
      double valk = -10000;

      // Find optimal curve at this point by maximizing of H(k)(.,1). 
      for (int a=0; a<p(k); a++)
	{
	  if (H(k)(a,1)>valk)
	    {
	      valk=H(k)(a,1);
	      indk=a;
	    }
	}
      //      
      I(k)=indk;
      V(k)=valk;
      cout << " Step " << k << " Total " << H(k)(indk,1) << " On " << 
	H(k)(indk,2) << " Off " << H(k)(indk,3) << 
	" Prior " << H(k)(indk,4) << endl;
      // Show optimal curve at this point in iteration.
      if (dumpdetections>=3)
	update_graphics(vp, S, I, H, k);
      
      timer1.stop("End Step");
    }
  // End dynamic programming iterations
  

  
  int indk=-1;
  double valk = -10000;
  int k=0;
  // Find optimal curve at the end.
  for (int a=0; a<p(k); a++)
    {
      if (H(k)(a,1)>valk)
	{
	  valk=H(k)(a,1);
	  indk=a;
	}
    }
  
  I(k)=indk;
  V(k)=valk;

  det.locs.init(length,2);
  //  Store optimal curve in det.locs.
  for (int a = k; a<length; a++)
   {
      if (a<length-1)
   	I(a+1) = (int)(H(a)(I(a),0));
      det.locs(a,0)=S(a)(I(a))[0];
      det.locs(a,1)=S(a)(I(a))[1];
   } 

  

  
  cout << " Final " << endl;
  if (dumpdetections >= 1)
    {
      update_graphics(vp1, S, I, H, 0);
    }
  empty_STAGS(edgelands);  
}

// Compute 4 features everywhere in the image. Locations are stored
// in binary arrays STAGS.
void makepoint(image &im, int dis, landmarks &edgelands)
{
  void detectedges(const image& im, landmarks& lmarks);
  double locedgetype=1;
  // Preserve existing edge parameters (the curve tracking routine may be called from
  // the sparse model detection routine.
  int locedgesizeblock=EdgeSizeBlock;
  int locedgenumberofpixels=EdgeNumberOfPixels;
  swap(EdgeType,locedgetype);
  EdgeRadius=dis;

  int ln = 8;

  STAGSdim=MAX(im.dimx(), im.dimy());
  detectedges(im, edgelands);
  swap(EdgeType,locedgetype);
  EdgeSizeBlock=locedgesizeblock;
  EdgeNumberOfPixels=locedgenumberofpixels;
  fill_STAGS(edgelands,0);
 
  // See the point processes.
  if (dum1==1)
    {
      viewports vp(ln);
      for (int k=0; k<ln; k++)
	{
	  int oldrevv=revvideo;
	  revvideo=0;
	  view_normalized(STAGS(k));
	  revvideo=oldrevv;
	}
    }
  
}

// Compute the cost of a particular segment, interms of deviation
//from model segment p0-p1, and log-probabilities for
// features on and off segment
void Psi(Pixel p0, Pixel p1,double h,Pixel  m0,Pixel  m1, double lp, double  lpb, double  ap , double apb, dgrid& vals)
{
  
  // p0 p1 points in image
  Pixel dp = p1-p0;
  
  // length of segment.
  double len=dp.Norm();
  if (len<1)
    {
      vals(0) = -10000;
      return;
    }
  
  Pixel dm = m1-m0;
  double lenm = dm.Norm();
  
  if (lenm==0)
    {
      vals(0)=-10000;
      return;
    }
  
  dgrid dpmr = pixratio(dp,dm);
  double dan=fabs(Angle(dpmr));
  
  // Prior computation. 
  double A = prior_angle;
  double B = prior_length;
  double tval =  B*fabs(log(Norm(dpmr))) + A*fabs(dan);
  
  double ang = dp.angle();
  if (ang<0)
    ang = 2*PI + ang;
  
  // Get index of the angle of the model segment p0-p1.
  double an=(ang*8)/(PI2);
  int kk = MIN((int)rint(an),7);
  kk= ((kk+2) % 8);
  

  // EXPERIMENT
  kk=kk % 4;
  int Ilen = int(len)+1;
  
  double unitdp[2];
  unitdp[0]=dp[0]/len;
  unitdp[1]=dp[1]/len;
  
  int oldrep = -1;
  int oldimp = -1;
  
  igrid count(8);
  count.clear(0);
  int ll  = 0;
  
  double p[2];
  p[0] = p0[0];
  p[1] = p0[1];
  
  // Loop on points along the segment and get counts of
  // features of type `kk' in STAGS(kk).
  for (int a=0; a < Ilen+1; a++)
    {
      int rep = (int)rint(p[0]);
      
      int imp = (int)rint(p[1]);
      if (rep != oldrep || imp != oldimp)
	{
	  count(kk) = count(kk) + STAGS(kk).get(rep,imp);
	  ll += 1;
	}
      oldrep = rep;
      oldimp = imp;
      p[0] = p[0] + unitdp[0];
      p[1] = p[1] + unitdp[1];
    }
  
  //cout << " count " << count(kk) << endl;
  
  vals(1) = count(kk)*(lp - lpb);
  
  vals(2) =  (ll-count(kk))*(ap-apb);
  
  vals(3) = tval;
  
  vals(0)=vals(1)+vals(2)-vals(3) + h;
 
}

// Some basic analytic geometry, could be done with the `complex' class.
// Ratio of 2 complex numbers.
dgrid pixratio(Pixel a, Pixel b)
{
  double l = b[0]*b[0] + b[1]*b[1];
  dgrid ans(2);
  ans(0) = (a[0]*b[0]+a[1]*b[1])/l;
  ans(1) = (a[1]*b[0]-a[0]*b[1])/l;
  return(ans);
}


double Angle(dgrid a)
{
return(atan2(a(1),a(0)));
}

double Norm(dgrid a)
{
return(sqrt(a(0)*a(0)+a(1)*a(1)));
}


// Show current optimal curve. Recovered recursively by finding
// optimal match to current point.
void update_graphics(viewport &vp, Pixelss& S, igrid &I, dgrids &H, int k)
{
  
      if (dumpdetections==4)
	vp.clear_geom();
      vp.add_point(S(k)(I(k))[0],S(k)(I(k))[1],basecol,5);
      for (int a = k; a<S.dim()-1; a++)
	{
	  I(a+1) = (int)(H(a)(I(a),0));
	  vp.add_line(S(a)(I(a))[0],S(a)(I(a))[1],
		      S(a+1)(I(a+1))[0],S(a+1)(I(a+1))[1], basecol);
	  vp.add_point(S(a+1)(I(a+1))[0],S(a+1)(I(a+1))[1], basecol,5);
	  //V(a+1) = H(a)(I(a),1);
	  
	  //cout << a << " " << H(a)(I(a),1) << " " << H(a)(I(a),2);
	  //cout << " " << H(a)(I(a),3) << " " << H(a)(I(a),4)<<endl;
	  //cout << S(a+1)(I(a+1)) - S(a)(I(a)) << endl;
	  //cout << " model " << cf(a+1)-cf(a) << endl;
	  
	}
      vp.view();

}


// curve_dyna
