// Copyright YALI AMIT, January 15, 2001.

#include <iostream>
#include <cstdio>

#include "includes.h"
#include "image.h"
#include "trainstat.h"
#include "detection.h"
#include "ltatree.h"
#include "viewgrid.h"
#include "viewport.h"
#include "globals.h"
#include "wave.h"

#define log2(a) log((double)a)/log(2.)
extern ucgrids STAGS;
void init_STAGS(int size, int dimx, int dimy);
void fill_STAGS(const landmarks& lm, int spread = 0, int numt = -1);
void empty_STAGS(const landmarks& lm);

double lamfun(int i, int j);
void init_twod( dgrids& itemp, dgrids& idat);
void calches();
void step_AFFINE( dgrids& itemp, dgrids& idat);
void incdims();
void xrestore1(int niters, dgrids& itemp, dgrids& idat, viewport &vpim);
void plotuv1(viewport &vp, dgrid &fx, dgrid &fy, int jump, int defpixsize);
void blur_subsample(dgrid &th, double ratio, double sd);
void twod_els();
void match_points(dgrid &u, dgrid &v, int bs1,int bs2, igrid &in, igrid &out);
 void full_search(dgrids &itemp, dgrids &idat, viewport &vpim);
double LIKEs(dgrids &data, dgrids &ltmp1, dgrids &ltmp0, double sx, double sy, int ix, int jy);
double LIKEr(dgrids &data, dgrids &ltmp1, dgrids &ltmp0, double sx, double sy, int ix, int jy);
void show_stuff(viewport &vpa, viewport &vp_field);
int trans_scale(int x, int ix, double sx, int cx, int fwd);
void detectedges(const image& im, landmarks& lmarks);
extern dgrid sumlam;
dgrids appls, appls1, lapps0, lapps1, stuff;
dgrid defs_gradx, defs_grady;
dgrids itemp_gradx, itemp_grady;

dgrid fields_x, fields_y, affines_x, affines_y;
dgrid lamdas_x, lamdas_y;
dgrids itemp, idat;

// lmarkslist fr_lands;
// int fr_index;

double scalex,scaley,transx,transy;
int xdim, ydim, xdimh, ydimh;
extern int dim_curr, ND, fd, size_dat;

int numtemps;

igrid in;

double sdtcfs=1.;
int DISPLAYNs=10;
double initscalex=1., initscaley=1.;
extern double HESS;
extern double dt;
double invdx, invdy;
double lg_one_minus_minval, lg_minval;
extern double xxxo;
extern int L;
extern dgrid freq_x, freq_y, lamda_x, lamda_y;
int nnn[8][2] = { {0, -1}, {0, 1}, {-1,0}, {1,0}, {1, -1}, {-1,1}, {1, 1}, {-1, -1}};
  void show_imvec(dgrids &dvec, viewport &lvp, int ref);
// ___________________________________________________________________________

// Run the Bernoulli model on an image
void twod_els(image &im)
{
  int dmx = defsizex;
  int dmy = defsizey;
  xdim = dmx;
  ydim = dmy;
  xdimh=xdim/2;
  ydimh=ydim/2;
  size_dat=xdim;
  spread=1;
  // Hardwired edge parameters (can be removed)
  int er=4, es=1, ei=1;
  double ee=.03;
  swap(er,EdgeGoodTests);swap(es,EdgeSizeBlock);
  swap(ei,edgeinc);swap(ee,EdgeEpsilon);
  landmarks lands;

  // Get edges only inside the region of interest
  image imt(2*eyex,2*eyey);
  imt.subcopy(xdimh-eyex,xdimh+eyex,ydimh-eyey,ydimh+eyey,im,0,0);
  detectedges(imt, lands);
  
  // Find lower and upper bounds of edge data
  int globalminx = 1000000, globalminy = 1000000, 
    globalmaxx = -1000000, globalmaxy = -1000000;
  for (int j = 0; j < lands.total_instances(); ++j)
  {
    // Readjust the edge locations in middle of the bigger defsizex by
    // defsizey lattice
    lands.location_x(j)+=(xdimh-eyex);
    lands.location_y(j)+=(ydimh-eyey);
    globalminx = MIN(globalminx, lands.location_x(j));
    globalminy = MIN(globalminy, lands.location_y(j));
    globalmaxx = MAX(globalmaxx, lands.location_x(j));
    globalmaxy = MAX(globalmaxy, lands.location_y(j));
  }

  // FIll the global buffers STAGS with the detected edges, with spread
  fill_STAGS(lands,spread);

  ifstream ifs(tempfile);

  // Get 8 templates.
  numtemps=8;
  itemp.init(numtemps);
  idat.init(numtemps+1);

  int dxt,dyt;
  int dxd = int(dmx - (globalmaxx - globalminx))/2;
  int dyd = int(dmy - (globalmaxy - globalminy))/2;
  lg_one_minus_minval=log(1-minval);
  lg_minval=log(minval);
  // Loop over numtemps frequency maps of edges.
  for (int s=0; s<numtemps; s++)
  {
    dgrid tt;
    // Read in frequency map for edge type `s'.
    tt.read(ifs);

    //    cout << tt.dimx() << " " << tt.dimy()  << endl;
    // Embedd templates inside facrtsizex by fcartsizey lattice
    dxt = int(dmx-tt.dimx())/2;
    dyt = int(dmy-tt.dimy())/2;
    //cout << dxt << " " << dyt << endl;
    itemp(s).init(dmx,dmy);
    idat(s).init(dmx,dmy);
    itemp(s).clear(minval);

    // Copy from STAGS(s) into idat(s), with possible shift of

    gridfun( ( ( fabs(i-xdimh)<eyex && fabs(j-ydimh)<eyey) ?  (STAGS(s).get(i-dxd+globalminx,j-dyd+globalminy)>0) : 0 ),idat(s));

    // Set minimum and maximum value of edge frequencies at minval maxval.
    for (int i=0; i<tt.dimx(); i++)
      for (int j=0; j<tt.dimy(); j++)
          itemp(s).set(dxt+i, dyt+j,MAX(minval,MIN(tt(i,j),maxval)));
  }

  empty_STAGS(lands);
  idat(numtemps).init(dmx,dmy);
  gridfun( ( ( fabs(i-xdimh)<eyex && fabs(j-ydimh)<eyey) ? im.get(i-dxd+globalminx,j-dyd+globalminy) : 0 ),idat(numtemps));

  // Run algorithm on `itemp' `idat'.
  twod_els();
  // Return original edge parameters.
  swap(er,EdgeRadius);swap(es,EdgeSizeBlock);
  swap(ei,edgeinc);swap(ee,EdgeEpsilon);

}

// Display a vector of grids.
void show_imvec(dgrids &dvec, viewport &lvp, int ref)
{
  ucgrid uct(xdim*numtemps, ydim);
  ucgrid ut(xdim,ydim);
  for (int t=0; t<numtemps; t++)
    {
      //Show truncated template to smaller region around center
     
      gridfun((uchar)(dvec(t)(i,j)*(numcolors-1)),ut);
      uct.subcopy(0,xdim,0,ydim,ut,t*xdim,0);
      
    }
  uct.normalize(imrange*numcolors,(1-imrange)*numcolors-1);

  
  lvp.add_grid(uct);
  for (int t=0; t<numtemps; t++)
    {
      lvp.add_line((t+1)*xdim,0,(t+1)*xdim,ydim,basecol);
    }
 
    lvp.refresh(ref);
}

void twod_els()
{

 
  viewport vpdata, vptemp, vpim, vpstemp;

  //Show data and frequency maps.

  if (opt!=230) {
    show_imvec(itemp, vptemp,1);
    show_imvec(idat, vpdata,1);
  }

  // The image itself
  ucgrid ut(xdim,ydim);
  idat(numtemps).normalize(0.,1.);
  gridfun((uchar)(idat(numtemps)(i,j)*(numcolors-1)),ut);
  ut.normalize(imrange*numcolors,(1-imrange)*numcolors);
  vpim.add_grid(ut);
  

  // First component of non-truncated template (just for showing).
  ucgrid ut1(xdim, ydim);

  gridfun((uchar)((itemp(0)(i,j)+itemp(1)(i,j)+itemp(2)(i,j)+itemp(3)(i,j)
                  +itemp(4)(i,j)+itemp(5)(i,j)+itemp(6)(i,j)+itemp(7)(i,j))
		  *(numcolors/8)),ut1);
  vpstemp.add_grid(ut1);
  if (opt!=230)
    vpstemp.view();

  // Display some template points
  if (opt!=230) {
    vpstemp.get_marked_points(in);
  }

  // If points are marked on image these are written to face_points.
  if (in.dimx()>0)
    in.write("face_points");
  // Other wise face points is read or if it doesnt exist,
  // The high freuqncy points in the edge frequency maps are taken.
  else
    {
      if(!in.read("face_points"))
	{
	  igrid tin(ut1.dimx()*ut1.dimy(),2);
          int p=0;
	  for (int x=0; x< ut1.dimx(); x++)
	    for (int y=0; y<ut1.dimy(); y++)
	      {
		if (itemp(0)(x,y) > objprop ||
		    itemp(1)(x,y) > objprop ||
		    itemp(2)(x,y) > objprop ||
		    itemp(3)(x,y) > objprop ||
		    itemp(4)(x,y) > objprop ||
		    itemp(5)(x,y) > objprop ||
		    itemp(6)(x,y) > objprop ||
		    itemp(7)(x,y) > objprop)
		  {
                    tin(p,0)=x; tin(p,1)=y;
		    p++;
		  }
	      }
	  in.init(p,2);
	  gridfun(tin(i,j),in);
	}
    }
 
  for (int i=0; i<in.dimx(); i++)
    {
      in(i,0)=int(double(in(i,0)));
      in(i,1)=int(double(in(i,1)));
      vpim.add_point(in(i,0),in(i,1),basecol,5);
    }
  
  // Prepare wavelet basis.
  makevec(base);
  if (opt!=230)
    vpim.view();
  vpim.clear_geom();
 
  init_twod(itemp, idat);

  fields_x.clear();
  fields_y.clear(); 
 
  // Loop through a range of scale and translation parameters
  // with direct computation of cost function to find optimum
  // The translation and scale are given as displacement fields
  // field_x, field_y.

 if (shift>0)
  full_search(itemp,idat,vpim);

  freq_x = fields_x;
  wavep2(freq_x,1); 
  freq_y= fields_y;
  wavep2(freq_y,1);

  void xrestores();
  // Start gradient descent on non-linear deformations.
  if (numiters>0)
    {
      xrestores(); 
      igrid out=in;
      
      // Show where face_points are mapped by final field
      match_points(fields_x, fields_y, 0, 0, in, out);
      vpim.clear_geom();  
      for (int p=0; p<out.dimx(); p++)
	vpim.add_point(out(p,0),out(p,1),basecol,5);
      
      if (opt!=230)
	vpim.view();
      
    }
}


// ___________________________________________________________________________
int trans_scale(int x, int ix, double sx, int cx, int fwd)
{
  if (fwd)
    return( int(rint( (x-cx)*sx + ix + cx)) );
  else
    return( int(rint( double(x-ix-cx)/sx + cx)));
}

// log-Likelihood of a scale and translation.
double LIKEs(dgrids &data, dgrids &ltmp0, dgrids &ltmp1, double sx, double sy, int ix, int jy)
{
  double xx=0;
  
  // A fixed window of the data is inspected and compared
  // to scaled and translated versions of the template.
  for (int x=-eyex; x<eyex; x++)
    for (int y=-eyey; y<eyey; y++)
      {
	int sinx=trans_scale(x, ix, sx , 0, 1) + xdimh;
	int siny=trans_scale(y, jy, sy , 0, 1) + ydimh;

	if (sinx >= 0 && sinx < xdim && siny >=0 && siny < ydim)
	  {
	    for (int s=0; s<numtemps; s++)
	      {
		int ld =  (int)(data(s)(xdimh+x,ydimh+y));
		xx -= (ld ? ltmp0(s)(sinx,siny) : ltmp1(s)(sinx,siny));
	      }
	  }
        else
	{
	  for (int s=0; s<numtemps; s++)
	      {
		int ld =  (int)(data(s)(xdimh+x,ydimh+y));
		xx -= (ld ? lg_minval : lg_one_minus_minval);
	      }
	}
	
      }
  
  return(xx);
}

double LIKEr(dgrids &data, dgrids &ltmp0, dgrids &ltmp1, double sx, double sy, int ix, int jy)
{
  double xx=0;
  
  // A fixed window of the data is inspected and compared
  // to scaled and translated versions of the template.
  for (int x=-eyex; x<eyex; x++)
    for (int y=-eyey; y<eyey; y++)
      {
	int sinx=trans_scale(x, ix, sx , 0, 1) + xdimh;
	int siny=trans_scale(y, jy, sy , 0, 1) + ydimh;

	if (sinx >= 0 && sinx < xdim && siny >=0 && siny < ydim)
	  {
	    for (int s=0; s<numtemps; s++)
	      {
		int ld =  (int)(data(s)(xdimh+x,ydimh+y));
		xx -= (ld ? ( ltmp0(s)(sinx,siny)-lg_minval ) : 
		            ( ltmp1(s)(sinx,siny)-lg_one_minus_minval ) );
	      }
	  }
      }
  
  return(xx);
}

// ___________________________________________________________________________


// Brute force search through scale and translation.
void full_search(dgrids &itemp, dgrids &idat, viewport &vpim)
{
  double outold=1000000.;
  double sxold=1., syold=1.;
  int iold=0,jold=0,iold2=0,jold2=0;
  int shifta = shift/2;

  image tim(xdim,ydim);
  tim.clear(0);
  

  // Prepare the log-probability maps
  for (int s=0; s< numtemps; s++)	
  {
    gridfun(log(1.-itemp(s)(i,j)),appls1(s));
    gridfun(log(itemp(s)(i,j)),appls(s));
  }

  // Loop over a range of translations and find optimum.
  for (int i=-shift; i<= shift; i+=1)
    for (int j= -shift; j<= shift; j+=1)
      {
	double out=LIKEr(idat,appls, appls1, 1., 1., i, j);
	if (out < outold)
	  { outold = out; iold = i; jold = j; }
      }
  
  //  cout << "new i j out " << iold << " " << jold << " " << outold << endl;
  outold=100000; 

  // Loop over a range of scales in x and y.
  for (double sx=lowscale; sx <= highscale+.01; sx+=.1)
    for (double sy=lowscale; sy <= highscale+.01; sy+=.1)
    {
      double out=LIKEr(idat,appls, appls1, sx, sy, iold, jold);
      //      cout << " sx sy cost " << sx << " " << sy << " "  << out << endl;
      if (out < outold)
	{ outold = out; sxold = sx; syold = sy;}
    }

  
  cout << "new sx sy out " << 1./sxold << " " << 1./syold << " " << outold << endl;

 outold=100000;
 // Reloop through a range of translations
 for (int i=iold-shifta; i<= iold+shifta; i+=1)
    for (int j= jold-shifta; j<= jold+shifta; j+=1)
      {
	double out=LIKEr(idat,appls, appls1, sxold, syold, i, j);
	if (out < outold)
	  { outold = out; iold2 = i; jold2 = j; }
      }

 iold=iold2;
 jold=jold2;
 
 cout << "new i j out " << iold << " " << jold << " " << outold << endl;
 if (in.dimx()==0)
   return;

 // Show where face_points are mapped. 
 for (int p=0; p<in.dimx(); p++)
   {
     int x1 = trans_scale( in(p,0)-xdimh, iold, sxold, 0, 0) + xdimh;
     int y1 = trans_scale( in(p,1)-ydimh, jold, syold, 0, 0) + ydimh;
     vpim.add_point(x1,y1,basecol,5);
   }
 if (opt!=230) 
   vpim.view();
 
 // create the field from the optimal scale and translation.
 gridfun( double(trans_scale(i-xdimh, iold, sxold, 0, 1)+xdimh-i)/xdim, fields_x);
 gridfun( double(trans_scale(j-ydimh, jold, syold, 0, 1)+ydimh-j)/ydim, fields_y);
 
} // full_search


// Likelihood given a deformation of the probability maps.
double costs()
{
  double xx=0;
  for (int s=0; s<numtemps; s++)
  {
  for (int x=xdimh-eyex; x<xdimh+eyex; x++)
    for (int y=ydimh-eyey; y<ydimh+eyey; y++)
      {
        xx -= (  (int)( idat(s)(x,y)) ? lapps0(s)(x,y) : lapps1(s)(x,y));
      }
  }
  return(xx);
}


// Gradient descent.
void xrestores()
{

 void plotuv1(viewport &vp, dgrid &fx, dgrid &fy, int jump, int defpixsize);


  
if (FIRSTDIM==0)
 FIRSTDIM=1;
if (post==0)
 post=1;
 dim_curr=FIRSTDIM;
 // Adjust time step.
  ND =  xdim-dim_curr;
  fd = L-dim_curr;
  dt = dt_scale_factor/(dim_curr*dim_curr*post*HESS+sumlam(dim_curr));
  cout << " Dim - " << dim_curr << " time-step - "<< dt << endl;
  
  int defpixsize=2;  

  viewport vpa;  
  viewport vpa1;
  viewport vp_field(defpixsize*fields_x.dimx()*10/8,defpixsize*fields_x.dimy()*10/8);

  for(int II=1; II<=numiters; II++)
    {
      double xxx;
      cout << " II " << II << " dimcurr " << dim_curr << endl;
      
      // Get the deformed log-probability maps.
      for (int s=0; s< numtemps; s++)
	{
	  applyuv_smt1(fields_x, fields_y, itemp(s), appls(s), 0, 0);
	  gridfun(log(appls(s)(i,j)),lapps0(s));
	  gridfun(log(1.-appls(s)(i,j)),lapps1(s));
	}
        
      if ( (II % displayinter) == 0)
	if (opt!=230)
	  show_stuff(vpa, vp_field);
          
	
      
      // Check least squares
      double costs();
      xxx = costs();
      printf("energy - %3.3f \n",xxx);

      void incdim();
      // If change in cost small increase dimension
      if (II>1 && fabs(xxx-xxxo)/xxxo < minenergychange) 
	{
	if (dim_curr < L) incdim();
        else if (xxx-xxxo > 0)
	  { 
	    if (opt!=230) {
	      show_stuff(vpa, vp_field);
	      getchar();
	    }
	    return;
	  }
	}
      xxxo = xxx;	      
      void step_FREQs();
      // One iteration on wavelet coeficients.
      step_FREQs();
      
      // Generate the field at the required points
      fields_x.clear();
      fields_x.subcopy(ND,size_dat,ND,size_dat,freq_x,ND,ND); 
      unwavep2(fields_x,1);
      fields_y.clear();
      fields_y.subcopy(ND,size_dat,ND,size_dat,freq_y,ND,ND); 
      unwavep2(fields_y,1);
    }
}

// SHow deformation field 
void show_stuff(viewport &vpa, viewport &vp_field)
	    {
	      int defpixsize=2;
	      int oldpixsize=1;

	      ucgrid ut(xdim,ydim);         
	      swap(pixsize,oldpixsize);
	      pixsize=4;
	      gridfun((uchar)((appls(0)(i,j)*1.+idat(0)(i,j)*0.)*(numcolors-1)),ut);
	      vpa.add_grid(ut);
	      vpa.refresh();
	      pixsize=1;
	      plotuv1(vp_field, fields_x, fields_y, arrow_spacing, defpixsize);
	      vp_field.view();
	      swap(pixsize,oldpixsize);
	    }   

// Compute gradient in coeficients and update
void step_FREQs()
{
  
  // Apply deformation to gradient
  // Calculate Horizontal and Vertical fields applied to  tempalte gradient
  
  dgrid addx(xdim,ydim);
  addx.clear();
  dgrid addy(xdim,ydim);
  addy.clear();
  
  // Loop over 8 edge types.
  for (int g=0; g<numtemps; g++)
  {
    applyuv_smt1(fields_x, fields_y,itemp_gradx(g), defs_gradx, 0,0);
    applyuv_smt1(fields_x, fields_y,itemp_grady(g), defs_grady, 0,0);
    gridfun(addx(i,j)+idat(g)(i,j)*defs_gradx(i,j)/appls(g)(i,j),addx);
    gridfun(addx(i,j)-(1-idat(g)(i,j))*defs_gradx(i,j)/(1-appls(g)(i,j)),addx);
    gridfun(addy(i,j)+idat(g)(i,j)*defs_grady(i,j)/appls(g)(i,j),addy);
    gridfun(addy(i,j)-(1-idat(g)(i,j))*defs_grady(i,j)/(1-appls(g)(i,j)),addy);
  }


  // Forward transform
  wavep2(addx,1);
  wavep2(addy,1);
  
  double xxx=0;
  // Update active coeficients between fd and L.
  for (int ii=fd; ii<L; ii++)
    for (int jj=fd; jj<L; jj++)
      {
	int iii = ii-fd+ND;
	int jjj = jj-fd+ND;
	xxx = -prior_fac*lamda_x(ii,jj)*freq_x(iii,jjj)+addx(iii,jjj)*post;
	freq_x(iii,jjj)+= xxx*dt;
	xxx = -prior_fac*lamda_y(ii,jj)*freq_y(iii,jjj)+addy(iii,jjj)*post;
	freq_y(iii,jjj)+= xxx*dt;
	
      }
}




void init_twod(dgrids &itemp, dgrids &idat)
{
  image im;
  viewport vp;

  if (LASTDIM==0)
    L = 3;
  else
    L = LASTDIM;
  // Initialize wavelet vector

  itemp_gradx.init(numtemps);
  itemp_grady.init(numtemps);
  appls.init(numtemps);
  appls1.init(numtemps);
  defs_gradx.init(xdim, ydim);
  defs_grady.init(xdim, ydim);
  lapps0.init(numtemps);
  lapps1.init(numtemps);
  stuff.init(numtemps);
  for (int s = 0; s < numtemps; s++)
  {
    appls(s).init(xdim,ydim);
    appls1(s).init(xdim,ydim);
    lapps0(s).init(xdim,ydim);
    lapps1(s).init(xdim,ydim);
    stuff(s).init(xdim,ydim);
    itemp_gradx(s).init(xdim,ydim);
    itemp_grady(s).init(xdim,ydim);
    // Gradient of probability maps.
    GRADIENT(itemp(s), itemp_gradx(s), itemp_grady(s));
  }
  calches();  


  // Field arrays
  fields_x.init(xdim,ydim);
  fields_x.clear();
  fields_y.init(xdim,ydim);
  fields_y.clear();

  // Wavelet Coeficient arrays
  freq_x.init(size_dat,size_dat);
  freq_x.clear();
  
  freq_y.init(size_dat,size_dat);
  freq_y.clear(); 
  
  // Prepare lamda multipliers in prior
  lamda_x.init(L+1,L+1);
  lamda_y.init(L+1,L+1);
  gridfun( lamfun(i,j),lamda_x);
  gridfun( lamfun(i,j),lamda_y);
  sumlam.init(L+1);
  sumlam(0) = 1;
  for (int i=1; i<L+1; i++) 
    sumlam(i) = prior_fac*lamda_y(i,i)+sumlam(i-1);
}

// ___________________________________________________________________________

// Subroutine for adhoc calculation of Hessian to determine timestep 
// Same routine as in regular Gaussian model.
void calches()
{
  double g = 0;
  int ubx = itemp_gradx(0).dimx();
  int uby = itemp_grady(0).dimx();
  for (int s=0; s<numtemps; s++)
    for (int i=0; i<ubx; i++)
      for (int j=0; j < uby; j++)
      {
        g += itemp_gradx(s)(i,j)*itemp_gradx(s)(i,j);
        g += itemp_grady(s)(i,j)*itemp_grady(s)(i,j);
      }

  HESS = g/(xdim*ydim);
  printf(" Hess - %g\n",HESS);
}

// Added by Elliot for face recognition project - 6/3/02
// Run the deformable image model and compute the landmarks on each face.
// The landmarks are stored in lands, and the number of detected landmarks
// is returned.
int fr_train(image& im, lmarklist& lands) {

  viewport vpim(im);
  int xGridDim, yGridDim;
  int gridNumPts;
  
  twod_els(im);
  
  im.normalize(imrange*numcolors, (1-imrange)*numcolors - 1);
  vpim.upd_grid();

  int xOffset=18;
  int yOffset=8;

  xGridDim = im.dimx()/gridSpacing;
  yGridDim = im.dimy()/gridSpacing;
  gridNumPts = xGridDim * yGridDim;

  cout << "xGridDim=" << xGridDim << endl;
  cout << "yGridDim=" << yGridDim << endl;
  
  igrid gridIn;
  gridIn.init(gridNumPts,2);

  for (int i=0; i<gridNumPts; i++) {
    gridIn(i,0) = (i%xGridDim)*gridSpacing + xOffset + gridSpacing / 2;
    gridIn(i,1) = (i/xGridDim)*gridSpacing + yOffset + gridSpacing / 2;
  }
  igrid gridOut=gridIn;
  match_points(fields_x,fields_y,0,0,gridIn,gridOut);

  // Extract the features
  
  // Set up an igrid with the cutoffs -- this should be changed to
  // import the cuts from a file.
  igrid cuts(numCuts);
//   cuts(0)=0.5;
//   cuts(1)=1.0;
//   cuts(2)=1.5;
//   cuts(3)=2.0;
//   cuts(4)=2.5;
  cuts(0)=1.5;
  cuts(1)=2.0;
  cuts(2)=2.5;
  
  //  lands.init(gridNumPts*4*numCuts);
  
  // Calculate the mean of all horizontal and vertical distances
  // for use as a reference.
  double refDist=0;
  int num=0;
  for (int i=0; i<gridNumPts; i++) {
    if (i%xGridDim<xGridDim-1) {
      refDist += sqrt((gridOut(i,0)-gridOut(i+1,0))
		      *(gridOut(i,0)-gridOut(i+1,0))
		      +(gridOut(i,1)-gridOut(i+1,1))
		      *(gridOut(i,1)-gridOut(i+1,1)));
      num++;
    }
    if (i/xGridDim<yGridDim-1) {
      refDist += sqrt((gridOut(i,0)-gridOut(i+xGridDim,0))
		      *(gridOut(i,0)-gridOut(i+xGridDim,0))
		      +(gridOut(i,1)-gridOut(i+xGridDim,1))
		      *(gridOut(i,1)-gridOut(i+xGridDim,1)));
      num++;
    }
  }
  refDist /= (double)num;
  cout << "refDist=" << refDist << endl;

  dgrid distances(4*gridNumPts); // Stores the distances between pairs of pts.
  igrid locations(4*gridNumPts,3); // Stores the pair coordinates & directions
  
  int count=0;
  numPairs=0;
  int pt2;
  for (int i=0; i<gridNumPts; i++) {
    for (int j=0; j<4; j++) {
      int OK=1;
      if ((j==0 || j==1) && (gridIn(i,0)+gridSpacing>=92+xOffset)) {
	OK=0;
      }
      if ((j==1 || j==2 || j==3) && (gridIn(i,1)+gridSpacing>=112+yOffset)) {
	OK=0;
      }
      if ((j==3) && (gridIn(i,0)-gridSpacing<xOffset)) {
	OK=0;
      }
      
      // Figure out the index of the next point in the grid
//       if (j==0) 
// 	pt2=i+1;
//       else if (j==1) 
// 	pt2=i+xGridDim+1;
//       else if (j==2) 
// 	pt2=i+xGridDim;
//       else if (j==3) 
// 	pt2=i+xGridDim-1;

//       if (gridOut(i,0)<0 || gridOut(i,1)<0 || 
//  	  gridOut(pt2,0)<0 || gridOut(pt2,1)<0 || 
// 	  gridOut(i,0)>=128 || gridOut(i,1)>=128 || 
// 	  gridOut(pt2,0)>=128 || gridOut(pt2,1)>=128) {
//  	OK=0;
//       }
      
      if (OK) {
	   
	if (j==0) 
	  pt2=i+1;
	else if (j==1) 
	  pt2=i+xGridDim+1;
	else if (j==2) 
	  pt2=i+xGridDim;
	else if (j==3) 
	  pt2=i+xGridDim-1;

	if (gridOut(i,0)<0 || gridOut(i,1)<0 || 
	    gridOut(pt2,0)<0 || gridOut(pt2,1)<0 || 
	    gridOut(i,0)>=128 || gridOut(i,1)>=128 || 
	    gridOut(pt2,0)>=128 || gridOut(pt2,1)>=128) {

	  locations(numPairs,0)=-1;
	  locations(numPairs,1)=-1;
	  locations(numPairs,2)=-1;
	  distances(numPairs)=-1;
	}
	else {
	  
	  distances(numPairs)=sqrt((gridOut(i,0)-gridOut(pt2,0))
				   *(gridOut(i,0)-gridOut(pt2,0))
				   +(gridOut(i,1)-gridOut(pt2,1))
				   *(gridOut(i,1)-gridOut(pt2,1)));
	  locations(numPairs,0)=gridIn(i,0)-xOffset;
	  locations(numPairs,1)=gridIn(i,1)-yOffset;
	  locations(numPairs,2)=j;
	}
	numPairs++;

      } // if (OK)
      
    } // for (int j=0; j<4;...);
  } // for (int i=0; i<gridNumPts...)
  
  cout << "numPairs=" << numPairs << endl;

  lands.init(gridNumPts*4*numCuts + gridNumPts*4*numCuts*numPairs);

  for (int i=0; i<numPairs; i++) {
    for (int k=0; k<numCuts; k++) {

      if (distances(i) / refDist > cuts(k)) {

	lands(count).cx=locations(i,0);
	lands(count).cy=locations(i,1);
	lands(count).index=k*4 + locations(i,2);
	count++;
      } // If feature passes test
    } // for (int k=0;.k<numCuts;...)
  } // for (int i=0; i<numPairs;...)

  cout << "Doing ratios." << endl;
  int maxindex=0;

  double distRatio=0;
  for (int i=0; i<numPairs; i++) {
    for (int j=i+1; j<numPairs; j++) {
      if (distances(i)>=0 && distances(j)>=0) {

	if (distances(i)==distances(j)) {
	  distRatio=1;
	}
	else if (distances(j)==0) {
	  distRatio=100000;
	}
	else {
	  distRatio = distances(i)/distances(j);
	}
	for (int k=0; k<numCuts; k++) {
	  
	  if (distRatio > cuts(k)) {
	    
	    lands(count).cx=locations(i,0);
	    lands(count).cy=locations(i,1);
	    lands(count).index=numCuts*4 + ((k*4+locations(i,2))*numPairs + j);
	    if (lands(count).index>maxindex) {
	      maxindex=lands(count).index;
	    }
	    // 	  cout << "Feature " << count << " / " << lands.dim() << endl;
	    count++;
	  }
	} // for (int k=0; k<numCuts; k++)
      } // if (distances(i)>=0 && distances(j)>=0)
      
    } // for (int j=i+1; j<numPairs; ...)
  } // for (int i=0; i<numPairs; ...)
  
  cout << "maxindex=" << maxindex << endl;

  return count;
  
}
