// Copyright YALI AMIT, January 15, 2001.

#ifndef FINDDETS_C
#define FINDDETS_C



// __________________________________________________________________________

#include <iostream>
#include <cstdio>

#include "includes.h"
#include "grid.h"
#include "simpstrg.h"
#include "timer.h"
#include "avlgrid.h"
#include "landmark.h"
#include "pixel.h"
#include "detection.h"
#include "carttree.h"
#include "ltatree.h"
#include "viewgrid.h"
#include "viewport.h"
#include "globals.h"
#include "detfuncs.h"

extern double ratios[];
extern int statpts[];
extern images IMV;
extern int CURRDET, CURRIM;

void viewdet(const image& im, partialtrans& pt, detection& det);
void show_hough_points(int oldg, int numf, int currratio, detections &tmpdetects, viewport &vp3, int &oldmax, int &imax, int &jmax);
void make_procmask(ucgrid& processmask, int masktype, int sx, int sy);
int cluster_triangles(detections& dets, int numtr);
int  process_template(image &imorig, detection &det, partialtrans fpt, ucgrid &fpmask);
 void show_instantiation(igrid &maxx, double currratio);
void detectedges(const image& im, landmarks& lmarks);
void eval_detect(detection &det, ltatrees &ltrees, int numlt);
int getscl(ltatrees& ltrees, int coun, detections &tmpdets, 
              int numlt, int acchits, int x1, int y1, 
	   int lcentx, int lcenty);
void show_landmarks(image &im, landmarks &edgelands, landmarks & ltalands);
int q1d1(landmarks& lmarks, igrids& instances);
// Global variable to store image 
image globim;
partialtrans PT;
extern ltatrees ltrees;
extern carttrees ctrees, ctreesw, ftrees;
typedef vector<carttrees> carttreess;
carttreess ftreess, ftreesss;
detections tmpdetects;
double cs[32], sn[32];
igrids hougharea;
int centx,centy, ecentx, ecenty;
ltatrees  etrees;
int maxnumltas;
double savefac=1.;
int ltacentx, ltacenty;


// __________________________________________________________________________

void finddets(image& imorig, const simpstrg& pgm, detections& dets,
	       const simpstrg& parf)
{

  if (coarse)
    savefac=ltacluster;

  // Read in a pgmfile if 'imorig' not already set up.
  if (imorig.dimx() == 0 && !imorig.read(pgm))
    exit(1);

  static partialtrans pt, fpt;  

    
  if (dumpdetections)
    cout << "Dimensions " << imorig.dimx() << " " << imorig.dimy() << endl;

  // Perform one time initializations.
  static int inited = 0;
  static ucgrid pmask, fpmask;

  // Read in model ltas and edges and get classification trees. 
  // This is done only for first image processed while program is running
  if (!inited)
  {
    inited = 1;

    // Read in the parameters
    if (parf != "" && !parse_parfile(parf)) {
      cerr << "Unable to open parameter file [" << parfile << "].\n";
      exit(1);
    }

    // Center point of model triangle not necessarily equal to
    // centeroid  of model  points.

    ltacentx=(ltatree1x+ltatree2x+ltatree3x)/3;
    ltacenty=(ltatree1y+ltatree2y+ltatree3y)/3;
    cout << " ltacenter " << ltacentx << " " << ltacenty << endl;
    // Dump the values of global variables to the dumpfile (if
    // specified).
    dump_globals(dumpfile);
    // Vector of detections
    tmpdetects.init(maxdetects);

    // Get model lta information and model edge information
    // Prepare areas for hough transform for each model point.
    if (hough)
      fill_hough_areas();

    // Get classification trees
    get_trees(pt, fpt);
   
    if (!onlyedges)
      {
	numang = ltrees[0].tnumang;
	if (near<0)
	  near = -near;
	else
	  near = ltrees[0].tnear;
      }

    // Masks to impose on reference grid, for classification stages,
    // for normalization purposes.
    make_procmask(pmask, masktype, cartsizex, cartsizey);
    make_procmask(fpmask, fmasktype, fcartsizex, fcartsizey);
  }
  else 
    {
      for (int i = 0; i < maxdetects; ++i)
	tmpdetects[i].reinit();
    }

 



 
  // Non flexible graphics with repeated access to image
  globim = imorig;


  timer timer1;
 if (basecol==0)
  globim.normalize(numcolors-1);
 else
  globim.normalize(imrange*numcolors,(1-imrange)*numcolors-1);

  viewport vp3;
  // Viewport for showing hough transform output
  if (hough>=3)
    vp3.add_grid(globim);

  int oldmax=-1;
  int imax=-1, jmax=-1;
  int numfound=0, currgood=0, oldgood=0;
  imorig.normalize(255);

  // Loop through different resolutions of the image to get objects at scales 4:1
  for (int currratio = firstratio; 
       currratio < firstratio + numratios; ++currratio)
    {
      if (dumpdetections)
	cout << endl << "RATIO " << ratios[currratio] << endl;

      image im = imorig;
      if (ratios[currratio]<1.)
	im.blur_downscale(ratios[currratio], sd);
      
    
   
      timer1.start();
      
      // Find the triangles
      // landmark classes to store locations of edges and lta's
      landmarks edgelands, ltalands;
      numfound=oldgood;

      // Get edges, lta's and candidate centers from hough transform
      find_triangles(im, ltrees, etrees, numltas, edgelands, ltalands, tmpdetects, numfound);
      if (showedges >= 1 || showltas >= 1 || hough==5)
	return;
      
      
      // Show result of hough transform.
      if (hough>=3)
	{
	  show_hough_points(oldgood, numfound, currratio, tmpdetects, vp3, oldmax, imax, jmax);
	  continue;
	}
      
      // Prepare data for clasification. 
      // cartstyle=0 - normalized gray level classification
      if (cartstyle == 2) // Classification trees based on edge locations
	fill_STAGS(edgelands, cartspread);
      else if (cartstyle==3) // Classification trees based on lta locations
        fill_STAGS(ltalands, cartspread);
      
      timer timer2;
      if (dumpdetections)
	timer2.start("Filtering Triangles");
      // Loop over detected linear poses stored in tmpdetects
      for (int j = oldgood; j < numfound; ++j)
	{
	  int ii;
	  ucgrids reglands,  reglands2;
	  dgrid regim, regim2;
	  int f1=1;
	  // If there are classification trees of the 
	  // class against the rest of the world
	  if (numcarts > 0)
	    {
	      // Register data (gray scale, edge or lta) 
	      // in ROI determined by detected linear pose.
	      register_image(im, pt, regim, reglands, 
			     tmpdetects[j].tripts, 0,
			     cartstyle, cartsizex, cartsizey, pmask, leastsquares);

	      // Run cart trees on registered data
	      f1 = cart_against_world(regim, ctrees, tmpdetects[j].cartprobs, reglands, numcarts / itersize);
	    }

	  // If detection passed cart test
	  if ( f1 )
	    {
	      
	      // Classify the detected object among several classes 
	      // fcartstyle determines type of data as cartstyle ( 0, 2,3).
	      if (fnumcarts > 0)
		{
		  // Possible to use a different reference grid with different
		  // anchor points, coded in fpt, for classification
		  register_image(im, fpt, regim2, reglands2, 
				 tmpdetects[j].tripts, 0,fcartstyle, fcartsizex, 
				 fcartsizey, fpmask, fleastsquares);

		  // Run classification trees on registered data
		  // This does not prune detections only labels them
		  cart_classify(regim2, ftrees, reglands2, tmpdetects[j].cartprobs2);

		  tmpdetects[j].findmax();		
		  // Possible third round. Double check classification
		  // Classify recognized class against rest of the world.
		  if (fcart2 && tmpdetects[j].counts[5]>=0) 
		    cart_classify(regim2, ftreess[tmpdetects[j].counts[5]], 
			   reglands2, tmpdetects[j].cartprobs3);
		  
		}
	      
	      // If classified detection survives third round or there is no third round
	      // Put some more information in the detection class
	      if (!fcart2 || 
		  (tmpdetects[j].cartprobs3[0] >  alpha1*tmpdetects[j].cartprobs3[1]))
		{
		  // A face was detected, so update the detection.
		  tmpdetects[j].good = 1;
		  tmpdetects[j].scale = ratios[currratio];
                  // Record numltas,hits and numedges,acchits for this detection
		  tmpdetects[j].counts[1]=numltas;
		  tmpdetects[j].counts[2]=hits;
		  tmpdetects[j].counts[3]=numedgetrees;
		  tmpdetects[j].counts[4]=acchits;
                  if (classlabel>=0 )
		    tmpdetects[j].counts[5]=classlabel;
                  else if (makecoords==1)
		    tmpdetects[j].counts[5]=edgelands.truth();

		  for (ii = 0; ii < 8; ++ii)
		    tmpdetects[j].tripts[ii] = 
		      int(rint(tmpdetects[j].tripts[ii]/ratios[currratio]));
		  if (showbest==5)
		    show_instantiation(tmpdetects[j].locs, ratios[currratio]);
		}
	      
	    }
	}
   
      if (cartstyle == 2)
	empty_STAGS(edgelands);
      else if (cartstyle==3)
        empty_STAGS(ltalands);      
      
      // Get rid of detections which may be faulty (good=0)  
      compress_dets(tmpdetects, oldgood, numfound, currgood);

      if (dumpdetections)
	cout << "After compress Left with [" << currgood-oldgood << "] " << endl;
	oldgood=currgood;

      if (dumpdetections)
	timer2.stop("Filtering Triangles");


     
    }
  
  timer1.stop("Find/Filter Triangles");
  
  // Cluster the detections. Among triangles within a certain distance
  //  of each other, pick the one with highest value according to some criterion
  int finalboxes = cluster_triangles(tmpdetects, currgood);
  finalboxes = cluster_triangles(tmpdetects, finalboxes);
  if (dumpdetections)
    cout << "Left with [" << finalboxes << "] boxes \n";
  
  
  // Return an array of detections
  int i, numdets = 0;

  for (i = 0; i < currgood; ++i)
    {
      // Show detected features - instantiation - of detections.
      if (showbest==6 && tmpdetects[i].good)
	{
	  cout << facenames[tmpdetects[i].counts[5]] << endl;
	  cout << tmpdetects[i].counts[0] << " " << 
	    tmpdetects[i].counts[3] << " " << tmpdetects[i].counts[4] << endl;

	 show_instantiation(tmpdetects[i].locs, tmpdetects[i].scale);
	}
      // If required process detection with some form of deformable template model
	if(tmpdetects[i].good && templatestyle>0)
	  {
	    if (!process_template(imorig, tmpdetects[i], fpt, fpmask))
	      tmpdetects[i].good=0;
	  }

      numdets += (tmpdetects[i].good != 0);
    }
  
  if (dumpdetections)
    cout << "numdets = " << numdets << endl;
 

 dets.init(numdets);
 int currdet=0;
 for (i = 0; i < numdets; ++i)
    {
      while (!tmpdetects[currdet].good)
	++currdet;
      dets[i] = tmpdetects[currdet];
      ++currdet;
    }
  
 
 // Show point with maximal hough count.
  if (hough>=3  && imax>=0 && jmax >= 0 )
    {
      if (hough>3)
	{
	  if (basecol==0)
	    vp3.add_point(imax,jmax,3,5);      
	  else
	    vp3.add_string(imax-2,jmax+7,"X",7,0);
	}
      vp3.view();
    } 
  

} // 


// Show points detected in hough transform
void show_hough_points(int oldg, int numf, int currratio, detections &tmpdetects, viewport &vp3, int &oldmax, int &imax, int &jmax)
    {
      for (int u=oldg; u<numf; u++)
      {
        if (tmpdetects[u].counts[0]>oldmax)
        { 
          oldmax= tmpdetects[u].counts[0];
          imax=int(tmpdetects[u].tripts[6]/ratios[currratio]);
          jmax=int(tmpdetects[u].tripts[7]/ratios[currratio]);
        }
            vp3.add_point(tmpdetects[u].tripts[6]/ratios[currratio],
			  tmpdetects[u].tripts[7]/ratios[currratio],basecol,3);

      }
      vp3.refresh();
    }      


// __________________________________________________________________________

// Show an instantiation
// SHow all the detected local featurees in the instantiation.
void show_instantiation(igrid &maxx, double currratio)
{
  image im;

  // Crop image if required
 if (LBX>0 && LBY >0 && regionsize>0)
   {
     im.init(regionsize,regionsize);
     im.subcopy(LBX, LBX+regionsize, LBY, LBY+regionsize, globim, 0, 0);
   }
 else
   {
     LBX=0; LBY=0;
     im=globim; 
     im.blur_downscale(1.,sd);
   }
im.normalize(imrange*numcolors, (1-imrange)*numcolors-1);
 viewport vp(im);
 vp.refresh();
 for (int t3 = 0; t3 < maxx.dimx(); ++t3)
   {
     cout << maxx(t3,0) << " " << maxx(t3,1) << " " << maxx(t3,2) << endl;
     int x1,y1;
     if (currratio<1)
       {
	 x1=int(maxx(t3, 0) / currratio);
	 y1=int(maxx(t3, 1) / currratio);
       }
     else
       {
	 x1=int(maxx(t3, 0));
	 y1=int(maxx(t3, 1));
       }
     x1-=LBX; y1-=LBY;
     int col=basecol;
     if (maxx(t3,2)==-1)
       {
	 //	 col=magcol;
	   vp.add_point(x1,y1,10,3);
	   //vp.add_string(x1-1,y1+1,"x",basecol,0);
       }
     else
       {
	 if (dum1==1)
	   vp.add_string(x1-2,y1+2,ltos(t3),basecol,0);
	 //	 vp.add_string(x1-1,y1+1,"+",basecol,0);
	 vp.add_point(x1,y1,col,3);
       }
     
     
   }

 vp.view();
}
// show

// Run cart trees of class against the rest of the world. There
// could be several iterations of such groups of (itersize) trees (numiters).
// Each iteration is a filter on detections

int cart_against_world(dgrid& roi, const carttree* ctrees,
          dgrid &cartprobs, ucgrids& reglands, int numiters)
{
  int i, j;

  // `1' is the correct clas, `0' is the background class. 
  for (i = 0; i < numiters; ++i)
  {
    dgrid agg(1, 2);
    agg.clear();
    for (j = 0; j < itersize; ++j)
    {
      ctrees[i * itersize + j].drop(roi, reglands, agg);
    }
    float numbgr = ctrees[i * itersize].top()->hist[0];
    double numface = ctrees[i * itersize].top()->hist[1];
    double dataratio;
    // Reweighting factor to avoid false negatives.
    if (purepar>0)
      dataratio= MAX(1., numbgr / numface) * purepar;
    else
      dataratio=1.;
    cartprobs[i] += dataratio*agg(0, 1) / (agg(0, 0) + agg(0, 1));
    if (dum1 == 92) 
      {
	cout << "dataratio " << dataratio << endl;
	cout << "iteration " << i << " [" 
	     << agg(0,0) << ", " << agg(0,1) << "]\n";
	cout << "CRT " << cartprobs[i] << endl;
      }
    if (dataratio * agg(0,1) < agg(0,0))
	return 0;

  }
  
  return 1;

} // cart_against_world

// __________________________________________________________________________
// cart_classify classifies object among several classes, does not filter.
int cart_classify(dgrid& roi, const carttree* ctrees, ucgrids& reglands, dgrid &cp)
{
  dgrid agg(1, numclass);
  agg.clear();
  for (int j = 0; j < fnumcarts; ++j)
  {
    ctrees[j].drop(roi, reglands, agg);
  }
  for (int i = 0; i < cp.dimx(); ++i)
    cp[i] = agg(0, i);

  return 1;
}


// __________________________________________________________________________
// View a detection, the registered ROI, there is also an option (viewdetections=2)
// to store each ROI in a global vector of images, and finally
// write out the collection of detections in the Images_* format
void viewdet(const image& im, partialtrans& pt, detection& det)
{
  static int inited = 0;
  static ucgrid fpmask;
  if (!inited) {
    inited = 1;
    make_procmask(fpmask, fmasktype, fcartsizex, fcartsizey);
  }
  ucgrids reglands;
  dgrid regim;
  image iml=im;
  if (det.scale < 1.)
    iml.blur_downscale(1, sd);
 
  register_image(iml, pt, regim, reglands, det.tripts, 0,
		 0, fcartsizex, fcartsizey, fpmask,
		 fleastsquares);
    regim.normalize(0,numcolors-1);

    image vim;
    if (viewdetections==2)
      {        
	if (CURRDET < numdata){
	IMV[CURRDET].init(regim.dimx(),regim.dimy());
	gridfun((uchar)(regim(i,j)),IMV[CURRDET]);
	//IMV[CURRDET].set_truth(im.truth());
	IMV[CURRDET].set_truth(0);
	// viewport vp(IMV[CURRDET]);
	//vp.view(); 
	CURRDET++;
	}
      }
    else
      {
     vim.init(regim.dimx(), regim.dimy());
      gridfun((uchar)(regim(i,j)),vim);
      viewport vp(vim);
       vp.view();   
      }

} // viewdet

// __________________________________________________________________________
// Output information regarding the detections
void output_detections(viewport& vp, image& im, detections& dets)
{
  int i;

  cout << "Left with " << dets.dim() << " Boxes " << endl;

  // Output the triangle of the detection.
  if (makecoords) {
    ofstream cfile("detcoords", ios::app);
    if (dets.dim()>0)
      {
	for (int j=0; j < dets.dim(); j++)
	  {
	    for (i = 0; i < 6; ++i) 
	      cfile << dets[j].tripts[i] << " ";
            cfile << CURRIM  << " " << dets[j].counts[5] << endl;
	  }
      }
    else
      cfile << "-1 -1 -1 -1 -1 -1 " << CURRIM <<  " -1" <<endl;
    return;
  }

  // Dump as much information as possible
  if (dumpdetections==2) {
    for (i = 0; i < dets.dim(); ++i)
    {
      dets[i].dump();
     if (dumpdetections==3)
      dets[i].locs.dump();

     // View the detection
      if (viewdetections)
      {
        cout << fpoint1x << " " << fpoint1y << endl;
        cout << fpoint2x << " " << fpoint2y << endl;
        cout << fpoint3x << " " << fpoint3y << endl;
	partialtrans pt(Pixel(fpoint1x, fpoint1y), Pixel(fpoint2x, fpoint2y),
			Pixel(fpoint3x, fpoint3y));
	viewdet(im, pt, dets[i]);
      }
    }
  }

 
  // Show the triangles on the image, and if there is, the class label of the detection
  if (showtriangles)
  {
    image tmpim = im;
    
    tmpim.normalize(imrange*numcolors,(1-imrange)*numcolors - 1);
    vp.add_grid(tmpim);

    vp.clear_geom();
    if (showtriangles==1 || showtriangles==3)
    make_triangles(vp, dets);
      for (i = 0; i < dets.dim(); ++i) 
	{
	  if (dets[i].counts[5]>=0)
	    {
	      int x1=dets[i].tripts[0];
	      int y1=dets[i].tripts[1]-5;
	     
	      if (basecol==0)
		{
		  vp.add_string(x1,y1,facenames[dets[i].counts[5]]);
		}
	      else
		{
		  vp.add_string(x1,y1,facenames[dets[i].counts[5]],basecol);
		}
	    }
        }
    vp.view();
  }

} // output_detections


// Add the detection triangle to the viewport
void make_triangles(viewport& vp, detections& dets)
{
  //   vp.clear_geom();
  for (int i = 0; i < dets.dim(); ++i)
  {
    if (showtriangles==1)
      {
	vp.add_line(MAX(0, dets[i].tripts[0]), MAX(0, dets[i].tripts[1]),
                 MAX(0, dets[i].tripts[2]), MAX(0, dets[i].tripts[3]),basecol);
	vp.add_line(MAX(0, dets[i].tripts[0]), MAX(0, dets[i].tripts[1]),
		    MAX(0, dets[i].tripts[4]), MAX(0, dets[i].tripts[5]),basecol);
	vp.add_line(MAX(0, dets[i].tripts[2]), MAX(0, dets[i].tripts[3]),
                 MAX(0, dets[i].tripts[4]), MAX(0, dets[i].tripts[5]),basecol);
      }
    if (showtriangles==3)
      {
	vp.add_point(MAX(0, dets[i].tripts[0]), MAX(0, dets[i].tripts[1]),basecol,3);
	vp.add_point(MAX(0, dets[i].tripts[4]), MAX(0, dets[i].tripts[5]),basecol,3);
	vp.add_point(MAX(0, dets[i].tripts[2]), MAX(0, dets[i].tripts[3]),basecol,3);
      }
  }

} // make_triangle


// Get rid of detections that havent survived some past filter
// they have good=0
void compress_dets(detections &tmpdets, int oldgood, int numfound, int &currgood)
{
  currgood=oldgood;
  for (int j = oldgood; j < numfound; j++)
    {
      if (tmpdets[j].good==1) 
	{
          if (j>currgood)
	    {
	      tmpdets(currgood) = tmpdets(j);
              tmpdets(j).good=0;
	    }
	  currgood++;
	}
    }

 
}

// Cluster detections according to some criterion
int cluster_triangles(detections& dets, int numtr)
{
  int i, k;
  int numtriangles=numtr;
  // boxcluster - maximial distance between surviving detection and next
  //              detection in cluster
  if (boxcluster) {
    for (i = 0; i < numtriangles - 1; ++i)
      {
	
	// If this is a large object, clustering radius should be larger accordingly
	int newboxcluster = int(boxcluster / dets[i].scale);
	
        double oldscore = 0;
        switch (boxmethod)
	  {

	    // Number of detected features in instantiation
	  case 0:
	    oldscore = dets[i].counts[0];
	    break;
	    // Score on classification filter of object against rest of the world
          case 1:
	    oldscore=dets[i].cartprobs[0];
	    break;
	    // Score according to weight on class label
	  case 2:
	    oldscore=dets[i].cartprobs2[dets[i].counts[5]];
	    break;
	    // Score according to a value in cartprobs3, used for various scores.
	  case 3:
	    oldscore = dets[i].cartprobs3[0];
	    break;
	  }
        for (k = i + 1; k < numtriangles; )
	  {
	    // Check if new triangle close to current one
	    if ((abs(dets[i].tripts[0] - dets[k].tripts[0]) < newboxcluster) &&
		(abs(dets[i].tripts[1] - dets[k].tripts[1]) < newboxcluster) &&
		(abs(dets[i].tripts[2] - dets[k].tripts[2]) < newboxcluster) &&
		(abs(dets[i].tripts[3] - dets[k].tripts[3]) < newboxcluster) &&
		(abs(dets[i].tripts[4] - dets[k].tripts[4]) < newboxcluster) &&
		(abs(dets[i].tripts[5] - dets[k].tripts[5]) < newboxcluster))
	      {

		// One less triangle - it will either replace current 
		// or be replaced by last on
		--numtriangles;
		double newscore = 0;
		// Get score of new triangle
		switch (boxmethod)
		  {
		  case 0:
		    newscore = dets[k].counts[0];
		    break;
		  case 1:
		    newscore=dets[k].cartprobs[0];
		    break;
		  case 2:
		    newscore=dets[k].cartprobs2[dets[k].counts[5]];
		    break;
		  case 3:
		    newscore = dets[k].cartprobs3[0];
		    break;
		  }
		// Have last triangle replace k'th which was not better than current
		// since curr
		if (oldscore >= newscore)
		  {
		    if (k != numtriangles)
		      dets[k] = dets[numtriangles];
		    dets[numtriangles].good = 0;
		  }
		// Have k'th triangle replace current - it is better
		else
		  {
		    dets[i] = dets[k];
		    dets[k] = dets[numtriangles];
		    oldscore=newscore;
		    dets[numtriangles].good = 0;
		  }
	      }
	    else
	      {
	      ++k;
	      }
	  }
      }
    
  }
  return numtriangles;
  
} // cluster_triangles


// __________________________________________________________________________
// Get edges, ltas.
// Find candidate locations using Hough transform, and get scale and rotation
// estimates
void find_triangles(image& im, ltatrees& ltrees, ltatrees& etrees, int numltas,
                    landmarks& edgelands, landmarks& ltalands,
		    detections& tmpdets, int& numfound)
{

 
  // Geet edges and ltas  
  extract_landmarks(im, ltrees, numltas, edgelands, ltalands);
  if (showedges || showltas )
    return;
  // Steps 1 and 2 of counting detector.
  hough_and_scale(im, edgelands, ltalands, tmpdets, numfound, ltrees, etrees);
  return;
  
  
  
} // find_triangles

  igrids PREINSTS;

void extract_landmarks(image& im,  ltatrees& ltrees, int numltas,
		       landmarks& edgelands, landmarks& ltalands)
{

  void edges_to_ltas_new(landmarks& edgelands, landmarks& ltalands, ltatree *ltrees, int numltas);
  if (ltrees.dim()==0 && onlyedges==0 && ltausedepth!=0)
    {
      numltas=read_ltatrees(ltrees, ltatreedir);
      maxnumltas=numltas;
    }
  timer timer1;
  if (dumpdetections)
    timer1.start("Finding Edges");

  detectedges(im, edgelands);
  if (onlyedges)
    {
      if (showedges>=1)
	show_landmarks(im, edgelands, edgelands);
      return;
    }
 if (dumpdetections)
    {
      cout << " Found [" << edgelands.total_instances() << "] edges... ";
      cout << " Density " << (double)(edgelands.total_instances())/
	((8/edgeinc)*im.dimx()*im.dimy()) << endl;
      timer1.stop("Finding Edges");
    }
  
  if (edgesforscale==0 || check_type==0 )
    fill_STAGS(edgelands,0,maxnumltas);
  else
    fill_STAGS(edgelands);
  if (onlyedges==0)
    q1d1(edgelands, PREINSTS);

 
  if (onlyedges==0 && !showedges)
    {
      if (dumpdetections)
	timer1.restart("Finding LTAs");
      if (ltausedepth==0)
	{
	  void detectpairs(landmarks& edges, landmarks& lmarks);
	  detectpairs(edgelands, ltalands);
	}
      else
	//	edges_to_ltas(edgelands, ltalands, ltrees, numltas);
         edges_to_ltas_new(edgelands, ltalands, ltrees, numltas);
      if (dumpdetections)
	{
	  cout << " Found [" << ltalands.total_instances() << "] ltas... ";
	  cout << " Density " << (double)(ltalands.total_instances())/
	    (numltas*im.dimx()*im.dimy()) << endl;
	  timer1.stop("Finding LTAs");
	}
    }

  if (showedges>=1 || showltas>=1)
  show_landmarks(im, edgelands, ltalands);
  empty_STAGS(edgelands);
} 
// Get ltas from edge info
void edges_to_ltas_new(landmarks& edgelands, landmarks& ltalands, ltatree *ltrees, int numltas)
{
 int ltas(ltaiter iter, const landmarks& lmarks, igrid& instances);


  ltalands.set_truth(edgelands.truth());
  for (int i = 0; i < 6; ++i)
    ltalands.coords[i] = edgelands.coords[i];
  ltalands.init_counts_offsets(numltas);


  int numlands = 0;
 
  static lmarklist land(maxlands);

 
if (dumpdetections==1)
  cout << " In new " << numltas << " " << endl; 

  // Computer ltas
  for (int i = 0; i < numltas; ++i)
  {
    ltaiter terminal;
    //    cout << " i " << i << endl;
    if (ltausedepth==10)
      terminal=ltrees[i].top();
    igrid insts;

  
   int logic;
   //int oldd=ltausedepth;
   //ltausedepth=0;

   ltrees[i].dropdata_fromlist(edgelands, terminal, insts, PREINSTS);
   logic= (((ltausedepth == -1 && terminal.leaf() && terminal.yn()) || 
	    terminal.depth()-1 == ltausedepth) && terminal.yn());
   //ltausedepth=oldd;

   // If at least one was found
   if (logic)
    {
      int numinsts = insts.dimy();

      if (numinsts + numlands >= maxlands)
	{
	  cerr << "Too many landmarks in edges_to_ltas routine.\n";
	  exit(1);
	}
      // Store in landmark list.
      for (int j = 0; j < numinsts; j++)
      {
        land[j + numlands].cx = insts(1, j);
        land[j + numlands].cy = insts(2, j);
	// Inherits the difference of the center edge.
        land[j + numlands].index = i;
      }
      numlands += numinsts;
    }
  }
  
 if (coarse)
    clust_landmarks(land,ltacluster,numltas,numlands);
  else
    cluster_landmarks(land, ltacluster, numltas, numlands);

  // Store remaining ltas in a landmark structure
  ltalands.setup_from_lmarklist(land, numlands);

 
  
} // edges_to_ltas



// Get ltas from edge info
void edges_to_ltas(landmarks& edgelands, landmarks& ltalands, 
		   const ltatree *ltrees, int numltas)
{
 int ltas(ltaiter iter, const landmarks& lmarks, igrid& instances);


  ltalands.set_truth(edgelands.truth());
  for (int i = 0; i < 6; ++i)
    ltalands.coords[i] = edgelands.coords[i];
  ltalands.init_counts_offsets(numltas);

  int numlands = 0;
 
  static lmarklist land(maxlands);

  // Prepare global arrays for recording edge or lta locations for fast?? processing
  // and input edge locations for computing ltas
  if (edgesforscale==0 || check_type==0 )
  fill_STAGS(edgelands,0,maxnumltas);
  else
  fill_STAGS(edgelands);

  // Computer ltas
  for (int i = 0; i < numltas; ++i)
  {
    ltaiter terminal;

    if (ltausedepth==10)
      terminal=ltrees[i].top();
    igrid insts;

  
   int logic;
   // Compute ltas depth first at each first edge location.
   if (ltausedepth==10)
       logic=ltas(terminal, edgelands, insts);
   // Compute each depth at all location.
   else
      {
	ltrees[i].dropdata(edgelands, terminal, insts);
	logic= (((ltausedepth == -1 && terminal.leaf() && terminal.yn()) || 
		 terminal.depth() == ltausedepth) && terminal.yn());
      }
   // If at least one was found
   if (logic)
    {
      int numinsts = insts.dimy();
      if (numinsts + numlands >= maxlands)
	{
	  cerr << "Too many landmarks in edges_to_ltas routine.\n";
	  exit(1);
	}
      // Store in landmark list.
      for (int j = 0; j < numinsts; j++)
      {
        land[j + numlands].cx = insts(0, j);
        land[j + numlands].cy = insts(1, j);
        land[j + numlands].index = i;
      }
      numlands += numinsts;
    }
  }
  empty_STAGS(edgelands);
  // Cluster ltas, either into coarser resolution, or taking
  // average location of nearby ltas of same type
  if (coarse)
    clust_landmarks(land,ltacluster,numltas,numlands);
  else
    cluster_landmarks(land, ltacluster, numltas, numlands);

  // Store remaining ltas in a landmark structure
  ltalands.setup_from_lmarklist(land, numlands);

} // edges_to_ltas


// __________________________________________________________________________
// Currently not used
void register_detection(detection &det,  ucgrids& reglands, int csizex, int csizey, int cst, int unions=0)
{
 affinetrans at(Pixel(det.tripts[0], det.tripts[1]),
                Pixel(det.tripts[2], det.tripts[3]),
		Pixel(det.tripts[4], det.tripts[5]),
		Pixel(point1x, point1y), 
		Pixel(point2x,point2y),
		Pixel(point3x, point3y));

 if (edgesforscale==1)
   {
     if (unions)
       reglands.init(numedges / 2);
     else
       reglands.init(numedges);
   }
 else 
   reglands.init(numltas);

 for (int ii = 0; ii < reglands.dim(); ++ii)
   reglands[ii].init(csizex, csizey);
 

 for (int k=0; k<det.locs.dimx(); k++)
   {
     Pixel P(det.locs(k,0),det.locs(k,1));
     Pixel Q=at*P;
     if (det.locs(k,2)>=0)
       {
	 for (int xx=Q[0]-cartspread; xx<=Q[0]+cartspread; xx++)
	   for (int yy=Q[1]-cartspread; yy<=Q[1]+cartspread; yy++)
	     {
	       reglands[det.locs(k,2)].set(xx,yy,1);
	     }
       }
   }
 if (dum1==18)
   {

      void view_normalized(const ucgrid& gr, int gfork=0, int min=0, int max=0);
      for (int i=0; i<reglands.dim(); i++)
	{
	  cout << " landmark " << i << endl;
	  view_normalized(reglands[i]);
	}
    }

}    


// Register ROI data into reference grid csizex by csizey with
// anchor points coded in `pt'
void register_image(const image& im,  partialtrans& pt, 
		    dgrid& regim, ucgrids& reglands, int *tripts, int j,
		    int cstyle, int csizex, int csizey, ucgrid& pmask,
                    int lsq, int unions/* = 0*/)
{
  affinetrans at(pt, Pixel(tripts[0], tripts[1]),
                     Pixel(tripts[2], tripts[3]),
		     Pixel(tripts[4], tripts[5]));
 
  switch (cstyle) 
  {
   case 0:
   case 1:
    regim.init(csizex, csizey);
    register_and_normalize(im,  at, regim, pmask, lsq);
    break;

   case 2:
    if (unions)
      reglands.init(numedges / 2);
    else
      reglands.init(numedges);
    for (int ii = 0; ii < reglands.dim(); ++ii)
      reglands[ii].init(csizex, csizey);
    register_landmarks(at, reglands, unions);
    break;

   case 3:
    reglands.init(numltas);
    for (int ii = 0; ii < numltas; ++ii)
      reglands[ii].init(csizex, csizey);
    register_landmarks(at, reglands);
    break;
  }

} // register_image






// Perform Hough transform and obtain best linear pose at each candidate location
void hough_and_scale(image &im, const landmarks& ed, const landmarks& lm, 
	   detections &tmpdets, int &currgood, ltatrees& ltrees, ltatrees &etrees)
{
  void  fill_stemp(ltatrees &vtrees, int numt, const landmarks &lm, 
		   igrid &stemp, int dimx, int dimy, int pad);




  timer h1, h2;
 if (dumpdetections)
    {
  h1.start();
  h2.start();
  }
  cout << " In Hough " << numltas << endl;

  // Tree for storing instances of candidate centers and sorting them
  static iavlgrid tmpinsts;
  static int inited= 0;
  if (!inited) 
    {
      inited = 1;
      tmpinsts.ainit(maxdetects, 2);
    }
  else
    tmpinsts.reinit();

    igrid stemp;
    igrid temp;
    // Grid for hough votes is padded to avoid checking boundaries.
    stemp.init(im.dimx()/savefac + 2 * pad, im.dimy()/savefac + 2 * pad);
    stemp.clear();

    // Perform hough transform counts either with edge model or lta model
    if (onlyedges)
      fill_stemp(etrees, numedgetrees, ed, stemp, im.dimx(), im.dimy(), pad);
    else
      fill_stemp(ltrees, numltas, lm, stemp, im.dimx(), im.dimy(), pad);
    
    if (dumpdetections)
      h1.stop("hough1");
   
  
  // Fill STAGS for scale estimate
  if (hough==2)
    {
      if (edgesforscale)
	fill_STAGS(ed,edgespread);
      else
	fill_STAGS(lm,edgespread);
    }
  int coun=0;
  int bc=0;
  // Find locations with more than accfactor*hits votes.
  for (int i=pad; i< stemp.dimx() - pad; i++)
    for (int j=pad; j< stemp.dimy() - pad; j++)
    {
      int hh = (stemp(i, j) > hits*accfactor);
      if ( hh && coun < maxdetects)
      {
        static ivector l(2);
        l[0] = (i-pad);
        l[1] = (j-pad);
	// Store detections on a coarse grid
        if (coarse)
	  {
	    l[0]=int(rint(l[0]*savefac));
	    l[1]=int(rint(l[1]*savefac));
	  }
        if (bc==0 || (coarse || !(tmpinsts.tooclose(l, tridist))))
	  {
	    if (!coarse)
	      tmpinsts.insert(l);
	    bc++;
	    tmpdets[currgood+coun].counts[0]=stemp(i,j);	    
	    if (hough>=3) 
	      {
		tmpdets[currgood+coun].tripts[6] = l[0];
		tmpdets[currgood+coun].tripts[7] = l[1];

		coun++;
	      }
	    // Estimage linear pose either using edge model or lta model

	    if (hough == 2)
	      {
		int logic=0;

		if (edgesforscale)
		  {
		    if (showbest>0)
		      logic= getscale(etrees, currgood+coun, tmpdets, 
				      numedgetrees, acchits, l[0],l[1], centx, centy);
		    else
		      logic= getscl(etrees, currgood+coun, tmpdets, 
				  numedgetrees, acchits, l[0],l[1], centx, centy);
		  }
  
		else
		  logic= getscale(ltrees, currgood+coun, tmpdets, 
				  numltas, acchits, l[0],l[1], centx, centy);
		coun+=logic;
		
	      }
	    
	  }
      }
    }
      cout << " without edge filter " << bc << endl;
      cout << " coun " << coun << endl;
      cout << "DETS " << coun << endl;
      if (dumpdetections)
	{     
	  if (hough>=3)
	    {
	      cout << " numdets " << coun << " -1 " << "-1 " << endl;
	      if (dumpdetections==2)
		{
		  for (int k=0; k<coun; k++)
		    {
		      cout << " detec " << numbg << " " << tmpdets[currgood+k].tripts[6] << " " << tmpdets[currgood+k].tripts[7] << " " ;
		      cout << "account " <<tmpdets[currgood+k].counts[0] << endl;
		    }
		}
	    }
	}
      currgood += coun;
      
      if (hough==2)
	{
	  if (edgesforscale)
	    empty_STAGS(ed);
	  else
	    empty_STAGS(lm);
	}
      if (dumpdetections)
	h2.stop("hough2");
      
} // hough_and_scale


// Perform voting for Hough transform
//___________________________________________________________________________
void  fill_stemp(ltatrees &vtrees, int numt, const landmarks &lm, igrid &stemp, int dimx, int dimy, int pad)
{
  igrid rid,rip;
  image ima;

  
  // FOr graphic puposes, showing how many times a pixel is accessed

   if (houghpar1!=0)
   {
    ima.init(dimx,dimy);
    rid.init(dimx,dimy);
    rip.init(dimx,dimy);
    rid.clear();
   }
  
  igrid temp;     
  temp.init(dimx/savefac + 2 * pad, dimy/savefac + 2 * pad);

  //Count number of hits at each pixel in stemp
  int totinst=0;
  for (int curr = 0; curr < numt; curr++)
    {
      
          if (houghpar1>0)
	    rip.clear();
      
	  int currtype = curr;
	  if (onlyedges || ltausedepth==0)
	    currtype=vtrees[curr].tnear;


	  temp.clear();
	  const int numins = lm.type_count(currtype);
	  totinst += numins;
	  for (int currinst = 0; currinst < numins; ++currinst)
	    {
	      int x1=lm.location_x(currtype, currinst);
	      int y1=lm.location_y(currtype, currinst);
	      
	      if (houghpar1>0)
		rip(x1,y1)=1;
	      
	      
	      for (int ind = 0; ind < hougharea[curr].dimx(); ind+=1)
		{
		  int x2 = pad + x1 + hougharea[curr](ind,0);
		  int y2 = pad + y1 + hougharea[curr](ind,1);
		  // Do not vote twice for the same pixel from two
		  // different features of the same type and same model locations
		  if (temp(x2, y2)==0) 
		    {
		      stemp(x2, y2)+=1;
		      temp(x2, y2)=1;
		      
		    }
		}
	      
	}

      
      if (houghpar1>0)
	{
	  viewport vpw;

       
	  gridfun((int)(pow(stemp(pad+i,pad+j),houghpar1)*houghpar2),rid);
	  rid.normalize(imrange*numcolors, (1-imrange)*numcolors-1);
	  gridfun(rid(i,j),ima);
	  vpw.add_grid(ima);
	  for (int r=0; r< rid.dimx(); r++)
           for (int s=0; s < rid.dimy(); s++)
	     {
	       if (rip(r,s)>0)
		 {
		 if (basecol>0)
		   vpw.add_string(r,s, "+",basecol,0);
		 else
		 if (basecol==0)
		   vpw.add_point(r,s);
		 }
		 
	     }
	  vpw.view();
	  }
     
    }

 if (houghpar1<0)
	{
	  viewport vpw;
	  gridfun((int)(pow(stemp(pad+i,pad+j),fabs(houghpar1))*houghpar2),rid);
	  rid.normalize(imrange*numcolors, (1-imrange)*numcolors-1);
	  gridfun(rid(i,j),ima);
	  vpw.add_grid(ima);
	  vpw.view();
	  }
 cout << lm.total_instances() << "total " << totinst << " density " << double(totinst)/(dimx*dimy*numt) << endl;  
}


int comph(const void *a,const  void *b)
{
  int *q=(int *)(a);
  int *r=(int *)(b);
  if ( q[0] < r[0])
    return 1;
  else if (q[0] > r[0])
    return -1;
  
  if (q[1]<r[1])
    return 1;
  else if (q[1]>r[1])
    return -1;
  
  return 0;
}
// __________________________________________________________________________


// Prepare vectores with Hough transform areas.
void fill_hough_areas()
{
  void comp_center(ltatrees &vtrees, int numt, int& centx, int& centy);
  void fill(ltatrees &vtrees, int numt);

simpstrg direc=ltatreedir;
 if (!onlyedges)  // Read in the lta trees
    numltas=read_ltatrees(ltrees, direc);
 maxnumltas=numltas;


 cout << " Numltas " << numltas << endl;
 direc=edgetreedir;
 numedgetrees=0;
 if (edgesforscale)
  numedgetrees=read_ltatrees(etrees, direc);

 comp_center(ltrees, numltas, centx, centy);
 comp_center(etrees, numedgetrees, ecentx, ecenty);
 /*
 if (edgesforscale)
   {
     ltatree1x += (ecentx-ltacentx);
     ltatree1y += (ecenty-ltacenty);
     ltatree2x += (ecentx-ltacentx);
     ltatree2y += (ecenty-ltacenty);
     ltatree3x += (ecentx-ltacentx);
     ltatree3y += (ecenty-ltacenty);
   }
 */

 if (onlyedges)
   {
     void fille(ltatrees &vtrees, int numt);
     numltas=numedgetrees;
     fille(etrees, numedgetrees);
   }
 else
   fill(ltrees, numltas);


 
}

void comp_center(ltatrees &vtrees, int numt, int& centx, int& centy)
{

  if (numt==0)
    return;
  centx=0;
  centy=0;
  for (int tt = 0; tt < numt; tt++) {
    
    centx += vtrees[tt].locationx;
    centy += vtrees[tt].locationy;
  }
  centx=centx/numt;
  centy=centy/numt;
  cout << " cent " << centx << " " << centy << endl;
}

// Linear fill areas along direction orthogonal to edge direction,
// +/- spread.
void fille(ltatrees &vtrees, int numt)
{

  hougharea.init(numt);

  
  //Calculate  areas of candidate origins relative to each lta
  // location in the model taking into account +/-20% scaling
  for (int currtype = 0; currtype < numt; currtype++)
    {

      int vecx = vtrees[currtype].locationx - ecentx;
      int vecy = vtrees[currtype].locationy - ecenty;
      hougharea[currtype].init(2*spread+1,2);
      int edgetype=vtrees[currtype].tnear;
      if (edgetype==0 || edgetype==1)
	{
	  for (int k=-spread;k<=spread; k++)
	    {
	      hougharea[currtype](k+spread,0)=-vecx;
	      hougharea[currtype](k+spread,1)=-vecy+k;
	    }
	}
      if (edgetype==2 || edgetype==3)
	{
	  for (int k=-spread;k<=spread; k++)
	    {
	      hougharea[currtype](k+spread,0)=-vecx-k;
	      hougharea[currtype](k+spread,1)=-vecy+k;
	    }
	}
      if (edgetype==4 || edgetype==5)
	{
	  for (int k=-spread;k<=spread; k++)
	    {
	      hougharea[currtype](k+spread,0)=-vecx+k;
	      hougharea[currtype](k+spread,1)=-vecy;
	    }
	}
      if (edgetype==6 || edgetype==7)
	{
	  for (int k=-spread;k<=spread; k++)
	    {
	      hougharea[currtype](k+spread,0)=-vecx+k;
	      hougharea[currtype](k+spread,1)=-vecy+k;
	    }
	}
 if (dum1==50)
       {
      cout << "area " << currtype << endl;
      cout << hougharea[currtype].dimx() << " " << hougharea[currtype].dimy() << endl;
       }
    }
}

// Areas determined according to 
// distance from center, to accomodate
// range of scales and rotations.

void fill(ltatrees &vtrees, int numt)
{

  hougharea.init(numt);
 
 
  if (toppercent>0)
    {
      double increment=.02;
      int numinc = (int)(rint((toppercent-botpercent)/increment));
      //Calculate  areas of candidate origins relative to each lta
      // location in the model taking into account +/-20% scaling
      int rroott=5;
      for (int currtype = 0; currtype < numt; currtype++)
	{
	  int ind = 0;
	  int tarea[500][2];
	  double vecx = vtrees[currtype].locationx - centx;
	  double vecy = vtrees[currtype].locationy - centy;
	  
	  double fn = sqrt(vecx*vecx+vecy*vecy);
	  // Unit vector orthogonal to [location-cent]
	  double orx = -vecy/fn;
	  double ory = vecx/fn;
	  // 2 % of vector.
	  double vrx=increment*vecx;
	  double vry=increment*vecy;
	  // At 73% of original length.
	  double x1 =  - botpercent*vecx - rroott*orx;
	  double y1 =  - botpercent*vecy - rroott*ory;
	  for (int a=0; a<numinc; a++)
	    {
	      x1 = x1-vrx; y1 = y1-vry;
	      double x2=x1;
	      double y2=y1;
	      for (int b = 0; b < rroott*2-1; b++) {
		x2 += orx;
		y2 += ory;
		tarea[ind][0] = int(rint(x2));
		tarea[ind][1] = int(rint(y2));
		ind ++;
	      }
	    }
	  if (coarse)
	    {
	      for (int ii=0; ii<ind; ii++)
		{
		  tarea[ii][0]=int(rint(tarea[ii][0]/savefac));
		  tarea[ii][1]=int(rint(tarea[ii][1]/savefac));
		}
	    }
	  int comph(const void *a,const  void *b);
	  qsort(tarea,ind,2*sizeof(int),comph);
	  int indf=0;
	  int ttarea[500][2];
	  
	  ttarea[0][0]=tarea[0][0];
	  ttarea[0][1]=tarea[0][1];
	  for (int ii=1; ii<ind && indf<500; ii++)
	    {
	      if (tarea[ii][0] != ttarea[indf][0] || tarea[ii][1] != ttarea[indf][1])
		{
		  indf++;
		  ttarea[indf][0]=tarea[ii][0];
		  ttarea[indf][1]=tarea[ii][1];
		}
	    }
	  
	  hougharea[currtype].init(indf,2);
	  gridfun(ttarea[i][j],hougharea[currtype]);
	
	  
	}
    }
  else
    {
      int sp=int(rint(-toppercent));
      cout << " SP " << sp << endl;
      for (int currtype = 0; currtype < numt; currtype++)
	{
	  int l = 0;
	  hougharea[currtype].init((2*sp+1)*(2*sp+1),2);
	  int vecx = int(rint(double(vtrees[currtype].locationx - centx)/savefac));
	  int vecy = int(rint(double(vtrees[currtype].locationy - centy)/savefac));
	  for (int s=-vecx-sp; s<=-vecx+sp; s++)
	    for (int t=-vecy-sp; t<=-vecy+sp; t++)
	      {
		hougharea[currtype](l,0)=s;
		hougharea[currtype](l,1)=t;
		l++;
	      }
	}
    }
  if (dum1==50)
    {
      for (int currtype=0; currtype<numt; currtype++)
	{
	  cout << double(vtrees[currtype].locationx-centx)/savefac << " ";
	  cout << double(vtrees[currtype].locationy-centy)/savefac << endl;
	  cout << "area " << currtype << endl;
	  cout << hougharea[currtype].dimx() << " " << hougharea[currtype].dimy() << endl;
	  hougharea[currtype].dump2();	
	}
    }
  
} // fill

// __________________________________________________________________________

void eval_detect(detection &det, ltatrees &ltrees, int numlt)
{
  double dx=0, dy=0, mx=0, my=0;
  int p=0;
  for (int t1 = 0; t1 < numlt; ++t1)
    {
      //      if (det.locs(t1,2)>0)
      //	{
	  dx += det.locs(t1,0);
	  dy += det.locs(t1,1);
	  mx += ltrees[t1].locationx;
	  my += ltrees[t1].locationy;
	  p++;
	  //	}
    }
  mx/=p; my/=p; dx/=p; dy/=p;
  double dmx=dx-mx;
  double dmy=dy-my;
  double cost=0;
  for (int t1 = 0; t1 < numlt; ++t1)
    {
      if (det.locs(t1,2)>0)
	{
	  cost += (det.locs(t1,0)-ltrees[t1].locationx-dmx)*
	    (det.locs(t1,0)-ltrees[t1].locationx-dmx);
	  cost +=  (det.locs(t1,1)-ltrees[t1].locationy-dmy)*
	    (det.locs(t1,1)-ltrees[t1].locationy-dmy);
	}
      else
	cost += 4*(spread)*(spread);
    }
  
  cout << " COST " << sqrt(cost/p) << endl;
  det.cartprobs3[0]=-sqrt(cost/p);
}

// Get the estimated scale, shift and locatin around candidate center x1,y1.
// This is done brute force, very inefficient. Also detected instantiation
// is recorded here
int getscale(ltatrees& ltrees, int coun, detections &tmpdets, 
              int numlt, int acchits, int x1, int y1, 
              int lcentx, int lcenty)
{
  int oldnumhits=0;
  double oldscale=1.;
  int oldcx = x1;
  int oldcy = y1;
  double oldang = 0.;
  igrid xt(numlt,3);
  igrid oxt(numlt,3);
  oxt.clear();


  if (acchits>=0 && numlt > 0)
  {
    for (int cx = x1-locshift; cx <= x1+locshift; cx++)
      for (int cy = y1-locshift; cy <= y1+locshift; cy++)
      {
	for (double scale=lowscale; scale <= highscale; scale += .1)
        {
	  xt.clear(-1);
	  int numhit = 0;
          for (int t1 = 0; t1 < numlt; ++t1)
          {
            const int loc1x = ltrees[t1].locationx - lcentx;
            const int loc1y = ltrees[t1].locationy - lcenty;
            int kc = cx + (lcentx-centx)+(int)rint(scale*loc1x);
            int lc = cy + (lcenty-centy)+(int)rint(scale*loc1y);
            int type=t1;
	    if (edgesforscale || ltausedepth==0 )
	      type = ltrees[t1].tnear;

            xt(t1,0)=kc; xt(t1,1)=lc;
            int logic=0;
            if (edgespread==0) 
	      // Get precise edge location, edges have not been spread.
	      {
		int curdis=spread+spread+1;
		int nd;
                for (int k=kc-spread; k<=kc+spread /*&& !logic*/; k++)
                for (int l=lc-spread; l<=lc+spread /*&& !logic */; l++)
		  {
                    if (STAGS[type].get(k,l) > 0 /*&& utemp.get(k,l) */)
		      {

			 if ((nd=abs(k-kc)+abs(l-lc)) < curdis)
			   {
			     xt(t1,0)=k; xt(t1,1)=l;
			     curdis=nd;
			   }
			 logic=1;
		      }
		  }
	      }
            else  // Edges have been spread, just check one location
	      {
		logic = (STAGS[type].get(kc,lc) > 0);       
	      }
	    if (logic)
	      {
		numhit++;
		xt(t1,2)=type;
	      }
          }
          if (numhit > oldnumhits)
          {
            oldscale=scale;
            oldcx=cx;
            oldcy=cy;
            oldnumhits=numhit;
	    oxt=xt;
          }
        }
      } 
  
    // Loop over possible rotations
    if (testrotations)
      for (double testang = -PI/10; testang <= PI/10 + FUZZ; testang += PI/20)
      {
	int numhit = 0;
	xt.clear(-1);
	for (int t1 = 0; t1 < numlt; ++t1)
	{
	  const int loc1x = ltrees[t1].locationx - centx;
	  const int loc1y = ltrees[t1].locationy - centy;
	  int kc = oldcx + (int)rint(oldscale * loc1x * cos(testang) -
				     oldscale * loc1y * sin(testang));
          int lc = oldcy + (int)rint(oldscale * loc1x * sin(testang) +
				     oldscale * loc1y * cos(testang));
	  int type=t1;
	  if (edgesforscale)
	    type = ltrees[t1].tnear;

	  xt(t1,0)=kc; xt(t1,1)=lc;
	  int logic=0;
	  if (edgespread==0) 
	    {
	      for (int k=kc-spread; k<=kc+spread && !logic; k++)
                for (int l=lc-spread; l<=lc+spread && !logic; l++)
		  {
                    if (STAGS[type].get(k,l) > 0)
		      {
			xt(t1,0)=k; xt(t1,1)=l;
			logic=1;
		      }
		  }
	    }
	  else 
	    {
	      logic = (STAGS[type].get(kc,lc) > 0);       
	    }
	  
	  if (logic)
	    {
	      numhit++;
              xt(t1,2)=t1;
	    }
	}
	if (numhit > oldnumhits)
	{
	  oldang = testang;
	  oldnumhits=numhit;
          oxt=xt;
	}
      }
  }
  else if (acchits==-1)
      oldnumhits=tmpdets[coun].counts[0];
  // If maximal number of hits greater than threshold detection survives
  if (oldnumhits>acchits*accfactor*boxfactorx)
  {
    double cs=1.;
    double sn=0.;
    if (testrotations)
      {
      cs=cos(oldang);
      sn=sin(oldang);
      }
    // Computer detection triangle.

    tmpdets[coun].tripts[0] = oldcx+
      (int)rint(oldscale*(ltatree1x - centx) * cs -
		oldscale*(ltatree1y - centy) * sn);
    tmpdets[coun].tripts[1] = oldcy+
      (int)rint(oldscale*(ltatree1x - centx) * sn +
		oldscale*(ltatree1y - centy) * cs);
    tmpdets[coun].tripts[2] = oldcx+
      (int)rint(oldscale*(ltatree2x - centx) * cs -
		oldscale*(ltatree2y - centy) * sn);
    tmpdets[coun].tripts[3] = oldcy+
      (int)rint(oldscale*(ltatree2x - centx) * sn +
		oldscale*(ltatree2y - centy) * cs);
    tmpdets[coun].tripts[4] = oldcx+
      (int)rint(oldscale*(ltatree3x - centx) * cs -
		oldscale*(ltatree3y - centy) * sn);
    tmpdets[coun].tripts[5] = oldcy+
      (int)rint(oldscale*(ltatree3x - centx) * sn +
		oldscale*(ltatree3y - centy) * cs);
    tmpdets[coun].tripts[6] = oldcx;
    tmpdets[coun].tripts[7] = oldcy;

    // Record number of detected model points.
    tmpdets[coun].counts[0]=oldnumhits;
    // Vector of locations of detected and estimated locations
    // of undetected model points.
    tmpdets[coun].locs=oxt;
    tmpdets[coun].initcart(numcarts, itersize, numclass);
    if (showbest==7)
      eval_detect(tmpdets[coun], ltrees, numlt);
    return(1);
  }
  
  return(0);
} // getscale
// __________________________________________________________________________
// Get the estimated scale, shift and locatin around candidate center x1,y1.
// This is done brute force, very inefficient. Also detected instantiation
// is recorded here
 void getsxy(double scale, int loc1x,int  loc1y,int & sx,int & sy)
	      {
		sx=(int)rint(scale*loc1x);
		sy=(int)rint(scale*loc1y);
	      }

int getscl(ltatrees& ltrees, int coun, detections &tmpdets, 
              int numlt, int acchits, int x1, int y1, 
              int lcentx, int lcenty)
{
  void getsxy(double scale, int loc1x,int  loc1y,int & sx,int & sy);
  int oldnumhits=0;
  double oldscale=1.;
  int oldcx = x1;
  int oldcy = y1;
  double oldang = 0.;

  int sx, sy;
  if (acchits>=0 && numlt > 0)
  {
    for (double scale=lowscale; scale <= highscale; scale += .1)
      {
	igrid numhitz(2*locshift+1, 2*locshift+1);
	numhitz.clear();
	for (int t1 = 0; t1 < numlt; ++t1)
          {
            const int loc1x = ltrees[t1].locationx - lcentx;
            const int loc1y = ltrees[t1].locationy - lcenty;
	    getsxy(scale, loc1x, loc1y, sx, sy);
	   
            
	      for (int x = -locshift; x <= locshift; x+=1)
		for (int y = -locshift; y <= locshift; y+=1)
		  {
		    int kc = x1 + x + (lcentx-centx)+sx;
		    int lc = y1 + y + (lcenty-centy)+sy;

		    int type=t1;
		    if (edgesforscale || ltausedepth==0 )
		    type = ltrees[t1].tnear;
		    //  xt(t1,0)=kc; xt(t1,1)=lc;
		    int logic=0;
		    if (edgespread==0) 
		      // Get precise edge location, edges have not been spread.
		      {
			int curdis=spread+spread+1;
			int nd;
			for (int k=kc-spread; k<=kc+spread && !logic; k++)
			  for (int l=lc-spread; l<=lc+spread && !logic; l++)
			    {
			      if (STAGS[type].get(k,l) > 0)
				{
				  logic=1;
				}
			    }
		      }
		    else  // Edges have been spread, just check one location
		      {
			logic = (STAGS[type].get(kc,lc) > 0);       
		      }
		    if (logic)
		      {
			numhitz(locshift+x,locshift+y)++;
			//xt(t1,2)=type;
		      }
		  }
	  }
	for (int x=-locshift; x<=locshift; x++)
	  for (int y=-locshift; y<=locshift; y++)
	  {
	    if (numhitz(locshift+x, locshift+y)>oldnumhits)
	      {
		oldnumhits=numhitz(locshift+x, locshift+y);
		oldscale=scale;
		oldcx=x1+x;
		oldcy=y1+y;
	      }
	  }
      }
  
    // Loop over possible rotations
    if (testrotations)
      for (double testang = -PI/10; testang <= PI/10 + FUZZ; testang += PI/20)
      {
	int numhit = 0;
	for (int t1 = 0; t1 < numlt; ++t1)
	{
	  const int loc1x = ltrees[t1].locationx - centx;
	  const int loc1y = ltrees[t1].locationy - centy;
	  int kc = oldcx + (int)rint(oldscale * loc1x * cos(testang) -
				     oldscale * loc1y * sin(testang));
          int lc = oldcy + (int)rint(oldscale * loc1x * sin(testang) +
				     oldscale * loc1y * cos(testang));
	  int type=t1;
	  if (edgesforscale)
	    type = ltrees[t1].tnear;

	  int logic=0;
	  if (edgespread==0) 
	    {
	      for (int k=kc-spread; k<=kc+spread && !logic; k++)
                for (int l=lc-spread; l<=lc+spread && !logic; l++)
		  {
                    if (STAGS[type].get(k,l) > 0)
		      {
			logic=1;
		      }
		  }
	    }
	  else 
	    {
	      logic = (STAGS[type].get(kc,lc) > 0);       
	    }
	  
	  if (logic)
	    {
	      numhit++;
	    }
	}
	if (numhit > oldnumhits)
	{
	  oldang = testang;
	  oldnumhits=numhit;
	}
      }
  }
  else if (acchits==-1)
      oldnumhits=tmpdets[coun].counts[0];
  // If maximal number of hits greater than threshold detection survives

  if (oldnumhits>acchits*accfactor*boxfactorx)
  {
    double cs=1.;
    double sn=0.;
    if (testrotations)
      {
      cs=cos(oldang);
      sn=sin(oldang);
      }
    // Computer detection triangle.

    tmpdets[coun].tripts[0] = oldcx+
      (int)rint(oldscale*(ltatree1x - centx) * cs -
		oldscale*(ltatree1y - centy) * sn);
    tmpdets[coun].tripts[1] = oldcy+
      (int)rint(oldscale*(ltatree1x - centx) * sn +
		oldscale*(ltatree1y - centy) * cs);
    tmpdets[coun].tripts[2] = oldcx+
      (int)rint(oldscale*(ltatree2x - centx) * cs -
		oldscale*(ltatree2y - centy) * sn);
    tmpdets[coun].tripts[3] = oldcy+
      (int)rint(oldscale*(ltatree2x - centx) * sn +
		oldscale*(ltatree2y - centy) * cs);
    tmpdets[coun].tripts[4] = oldcx+
      (int)rint(oldscale*(ltatree3x - centx) * cs -
		oldscale*(ltatree3y - centy) * sn);
    tmpdets[coun].tripts[5] = oldcy+
      (int)rint(oldscale*(ltatree3x - centx) * sn +
		oldscale*(ltatree3y - centy) * cs);
    tmpdets[coun].tripts[6] = oldcx;
    tmpdets[coun].tripts[7] = oldcy;

    // Record number of detected model points.
    tmpdets[coun].counts[0]=oldnumhits;
    // Vector of locations of detected and estimated locations
    // of undetected model points.
    tmpdets[coun].initcart(numcarts, itersize, numclass);
    return(1);
  }
  
  return(0);
} // getscale
// __________________________________________________________________________


// Read in cart trees and compute information on relevant reference
// grid and anchor points stored in these trees
void get_trees(partialtrans &pt, partialtrans &fpt)
{
  int i,j;
  ctrees.init(numcarts);

  // Read in object against rest of the world cart trees
  for (i = 0; i < numcarts; ++i)
    {
    if (!ctrees[i].read(carttreedir + simpstrg("/ctree") + ltos(i))) 
      {
	cerr << "Unable to read in carttrees.\n";
	exit(1);
      }
    if (i==0)
      {
        cartstyle=ctrees[0].carttype;
        cartspread=ctrees[0].cartsprd;
        cartsizex=ctrees[0].cartsx;
        cartsizey=ctrees[0].cartsy;
        cout << "cartstyle " << cartstyle << endl;
        cout << "cartspread " << cartspread << endl;
        cout << cartsizex << endl;
        ctrees[0].points.dump(); 
	
	pt.ptr(Pixel(ctrees[0].points(0),ctrees[0].points(1)),
	       Pixel(ctrees[0].points(2),ctrees[0].points(3)),
	       Pixel(ctrees[0].points(4),ctrees[0].points(5)));
      }
    }
  
  // Read in the classification cart trees
  ftrees.init(fnumcarts);
  carthistsize=numclass;
  for (i = 0; i < fnumcarts; ++i)
    {
      if (!ftrees[i].read(fcarttreedir + simpstrg("/ctree") + ltos(i))) 
	{
	  cerr << "Unable to read in face carttrees.\n";
	  exit(1);
	}
      if (i==0)
	{
	  fcartstyle=cartstyle=ftrees[0].carttype;
	  cartspread=ftrees[0].cartsprd;
	  fcartsizex=ftrees[0].cartsx;
	  fcartsizey=ftrees[0].cartsy;
          fpoint1x=ftrees[0].points(0);
          fpoint1y=ftrees[0].points(1);
	  fpoint2x=ftrees[0].points(2);
          fpoint2y=ftrees[0].points(3);
	  fpoint3x=ftrees[0].points(4);
          fpoint3y=ftrees[0].points(5);
	  cout << "fcartstyle " << fcartstyle << " " << cartstyle << endl;
        cout << "fcartspread " << cartspread << endl;
	if (dum2 < 0)
	  cartspread=-dum2-1;
        cout << fcartsizex << " " << fcartsizey << endl;
        ftrees[0].points.dump();
	fpt.ptr(Pixel(ftrees[0].points(0),ftrees[0].points(1)),
		Pixel(ftrees[0].points(2),ftrees[0].points(3)),
		Pixel(ftrees[0].points(4),ftrees[0].points(5)));
	}
    }

  // Read in object class against rest of classes cart trees
  if (fcart2)
    {
      ftreess.init(numclass);
      for (i = 0; i < numclass; ++i)
	{
	  ftreess[i].init(fnumcarts);
	  simpstrg tmpdir;
	  tmpdir = fcarttreedir + ltos(i);
	  for (j = 0; j < fnumcarts; ++j) 
	    {
	      if (!ftreess[i][j].read(tmpdir + simpstrg("/ctree") + ltos(j))) 
		{
		  cerr << "Unable to read in face carttrees.\n";
		  exit(1);
		}
	    }
	}
    }
}

#endif // FINDDETS_C
