// Copyright YALI AMIT, January 15, 2001.

#ifndef HISTOGRAM_H
#define HISTOGRAM_H

// __________________________________________________________________________
// Include files

#include <iostream>

// __________________________________________________________________________
// Local declarations and typedefs

class histogram;
typedef double (histogram::*impfunction)();

// __________________________________________________________________________

class histogram
{
 private:
  int size;
  float *counts;

 public:

  histogram(int s);
  ~histogram();
  void operator=(const histogram& m);
  float& operator[](int i);
  int dim() const { return size; }
  float total();
  int mode();
  int mode2();
  int pure(double p = 1.0);
  double unsup();
  double entropy();
  double gini();
  double kolsmir();
  void write(ostream& os);
  void read(istream& is);
  void clear();
  void cop(histogram &m);
};


// __________________________________________________________________________

inline histogram::histogram(int s)
{
  assert(s > 0);
  size = s;
  counts = new float[size];
  for (int i = 0; i < s; ++i)
    counts[i] = 0;
}

// __________________________________________________________________________

inline histogram::~histogram()
{
  delete [] counts;
}

// __________________________________________________________________________

inline void histogram::operator=(const histogram& m)
{
  assert(size == m.size);

  if (size != m.size)
    return;

  for (int i = 0; i < size; ++i)
    counts[i] = m.counts[i];
}

// __________________________________________________________________________

inline float& histogram::operator[](int i)
{
  assert(i >= 0 && i < size);
  return counts[i];
}

// __________________________________________________________________________

inline float histogram::total()
{
  float t = 0;
  for (int i = 0; i < size; ++i)

    t += counts[i];
  return t;
}

// __________________________________________________________________________

inline int histogram::mode()
{
  int m = 0;
  float count = -1;
  for (int i = 0; i < size; ++i)
    if (counts[i] > count)
    {
      count = counts[i];
      m = i;
    }
  return m;
}
// __________________________________________________________________________

inline int histogram::mode2()
{
  int m1 = mode();
  int m2 = 0;
  float count2 = -1;
  for (int i = 0; i < size; ++i)
    if (counts[i] > count2 && i != m1)
    {
      count2 = counts[i];
      m2 = i;
    }
  return m2;
}

// __________________________________________________________________________

inline int histogram::pure(double p)
{
  double t = total();
  for (int i = 0; i < size; ++i)
    if (counts[i] / t >= p)
      return 1;
  return 0;
}

// __________________________________________________________________________

inline void histogram::write(ostream& os)
{
  varwrite(os, size);
  arraywrite(os, counts, size);
}

// __________________________________________________________________________

inline void histogram::read(istream& is)
{
  //static int sizewarn = 0;
  int oldsize;
  varread(is, oldsize);
  //if (oldsize > 10 || oldsize < 0)
  //  oldsize = size;
  if (oldsize > size)
   {
     /*   
    if (!sizewarn)
    {
      sizewarn = 1;
      cerr << "Current histogram is too small to read all entries.\n"
           << "Current size is " << size << ", and read in size is " 
           << oldsize << ".\n";
    }
     */
    oldsize = size;
  }
  arrayread(is, counts, oldsize);
}

// __________________________________________________________________________

inline void histogram::clear()
{
  for (int i = 0; i < size; ++i)
    counts[i] = 0;
}

// __________________________________________________________________________
// in histgram.C

extern double ave_impurity(histogram& h1, histogram& h2);
extern void set_impurity_func(const char * impmeasure, 
			      int exitflag = 1);

// __________________________________________________________________________
// histgram.h
#endif
