// Copyright YALI AMIT, January 15, 2001.

#define LANDMARK_C



// __________________________________________________________________________
// Includes

#include <cctype>
#include <cstdio>
#include <iostream>
#include <iomanip>
#include <fstream>

#include "includes.h"
#include "simpstrg.h"
#include "image.h"
#include "landmark.h"
#include "globals.h"
#include "pixel.h"
#include "avlgrid.h"
// ________________________________________________________________________*/
// Finds the 'index' landmark in the file 'lmfilename' and puts it's
// data in 'currlm'.  'index' starts at 0.
extern ucgrids STAGS;




int landmarks::read(int index)
{
  // Simple error check
  if (index < 0)
    return 0;

  // Initialize some parameters.
  int land_index = index % 100;
  int file_index = index / 100;
  simpstrg data_file = landdir + "/Lands_" + ltos(file_index);

  // Open the landmark file.
  ifstream ifs;
  if (!data_dir_open(ifs, data_file, 0))
    {
      cerr << " Can not read landmarks from file " << data_file << endl;
      cerr << " You may be able to make them by going to the directory " << endl;
      cerr << " and running make_data " << endl;
      exit(1);
    }

  // Make sure we're getting the data for a landmark that really exists.
  int entries;
  varread(ifs, entries);

  if (land_index >= entries)
    return 0;

  // First skip to where the offset to the landmarks are, read in the offset,
  // and move to the offset.
  ifs.seekg((land_index + 1) * Sint);
  int offset;
  varread(ifs, offset);
  ifs.seekg(offset);
  // Determine the true value of the data with these landmarks.
  varread(ifs, value);

  // Read in the coordinates of the three distinguished points.
  arrayread(ifs, coords, 6);
  // Read in the typecounts
  if (!typecounts.read(ifs))
    return 0;

  // Determine the typeoffsets
  offset = 0;
  typeoffsets.init(typecounts.dimx());
  for (int i = 0; i < typecounts.dimx(); ++i)
  {
    typeoffsets(i) = offset;
    offset += typecounts(i);
  }

  // Read in the locations
  if (!locations.read(ifs))
    return 0;

  return 1;

} // read

// __________________________________________________________________________

int landmarks::size()
{
  return (Sint + 6 * Sint + typecounts.size() + locations.size());

} // landmarks::size

// __________________________________________________________________________

int landmarks::write(ofstream& ofs)
{
  varwrite(ofs, value);
  arraywrite(ofs, coords, 6);
  if (typecounts.write(ofs) && locations.write(ofs))
    return 1;
  else
    return 0;

} // landmarks::write

// __________________________________________________________________________

int landmarks::read(ifstream& ifs)
{
  ifs.read((char*)&value, sizeof(int));
  int nument;
  ifs.read((char*)&nument, sizeof(int));
  init_counts_offsets(nument);
  for (int i=0; i<nument; i++)
    {
      ifs.read((char*)&typecounts(i), sizeof(int));
      ifs.read((char*)&typeoffsets(i), sizeof(int));
    }
  // Read in the bounds.
  int bounds[4];
  ifs.read((char*)bounds, 4 * sizeof(int));
  int numlocs=bounds[1]+1;
  
  if (numlocs>0)
    {

      
      init_locations(numlocs);
      for (int i = 0; i < numlocs; ++i)
	{
          unsigned char uc;
	  ifs.read((char*)&uc, sizeof(unsigned char));
          locations(i,0)=uc;
	  ifs.read((char*)&uc, sizeof(unsigned char));
          locations(i,1)=uc;
	}
    }

  return 1;
}

/*
int landmarks::read(ifstream& ifs)
{
  varread(ifs, value);
  arrayread(ifs, coords, 6);
  if (typecounts.read(ifs) && locations.read(ifs))
    return 1;
  else
    return 0;

} // landmarks::write
*/

void landmarks::merge(landmarks &ne)
{

 // First set up the type counts;
  igrid temp_co(total_types()+ne.total_types());
  igrid temp_of(total_types()+ne.total_types());
  
  
  
  int i;
  for (i = 0; i < total_types(); i++)
    {
      temp_co(i) = type_count(i);
      temp_of(i) = type_offset(i);
    }
  int offs = total_instances();
  for (i = total_types(); i < total_types()+ne.total_types(); i++)
    {  
      temp_co(i) = ne.type_count(i-total_types());
      temp_of(i) = offs;
      offs += temp_co(i);
    }

  typecounts=temp_co;
  typeoffsets=temp_of;

  igrid templo(locations.dimx()+ne.total_instances(),2);
  
  for (i = 0; i < locations.dimx(); i++)
    {
     templo(i,0) = locations(i,0);
     templo(i,1) = locations(i,1);
    }
  for (i = locations.dimx(); i < locations.dimx() + ne.total_instances(); i++)
    {
  templo(i,0) = ne.location_x(i-locations.dimx());
  templo(i,1) = ne.location_y(i-locations.dimx());
    }
  locations=templo;
}


// __________________________________________________________________________
// Prints out the contents of a landmark

void landmarks::dump()
{
  int i;

  // Print out the landmark data.
  cout << "Truth is [" << truth();
  if (isprint(truth())) 
    cout << " (" << char(truth()) << ")";
  cout << "]\n";
  cout << "Registered coordinates are [" << coords[0] << ", " << coords[1] 
    << "], [" << coords[2] << ", " << coords[3] << "] and [" << coords[4]
      << ", " << coords[5] << "]\n";
  cout << "Number of entries is [" << total_types() << "]\n";
  if (total_types() > 0)
  {
    for (i = 0; i < total_types(); ++i)
    {
      cout << "  (" << i << ") - Number = [" << type_count(i)
	<< "], Offset = [" << type_offset(i) << "]\n";
    }
    cout << "There are " << total_instances() << " total instances.\n";
    if (total_instances() > 0)
    {
      cout << "The locations are the following:\n";
      for (i = 0; i < total_instances(); i += 7)
      {
	for (int j = 0; j < 7 && i + j < total_instances(); ++j)
	  cout << "  [" << setw(2) << location_x(i + j) << ", " 
            << setw(2) << int(location_y(i + j)) << "] ";
	cout << endl;
      }
    }
  }

} // dump



// __________________________________________________________________________

void write_lands(const char* landfile, int count, lmarkslist& lands)
{
  int namecount=firstiter;
  int fsize=100;
  for (int nc=0; nc < count; nc+=100)
    {
      simpstrg filen = simpstrg(landfile) + "/Lands_" + ltos(namecount++);
      Ofstream(ofs, filen);
      int lsize=MIN(fsize, count-nc);
      int offset = (1 + lsize) * Sint;
      varwrite(ofs, lsize);
      varwrite(ofs, offset);
      for (int i = nc+1; i < nc+lsize; ++i)
	{
	  offset += lands[i - 1].size();
	  varwrite(ofs, offset);
	}
      for (int i = nc; i < nc+lsize; ++i)
	lands[i].write(ofs);
    }

} // write_lands

// __________________________________________________________________________

void read_lands(const char* landfile, int count, lmarkslist& lands)
{
  Ifstream(ifs, landfile);
  if (!ifs)
  {
    cerr << "Unable to read landmark file\n";
    exit(1);
  }

  int dummy;
  varread(ifs, dummy);
  int i;
  for (i = 0; i < dummy; ++i)
    varread(ifs, dummy);
  for (i = 0; i < count; ++i)
    lands[i].read(ifs);

} // read_lands

// __________________________________________________________________________
// Transform an array of landmark locations into a landmark structure

void landmarks::setup_from_lmarklist(lmarklist& land, int numlands)
{
  // First set up the type counts;
  typecounts.clear();

  int i;
  for (i = 0; i < numlands; ++i)
    ++typecounts(land[i].index);

  // Next set up the type offsets;
  typeoffsets(0) = 0;
  for (i = 1; i < total_types(); ++i)
    typeoffsets(i) = typeoffsets(i - 1) + typecounts(i - 1);

  // Initialize and fill in the locations
  init_locations(numlands);
  igrid offsets(total_types());
  offsets.clear();
  for (i = 0; i < numlands; ++i)
  {
    int ind = land[i].index;
    location_x(ind, offsets(ind)) = land[i].cx;
    location_y(ind, offsets(ind)) = land[i].cy;
    ++offsets(ind);
  }

} // setup_from_lmarklist

void find_maxmin(lmarkslist &lmarks,int &globalminx,int &globalmaxx,int &globalminy,int &globalmaxy)
{
  globalminx = 1000000, globalminy = 1000000,
    globalmaxx = -1000000, globalmaxy = -1000000;
  
  for (int ii = 0; ii < lmarks.dim(); ++ii) 
    {
      for (int j = 0; j < lmarks(ii).total_instances(); ++j) 
	{
	  globalminx = MIN(globalminx, lmarks(ii).location_x(j));
	  globalminy = MIN(globalminy, lmarks(ii).location_y(j));
	  globalmaxx = MAX(globalmaxx, lmarks(ii).location_x(j));
	  globalmaxy = MAX(globalmaxy, lmarks(ii).location_y(j));
	}
    }
  
  cout << "global extremes " << globalminx << " " << globalmaxx << " " << globalminy << " " <<
    globalmaxy << endl;
  cout << globalmaxx - globalminx+2<< " " << globalmaxy -globalminy+2<< endl;
}


void spread_and_aggreg(landmarks &lands, int numt, igrids &temp)
{


  ucgrid stag(temp[0].dimx(), temp[0].dimy());
  stag.clear();
  for (int currtype=0; currtype < numt; currtype++)
    {
      const int numins = lands.type_count(currtype);
      
      for (int currinst = 0; currinst < numins; ++currinst)
	{
	  const int x = lands.location_x(currtype, currinst);
	  const int y = lands.location_y(currtype, currinst);
	  const int minx = MIN(spread, x);
	  const int miny = MIN(spread, y);
	  const int maxx = MIN(spread, stag.dimx() - x - 1);
	  const int maxy = MIN(spread, stag.dimy() - y - 1);
	  for (int ii = -minx; ii <= maxx; ++ii)
	    for (int j = -miny; j <= maxy; ++j)
	      stag(x + ii, y + j) = 1;
	  
	}
      gridfun(  stag(i,j)+temp[currtype](i,j)  ,temp[currtype]);
      stag.clear();
    }
}
void shift_lm(landmarks &l, int sftx, int sfty)
{

  for (int i=0; i<l.total_instances(); i++)
    {
      l.location_x(i)+=sftx;
      l.location_y(i)+=sfty;
    }
}



 void affine_trans_lm(landmarks &l, affinetrans &pqtrans, lmarklist &l2, int &numlands)
{
  numlands=0; 
  static ucgrid check(STAGSdim,STAGSdim);
  for (int j = 0; j < l.total_types(); ++j)
    {
   
      check.clear();
      for (int k = 0; k < l.type_count(j); ++k)
	{
	  l2[numlands].index = j;
	  Pixel t(l.location_x(j, k), l.location_y(j, k)), newt;
	  newt = pqtrans * t;
          
	  //newt = newt.TransXY(shift, shift);
	  if (++check(newt[0]+STAGSdim/4,newt[1]+STAGSdim/4) == 1)
	    {
	      l2[numlands].cx = newt[0];
	      l2[numlands].cy = newt[1];

	      ++numlands;

	    }
	}
    }
}


void cluster_landmarks(lmarklist& lands, int clustsize, 
		       int numtypes, int& total)
{
  // Return now if there is no clustering
  if (clustsize < 1 || total == 0)
    return;

  // Determine extents
  int minx = 10000;
  int miny = 10000;
  int maxx = 0;
  int maxy = 0;
  int i;
  for (i = 0; i < total; ++i)
  {
    if (lands[i].cx < minx) minx = lands[i].cx;
    if (lands[i].cx > maxx) maxx = lands[i].cx;
    if (lands[i].cy < miny) miny = lands[i].cy;
    if (lands[i].cy > maxy) maxy = lands[i].cy;
  }



  int xclusters = (maxx - minx + 1) / clustsize + 1;
  int yclusters = (maxy - miny + 1) / clustsize + 1;
  igrid numfound(xclusters, yclusters);
  igrid orig(xclusters, yclusters);
  for (int currtype = 0; currtype < numtypes; ++currtype)
  {

    numfound.clear();
    for (i = 0; i < total; ++i)
    {
      if (lands[i].index != currtype)
	continue;

      // Determine which cluster the instance belongs to
      int x = (lands[i].cx - minx) / clustsize;
      int y = (lands[i].cy - miny) / clustsize;
      if (++numfound(x, y) == 1)
	orig(x, y) = i;
      else
      {
	lands[orig(x, y)].cx += lands[i].cx;
	lands[orig(x, y)].cy += lands[i].cy;
	swap(lands[i--], lands[--total]);
      }
    }
    for (i = 0; i < xclusters; ++i)
    {
      for (int j = 0; j < yclusters; ++j)
      {
	if (numfound(i, j) > 0)
	{
        
  	  lands[orig(i, j)].cx /= numfound(i, j);
	  lands[orig(i, j)].cy /= numfound(i, j);

	}
      }
    }
  }

} // cluster_landmarks

// ___________________________________________________________________________
/*
void clust_landmarks(lmarklist& lands, int clustsize, 
                     int numtypes, int& total)
{
  // Return now if there is no clustering
  if (clustsize < 1 || total == 0)
    return;
 

  vector<iavlgrid> tmpinsts(numtypes);
  static ivector l(2);
  for (int k=0; k<numtypes; k++)
    tmpinsts[k].ainit(10000, 2);


  static lmarklist tlands(maxlands); 
  for (int i = 0; i < total; ++i)
  {
      l[0] = (lands[i].cx) / clustsize;
      l[1] = (lands[i].cy) / clustsize;
      int currtype=lands[i].index;
      tmpinsts[currtype].insert(l);
  }
  int ind=0;
  for (int currtype = 0; currtype<numtypes; ++currtype)
    {
      for (int i = tmpinsts[currtype].first(); 
	   i != -1; i = tmpinsts[currtype].next(i)) 
	{
	  lands[ind].cx=tmpinsts[currtype](i,0);
	  lands[ind].cy=tmpinsts[currtype](i,1);
	  lands[ind].index=currtype;
	  ind++;
	}
    }
  total=ind;


} // clust_landmarks
*/
/*
void clust_landmarks(lmarklist& lands, int clustsize, 
                     int numtypes, int& total)
{
  // Return now if there is no clustering
  if (clustsize < 1 || total == 0)
    return;
  // Determine extents
  int maxx = 0;
  int maxy = 0;
  int i;
  for (i = 0; i < total; ++i)
  {
    if (lands[i].cx > maxx) maxx = lands[i].cx;
    if (lands[i].cy > maxy) maxy = lands[i].cy;
  }

//   << minx << " " << miny << endl;


  int xclusters = (maxx + 1) / clustsize + 1;
  int yclusters = (maxy + 1) / clustsize + 1;
  igrid numfound(xclusters, yclusters);
  const int maxlands = 100000;
  //const int maxlands = 10000;
  static lmarklist tlands(maxlands);
  int ind=0;
  numfound.clear(-1);
  for (int currtype = 0; currtype < numtypes; ++currtype)
  {

    for (i = 0; i < total; ++i)
    {
      if (lands[i].index != currtype)
	continue;

      // Determine which cluster the instance belongs to
      int x = (lands[i].cx) / clustsize;
      int y = (lands[i].cy) / clustsize;
      numfound(x, y)=currtype;
    }
    for (i = 0; i < xclusters; ++i)
    {
      for (int j = 0; j < yclusters; ++j)
      {
	if (numfound(i, j) == currtype)
	{
	  tlands[ind].cx =i;
	  tlands[ind].cy =j;
          tlands[ind].index=currtype;
          ind++;
	}
      }
    }
  }
  total=ind;
  for (i=0; i<ind;i++)
    {
       lands[i].cx=tlands[i].cx;
       lands[i].cy=tlands[i].cy;
       lands[i].index=tlands[i].index;
    }

} // clust_landmarks
*/
void clust_landmarks(lmarklist& lands, int clustsize, 
                     int numtypes, int& total)
{
  // Return now if there is no clustering
  if (clustsize < 1 || total == 0)
    return;
  // Determine extents
  int maxx = 0;
  int maxy = 0;
  int i;
  for (i = 0; i < total; ++i)
  {
    if (lands[i].cx > maxx) maxx = lands[i].cx;
    if (lands[i].cy > maxy) maxy = lands[i].cy;
  }

//   << minx << " " << miny << endl;


  int xclusters = (maxx + 1) / clustsize + 1;
  int yclusters = (maxy + 1) / clustsize + 1;
  igrids numfound(numtypes);
  
  const int maxlands = 100000;
  //const int maxlands = 10000;
  static lmarklist tlands(maxlands);
  int ind=0;

  for (int currtype = 0; currtype < numtypes; ++currtype)
  {
    numfound[currtype].init(xclusters, yclusters);
    numfound[currtype].clear();
  }
  for (i = 0; i < total; ++i)
    {
      int x = (lands[i].cx) / clustsize;
      int y = (lands[i].cy) / clustsize;
      if (!numfound[lands[i].index](x,y))
	{
	  tlands[ind].cx =x;
	  tlands[ind].cy =y;
          tlands[ind].index=lands[i].index;
	  numfound[lands[i].index](x,y)=1;
          ind++;
	}
    }
  total=ind;
  for (i=0; i<ind;i++)
    {
       lands[i].cx=tlands[i].cx;
       lands[i].cy=tlands[i].cy;
       lands[i].index=tlands[i].index;
    }

} // clust_landmarks


// __________________________________________________________________________
void register_landmarks(affinetrans& at, ucgrids& newlands, int unions)
{
  int i, j, k;
  // Loop over landmark types
  for (i = 0; i < newlands.dim(); ++i)
    newlands[i].clear();

  // Loop over x and y coordinates
  if (!unions) 
    {
    Pixel P;
    for (j = 0; j < newlands[0].dimx(); ++j)
      for (k = 0; k < newlands[0].dimy(); ++k) 
	{
	  P = at * Pixel(j,  k);
	  for (i = 0; i < newlands.dim(); ++i)
	    newlands[i](j, k) = (STAGS[i].get(P[0], P[1]) != 0);
	}
    }
  else 
    {
    Pixel P;
    for (j = 0; j < newlands[0].dimx(); ++j)
      for (k = 0; k < newlands[0].dimy(); ++k) 
	{
        P = at * Pixel(j,  k);
        for (i = 0; i < newlands.dim(); ++i)
          newlands[i](j, k) = (STAGS[2 * i].get(P[0], P[1]) != 0 ||
                               STAGS[2 * i + 1].get(P[0], P[1]) != 0);
	}
    
      }
  if (dum1==2 || dum1==3)
    {
      
      void view_normalized(const ucgrid& gr, int gfork=0, int min=0, int max=0);
      for (i=0; i<newlands.dim(); i++)
	{
	  cout << " landmark " << i << endl;
	  if (dum1==3)
	      view_normalized(STAGS[i]);
	  view_normalized(newlands[i]);
	}
    }
} // register_landmarks



// landmark.c
