// Copyright YALI AMIT, January 15, 2001.

#define LTATREE_C



// __________________________________________________________________________
// Include files

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>
#include "simpstrg.h"
#include "includes.h"
#include "pixel.h"
#include "viewport.h"
#include "landmark.h"
#include "trainstat.h"
#include "ltatree.h"
#include "globals.h"
extern ucgrids STAGS;
void init_STAGS(int size, int dimx, int dimy);
void fill_STAGS(const landmarks& lm, int spread = 0, int numt = -1);
void empty_STAGS(const landmarks& lm);
extern ltatrees ltrees;
extern dgrid weits;


// __________________________________________________________________________
// Typedefs, etc.

static impfunction impfn = &histogram::entropy;
void make_vertlist(igrid& newverts, igrid& oldverts, question& que);
void get_best_question(const ltaiter& diter, const lmarkslist& lmarks, 
		       lgrid& quedata, const igrids& instances);
int get_answer(question& que, const landmarks& lmarks, 
	       const igrid& oldinst, igrid& newinst, int ii=-1);
void get_best_q0(question& que, lgrid& qlist, lgrid& dlist, 
		 const lmarkslist& lmarks, double& ent); 
int check_q0(relation& rel, const landmarks& lmarks, int ii=-1);
int check_q5(relation& rel, const landmarks& lmarks,igrid& instances, int upd);
int check_q6(relation& rel, const landmarks& lmarks, const igrid& instances, igrid& newinst, int upd);
void get_best_q1(question& que, lgrid& qlist, lgrid& dlist, 
		 const lmarkslist& lmarks, double& ent); 
void get_best_q2(question& que, lgrid& qlist, lgrid& dlist, 
		 const lmarkslist& lmarks, igrid& vertlist, 
		 const igrids& instances, double& ent);
void get_best_q5(question& que, lgrid& qlist, lgrid& dlist, 
		 const lmarkslist& lmarks, double& ent); 
void get_best_q6(question& que, lgrid& qlist, lgrid& dlist, 
		 const lmarkslist& lmarks, igrid& vertlist, 
		 const igrids& instances, double& ent);
int upd_inst1(relation& rel, const landmarks& lmark, igrid& instances);
int upd_inst2(relation& rel, const landmarks& lmarks, 
	      const igrid& oldinst, igrid& newinst);


// __________________________________________________________________________
// Grow an ltatree - this is used both for local lta's and for classification
//                   trees based on absolute or relational arrangements.
void ltatree::grow(const lgrid& datalist, const lmarkslist& lmarks)
{

  anginc = 2 * PI / numang;
  reltolerance = anginc;
  // Initialize the tree
  maketop();

  // Fill in the histogram information for the top node
  ltahist& dhist = top()->hist;
  dhist.clear();
  int i;
  for (i = 0; i < numtreedata; ++i)
      ++dhist[lmarks[i].truth(nist)];
  //   top()->datalist.init(numtreedata);
  //   for (i = 0; i < numtreedata; ++i)
  //     top()->datalist(i) = i;
  top()->datalist.init(datalist.dimx());
  for (i=0; i<datalist.dimx();++i) {
    top()->datalist(i)=datalist(i);
  }

  // Create an array for storing instances
  igrids instances(numtreedata);

  // Deepen the tree
  deepen(top(), lmarks, instances);

} // ltatree::grow


void ltatree::deepen(const ltaiter& diter, const lmarkslist& lmarks,
		       const igrids& instances)
{

  // Set up an alias for the node data and find out how much data there is
  int numnodedata = diter->datalist.dimx();
 
  int i;
  // What we do depens on whether or not we're at a leaf node or not
  if (diter.leaf())
  {
    // Test the stopping criterion for this node
    if (prestop(diter, mindata))
    {
     if (diter==diter.parent().yes() && !class_tree)
      cout << "final number " <<  diter->hist.total() << endl;

      return;
    }
    
    lgrid loclist;

    if (!class_tree)
      numloc=numnodedata;
    // Get the random subset of data points used for estimating probabilities in this node.
    void get_dig_list(lgrid &loclist,int numloc, lgrid &datalist);
    get_dig_list(loclist, numloc, diter->datalist);
    
    
    igrids locinst(loclist.dimx());
    
    for (int i=0; i<loclist.dimx(); i++)
      {
	if (class_tree!=2 && class_tree!=3)	
	  locinst(i)=instances(loclist(i));
	loclist(i)=diter->datalist(loclist(i));
	
      }
    

    // Get the best question for this node from a random subset.
    get_best_question(diter, lmarks, loclist, locinst);
    // Return now if no good question was found
    if (diter->que.typ == 0)
      return;

    // Prepare to ask the new question
    diter.makechildren();
    diter.no()->vertlist = diter->vertlist;
    make_vertlist(diter.yes()->vertlist, diter->vertlist, diter->que);

  }
  else
  {
    // Clear the histograms and facelists of the children
    diter.yes()->datalist.init();
    diter.yes()->hist.clear();
    diter.no()->datalist.init();
    diter.no()->hist.clear();
  }

  // Ask the node question of all the data at the current node
  int numyes = 0;
  igrids newinst(numnodedata);
  cgrid answers(numnodedata);
  for (i = 0; i < numnodedata; ++i)
  {
    if (class_tree == 0) // local lta tree - only yes nodes - 
      fill_STAGS(lmarks[diter->datalist(i)]); // STAGS used to make computation efficient
    answers(i) = get_answer(diter->que, lmarks[diter->datalist(i)], 
			    instances[i], newinst[i], diter->datalist(i));

    numyes += (answers(i) > 0);
    if (class_tree == 0)
      empty_STAGS(lmarks[diter->datalist(i)]);
  }


  // Set up the children
  igrids yesinstances(numyes), noinstances(numnodedata - numyes);
  diter.no()->datalist.init(numnodedata - numyes);
  diter.yes()->datalist.init(numyes);

  int noind = 0, yesind = 0;
  // Update which data points go to which child node.
  for (i = 0; i < numnodedata; ++i)
  {
    if (answers(i))
    {
      diter.yes()->datalist(yesind) = diter->datalist(i);
      if (!boost)
	++diter.yes()->hist[lmarks[diter->datalist(i)].truth(nist)];
      else
	diter.yes()->hist[lmarks[diter->datalist(i)].truth(nist)]+=
	  weits(diter->datalist(i));
      if (class_tree!=2 && class_tree!=3) // Absolute arrangement tree, 
	                                  //no instances to keep track of.
      yesinstances[yesind] << newinst[i];

      ++yesind;
    }
    else
      {
	diter.no()->datalist(noind) = diter->datalist(i);
	if (!boost)
	  ++diter.no()->hist[lmarks[diter->datalist(i)].truth(nist)];
	else
	  diter.no()->hist[lmarks[diter->datalist(i)].truth(nist)]+=
	    weits(diter->datalist(i));
       if (class_tree!=2)
	  noinstances[noind] << newinst[i];
	++noind;
      }
  }

  assert((yesind == numyes) && (yesind + noind == numnodedata));
  // Recursively proceed to deepen
  deepen(diter.yes(), lmarks, yesinstances);
  deepen(diter.no(), lmarks, noinstances);
 

}  // ltatree::deepen

// Find the best question of appropriate type at this node.
void get_best_question(const ltaiter& diter, const lmarkslist& lmarks, 
		       lgrid& quedata, const igrids& instances)
{

  lgrid maxq(20), numq(20), doneques(20), qlist;
  maxq.clear(); numq.clear(); doneques.clear();
  int numverts = diter->vertlist.dimx();
  // Maximum number of possible questions of the different types
  maxq(0) = cartsizex * cartsizex * numland;
  maxq(1) = numang * numdis * numtol * numland * numland;
  maxq(5) = numang * numdis * numtol * numland * numland;
  maxq(2) = numang * numdis * numtol * numland * numverts;
  maxq(6) = numang * numdis * numtol * numland * numverts;
  maxq(3) = cartsizex * cartsizey * numland * numland;
  maxq(4) = numland * cartsizex * cartsizey;

  
  double bestent = 1000000;
  if (class_tree)
    {
      if (boost==0)
	bestent = diter->hist.entropy();
      else
	{
	  bestent=0;
	  dgrid temp(numclass);
	  temp.clear();
	  int numnodedata = quedata.dimx();
	  for (int i = 0; i < numnodedata; ++i)
	    temp(lmarks(quedata(i)).truth())+=weits(quedata(i));
	  double tot=0;
	  for (int i=0; i<temp.dimx(); i++)
	    tot+=temp(i);
	  for (int i=0; i<temp.dimx(); i++)
	    if (temp(i)>0)
	      bestent -= log(temp(i)/tot)*temp(i)/tot;
	}
      //      cout << " BESTENT " << bestent << endl;
      if (bestent==0)
	return;
    }   
  // Absolute arrangements question `0'
  if (class_tree==2)
    {
      numq(0)=numquesask;
      random_select(qlist, maxq(0), numq(0)); 
      get_best_q0(diter->que, qlist, quedata, lmarks, bestent);
    }

  // Relative arrangements  1 are 5 are the same thing - two feature arrangements,
  //  asked if no pending arrangement exists - nodes with only no's in history.
  // Computed differently for efficiency consideratrions. In classification trees
  // use q3, for ltas use q1
  else
    {
      if (numverts == 0)
	{
	  if (useques[1])
	    {
	      numq(1) =  numquesask;
	      random_select(qlist, maxq(1), numq(1));
	      get_best_q1(diter->que, qlist, quedata, lmarks, bestent);
	    }
	  if (useques[5])
	    {
	      numq(5) =  numquesask;
	      random_select(qlist, maxq(5), numq(5));
	      get_best_q5(diter->que, qlist, quedata, lmarks, bestent);
	    }
	}
      else
	// If a pending arrangement exists - add to it. 2 and 6 are the same
	// computed differently.
	{
	  if (useques[2])
	    {
              numq(2)=numquesask;
	      random_select(qlist, maxq(2), numq(2));
	      get_best_q2(diter->que, qlist, quedata, lmarks, 
			  diter->vertlist, instances, bestent);
	    }
	  if (useques[6])
	    {
	      numq(6)=numquesask;
	      random_select(qlist, maxq(6), numq(6));
	      get_best_q6(diter->que, qlist, quedata, lmarks, 
			  diter->vertlist, instances, bestent);
	    }
	}
    }
      
} // get_best_question

// Get the answer to a question 
int get_answer(question& que, const landmarks& lmarks, 
	       const igrid& oldinst, igrid& newinst, int ii/*=-1*/)
{
  int answer=-1;
  switch (que.typ)
  {
  case 10:
    answer=check_q0(que.r1, lmarks, ii);
   break;
  case 1:
    answer = upd_inst1(que.r1, lmarks, newinst);
    break;
  case 5:
    answer = check_q5(que.r1, lmarks, newinst, 1);
    break;
  case 2:
    answer = upd_inst2(que.r1, lmarks, oldinst, newinst);
    break;
  case 6:
    answer = check_q6(que.r1, lmarks, oldinst, newinst, 1);
    break;
  default:
    cerr << "Unknown question type [" << que.typ << "] in "
	 << "get_answer routine\n";
    exit(1);
    break;
  }

  return answer;

} // get_answer

void ltatree::updatehists(const ltaiter& curr)
{
  if (curr.branch())
  {
    updatehists(curr.no());
    updatehists(curr.yes());
    for (int i = 0; i < curr->hist.dim(); ++i)
      {
      curr->hist[i] = curr.no()->hist[i] + curr.yes()->hist[i];
      //      cout << curr->hist[i] << " ";
      }
    //cout << endl;

  }

} // ltatree::updatehists


// For testing - drop a data point down a tree 
void ltatree::dropdata(const landmarks& lmarks, ltaiter& terminal, 
		       igrid& newinst, int i/*=-1*/) const
{

  anginc = 2 * PI / tnumang;
  reltolerance = anginc;
  terminal = top();
  igrid tmpinst;
  int currdepth = 0;
  while (terminal.branch() && (ltausedepth == -1 || ltausedepth > currdepth++))
  {
    if (get_answer(terminal->que, lmarks, newinst, tmpinst, i))
      terminal = terminal.yes();
    else 
      terminal = terminal.no();
    newinst << tmpinst;
    //    newinst.dump();
  }

} // ltatree::dropdata


// __________________________________________________________________________
// Update vertex lists of pending arrangements, what landmark types 
// with what angle.
void make_vertlist(igrid& newverts, igrid& oldverts, question& que)
{
  int i;
  switch(que.typ)
  {
  case 1:
  case 5:
    newverts.init(oldverts.dimx() + 3);
    for (i = 0; i < oldverts.dimx(); ++i)
      newverts(i) = oldverts(i);
    newverts(oldverts.dimx()) = que.r1.land1;
    newverts(oldverts.dimx() + 1) = que.r1.land2;
    newverts(oldverts.dimx() + 2) = que.r1.ang;
    break;
    
  case 2:
  case 6:
    newverts.init(oldverts.dimx() + 2);
    for (i = 0; i < oldverts.dimx(); ++i)
      newverts(i) = oldverts(i);
    newverts(oldverts.dimx()) = que.r1.land1;
    newverts(oldverts.dimx()+1) = que.r1.ang;
    break;
    

   default:
    break;
  }

} // make_vertlist

// __________________________________________________________________________

void ltatree::update()
{
  for (int i = first; i < first + numdata; ++i)
  {
    landmarks lmarks;
    igrid insts;
    lmarks.read(i);
    ltaiter terminal;
    dropdata(lmarks, terminal, insts);
    ++terminal->hist[lmarks.truth()];
  }

} // ltatree::update

// __________________________________________________________________________

void ltatree::clearhists(const ltaiter& curr)
{
  if (curr.branch())
  {
    clearhists(curr.yes());
    clearhists(curr.no());
  }
  curr->hist.clear();

} // ltatree::clearhists

// __________________________________________________________________________


// Return 1 if we can not continue growing the tree from node 't'.

int prestop(const ltaiter& t, int)
{

  // Stop if tree depth is too large
  if (t.depth() >= maxdepth)
    return 1;

  // Stop if a no node.
  if (!class_tree && (t.depth() > 0) && (t == t.parent().no()))
    return 1;

  // Stop if not enough data to continue
   if (!class_tree && t->hist.total() < numdata * objprop)
    return 1;

  // Stop if the node is pure
  if (class_tree && t->hist.pure(purepar))
     return 1;

  // Stop if there is not enough data to make a split
  if (class_tree && t->hist[t->hist.mode()] < mindata)
      return 1;

  // Stop if the second mode doesn't have a reasonable amount of data
  if (class_tree && t->hist[t->hist.mode2()] < modetwo)
     return 1;



  return 0;

} // prestop

// Different impurity functions for classification trees.
void set_impurity_func(const char* imeasure, int exitflag)
{

  
  if (strcmp(imeasure, "entropy") == 0)
    impfn = &histogram::entropy;
  else if (strcmp(imeasure, "gini") == 0)
    impfn = &histogram::gini;
  else if (strcmp(imeasure, "kolsmir") == 0)
    impfn = &histogram::kolsmir;
  else if (strcmp(imeasure, "unsup") == 0)
    impfn = &histogram::unsup;
  else if (exitflag) {
    cerr << "Invalid impurity measure " << imeasure << " specified.\n"
      "Exiting.\n";
    exit(1);
  }
  else
    impfn = &histogram::entropy;

} // set_impurity_func
// EStimate average entropy or impurity of a given split which
// produces two histograms h1,h2
double ave_impurity(histogram& h1, histogram& h2)
{
  double t1 = h1.total();
  double t2 = h2.total();
  double t = t1 + t2;
  return ((t == 0.0) ? 0.0 : (t1 * (h1.*impfn)() + t2 * (h2.*impfn)()) / t);

} // ave_impurity

void get_dig_list(lgrid &loclist, int numloc, lgrid &datalist)
{
  int numnodedata=datalist.dimx();
  if (numloc>=numnodedata)
    {
      loclist.init(numnodedata);
      for (int i=0; i<numnodedata; i++)
	{
	  loclist(i)=i;
	}
      return;
    }


  ucgrid tt(datalist.dimx());
  tt.clear();
  double tot=0;
  for (int i=0; i<datalist.dimx(); i++)
    tot+=weits(datalist(i));
  double fac=datalist.dimx()/double(tot);
  double pcnt = double(numloc)/double(datalist.dimx());
  int cc=0;
  for (int j=0; j<datalist.dimx(); j++)
    {   
      float x = drand48();
      if (x<=pcnt*weits(datalist(j))*fac)
	{
	  tt(j)=1;
	  cc++;
	}

    }
  loclist.init(cc);
  cc=0;
  for (int j=0; j<datalist.dimx(); j++)
    {
      if (tt(j))
	{
	  loclist(cc)=j;
	  cc++;
	}  
    }

}

//void get_dig_list(lgrid &loclist, int numloc, lgrid &datalist)
// {
//   int numnodedata=datalist.dimx();
//   if (numloc>=numnodedata)
//     {
//       loclist.init(numnodedata);
//       for (int i=0; i<numnodedata; i++)
//         {
//           loclist(i)=i;
//         }
//       return;
//     }
//   ucgrid selections(numnodedata);
//   dgrid distrib(numnodedata);
//   dgrid wtVec(numnodedata);

//   // Initialize the selections to zero
//   for (int i=0; i<numnodedata; i++) {
//     selections[i]=0;
//   }

//   // Initialize wtVec to contain the weights of the data at the node
//   for (int i=0; i<numnodedata; i++) {
//     wtVec(i)=weits(datalist(i));
//   }

//   // Assign values to the distrib vector
//   distrib(0)=wtVec(0);
//   for (int i=1; i<numnodedata; i++) {
//     distrib(i)=wtVec(i)+distrib(i-1);
//   }

//   // Normalize the distrib vector
//   for (int i=0; i<numnodedata; i++) {
//     distrib(i)/=distrib(numnodedata-1);
//   }

//   // Draw the sample
//   for (int i=0; i<numloc; i++)
//     {
//       int sel=0;
//       int done=0;
//       while (!done)
//         {
//           double x = drand48();
// 	  for (int j=0; j<numnodedata; j++)
//             {
//               if (x<=distrib(j))
//                 {
//                   if (selections(j)==0)
//                     {
//                       selections(j)=1;
// 		      sel=j;
//                       done=1;
//                     }
//                   break;
//                 }
//             }
//         }
//       // Once the a point has been chosen, set its weight to zero
//       wtVec(sel)=0;
//       // Recompute the distribution and renormalize
//       distrib(0)=wtVec(0);
//       for (int j=1; j<numnodedata; j++) {
// 	distrib(j)=wtVec(j)+distrib(j-1);
//       }
//       for (int j=0; j<numnodedata; j++) {
// 	distrib(j)/=distrib(numnodedata-1);
//       }
//     }
  
//   // Assign the selected digits to loclist
//   loclist.init(numloc);
//   int j=0;
//   for (int i=0; i<numnodedata; i++) {
//     if (selections(i)==1) {
//       loclist(j)=i;
//       j++;
//     }
//      if (j>=numloc)
//        break;
//   }
// }

// ltatree.C









