// Copyright YALI AMIT, January 15, 2001.

#ifndef LTATREE_H
#define LTATREE_H



#include <iostream>
#include "grid.h"
#include "question.h"
#include "histgram.h"
#include "landmark.h"
#include "tree.h"

extern int mindata;
extern int ltahistsize;
// __________________________________________________________________________

class ltahist : public histogram
{
 private:

 public:
  ltahist() : histogram(ltahistsize) { }

}; // class ltahist

// __________________________________________________________________________

class ltadata
{ 
 private:

 public:
  question que;
  igrid vertlist;
  lgrid datalist;
  ltahist hist;

  ltadata() { }
};

class ltatree;

typedef treeiterator<ltadata> ltaiter;

const int maxtags = 300;

class ltatreestats
{
 private:
  float total;
  float correct;
  int wtdepth;
  double wtent;
  igrid queuse;
  igrid taguse;
  igrid anguse;
  igrid toluse;
  igrid disuse;

 public:
  ltatreestats() : total(0), correct(0), wtdepth(0), wtent(0), queuse(20), 
  taguse(maxtags), anguse(40), toluse(5), disuse(3) { 
    queuse.clear(); taguse.clear(); anguse.clear(); 
    toluse.clear(); disuse.clear(); 
  }
  void operator+=(const ltatreestats& d) { 
    total += d.total; correct += d.correct; wtdepth += d.wtdepth; 
    wtent += d.wtent; queuse += d.queuse; taguse += d.taguse; 
    anguse += d.anguse; toluse += d.toluse; disuse += d.disuse; 
  }
  void dump(int n) const;

  friend class ltatree;
};

class ltatree : public tree<ltadata>
{
 private:

 public:
  ltatreestats *stats;
  int locationx, locationy;
  int tnumang, tnear;
  int maxyesdepth() { int md = 0; 
		      for (ltaiter t = top(); t.yes();
			   t = t.yes()) { ++md; } return md; }
  float maxyescount() { ltaiter t; for (t = top(); t.yes(); t = t.yes()); 
		      return (t ? t->hist.total() : 0); }

  ltatree() : stats(0) { }
  ~ltatree() { delete stats; }

  void calc_stats(const ltaiter& diter);

  void dump(ostream& os = cout);
  void dumpnodes(const ltaiter& t, ostream& os);
  void dumpnode(const ltaiter& t, ostream& os);
  void read(ifstream& ifs);
  void write(ofstream& ofs);
  void write(const char * treename);
  void writenodes(const ltaiter& t, ofstream& ofs);

  int read(const char * treename);
  void readnodes(const ltaiter& t, ifstream& ifs);
  void updatehists(const ltaiter& t);

  void deepen(const ltaiter& diter, const lmarkslist& lmarks, 
	      const igrids& instances);
  void dropdata(const landmarks& lmarks, ltaiter& diter, 
		igrid& newinst, int i=-1) const;
  void grow(const lgrid& datalist, const lmarkslist& lmarks);
  void update();
  void calcsummary(const ltaiter& t);
  void dumpsummary(ostream& os);
void dropdata_fromlist(landmarks& lmarks, ltaiter& terminal, 
			 igrid& newinst, igrids& preinsts);
  void drop(dgrid& traingrid, dgrid& targetgrid,
	    const ltaiter& t, int numdates, int& yescount, int& nocount);
  void clearhists(const ltaiter& curr);
};



// __________________________________________________________________________
// Typedefs, Function prototypes and inline functions

extern int prestop(const ltaiter& qiter, int mind = mindata);
extern int poststop(const ltaiter& qiter, double bestent);
void find_good_trees();
void choose_ltas(const igrid* temp);
typedef vector<ltatree> ltatrees;
// __________________________________________________________________________
// ltatree.h
#endif
