// Copyright YALI AMIT, January 15, 2001.
// __________________________________________________________________________
// Include files

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>

#include "includes.h"
#include "simpstrg.h"
#include "pixel.h"
#include "viewport.h"
#include "trainstat.h"
#include "ltatree.h"
#include "globals.h"

extern ltatrees ltrees;
void find_maxmin(lmarkslist &lmarks,int &globalminx,int &globalmaxx,int &globalminy,int &globalmaxy);

// __________________________________________________________________________
// option 46

// Grow lta trees based on registered edge data using
// the relational tree structure. Highest frequency in
// each box picked at each depth. That is the selection
// criterion at each node. 
void make_ltas()
{
  // Determine the data which will be used to build the lta trees.
  // For lta's, typically all the data is used.
  lgrid datalist;
  switch (initdataopt)
  {
   case 0:
    random_select(datalist, numdata, numtreedata);
    break;
  case 1:
     break;
  }

  // Read in the lists of landmarks
  int i, j;
  lmarkslist lmarks(numtreedata);
  for (i = 0; i < numtreedata; ++i)
    lmarks[i].read(datalist(i));

  numland=lmarks[0].total_types();
  anginc = 2 * PI / numang;
  reltolerance = anginc;
  // Determine landmark extents
  int globalminx, globalminy, globalmaxx, globalmaxy;
  find_maxmin(lmarks, globalminx, globalmaxx, globalminy, globalmaxy);
  
  cout << "[minx, maxx] [miny, maxy] = [" << globalminx << ", " 
    << globalmaxx << "] [" << globalminy << ", " << globalmaxy << "]\n";

  int lastregx = (globalmaxx-globalminx-1)/regionsize - firstreg ;
  int lastregy = (globalmaxy-globalminy-1)/regionsize - firstreg ;



  regionminx = globalminx + regionsize * firstreg;
  regionmaxx = regionminx + regionsize - 1;
  int currind = 0;
  //Loop over disjoint boxes.
  for (i = firstreg; regionminx < globalmaxx && i < lastregx; ++i)
  {
    // These are global variables constraining the location
    // of the points in the relational arrangement computed in q1.C q2.C
    regionminy = globalminy + regionsize * firstreg;
    regionmaxy = regionminy + regionsize - 1;
    
    for (j = firstreg; regionminy < globalmaxy && j < lastregy; ++j)
    {
      if (++currind >= first)
      {
	ltatree ltree;
	ltree.tnumang = numang;
	ltree.tnear = near;
	simpstrg currtree = simpstrg("tree") + ltos(i) + "." + ltos(j);
	ltree.locationx = regionminx + regionsize / 2;
	ltree.locationy = regionminy + regionsize / 2;
	cout << "Building tree " << currtree << endl;
	// Grow the relational tree.
	ltree.grow(datalist, lmarks);
	cout << ltree.locationx << " " << ltree.locationy << endl;
	cout << "Tree depth " << ltree.maxyesdepth() << endl;
	cout << "Frequencies: " << ltree.maxyescount() << endl;
	// Write the tree.
	if (ltree.maxyesdepth() >= mindepth)
	  ltree.write(currtree);
      }
      regionminy += regionsize;
      regionmaxy += regionsize;
    }
    regionminx += regionsize;
    regionmaxx += regionsize;
  }

} // make_ltas



// __________________________________________________________________________
// Option 85
// Pick a good collection of lta trees to use.  They are chosen so that
// the average distance to a nearest neighbor is maximized.  Many
// parameters are currently hardwired here.

void find_good_trees()
{
void find_maxmin(lmarkslist &lmarks,int &globalminx,int &globalmaxx,int &globalminy,int &globalmaxy);
  int i;
 lgrid datalist;
  switch (initdataopt)
  {
   case 0:
    random_select(datalist, numdata, numtreedata);
    break;
  case 1:
     break;
  }

  lmarkslist lmarks(numtreedata);


  for (i = 0; i < numtreedata; ++i)
    lmarks[i].read(datalist(i));
  int firstregx, firstregy, lastregx, lastregy;

  // Determine landmark extents
  int globalminx, globalminy, globalmaxx, globalmaxy;
  find_maxmin(lmarks, globalminx, globalmaxx, globalminy, globalmaxy);
  
  firstregx=firstregy=firstreg;
  lastregx=lastregy=lastreg;
  

  cout << firstregx << " " << lastregx << endl;
  cout << firstregy << " " << lastregy << endl;

  typedef ltatree* ltatreeptr;
  int numgood = 0;
  ltatreeptr good[400];
  int goodx[400], goody[400];
  ltatree ltreess[lastregx - firstregx][lastregy - firstregy];
  // Loop through all arrangements coded in ltatrees in ltatreedir
  // Accumulate those with high enough counts.
  // CHoose a random subset with locations most spread out.
  for (i = firstregx; i < lastregx; ++i)
  {
    for (int j = firstregy; j < lastregy; ++j)
    {
      simpstrg ltaname = ltatreedir + simpstrg("/tree") + 
	ltos(i) + "." + ltos(j);

      cout << ltaname << endl;
      if (ltreess[i-firstregx][j-firstregy].read(ltaname))
      {
	cout << " Read " << endl;
        cout << i << " " << j << " " << 
	  ltreess[i-firstregx][j-firstregy].maxyescount() << endl;
	cout << i << " " << j << " " << 
	  ltreess[i-firstregx][j-firstregy].maxyesdepth() << endl;

      
      if (ltreess[i-firstregx][j-firstregy].maxyesdepth() >= mindepth &&
	  ltreess[i-firstregx][j-firstregy].maxyescount() > numdata * objprop)
      {
	goodx[numgood] = i;
	goody[numgood] = j;
	good[numgood++] = &(ltreess[i-firstregx][j-firstregy]);
        cout << numgood << endl;
      }
      } 
    }
  }
  double bestmin = 0;
  double bestmin1 = 0;
  lgrid bestl;
  cout << " numgood " << numgood << endl;
  if (numltas >= numgood)
     {
       numltas=numgood;
       cout << "Only " << numgood << "ltas " << endl;
       bestl.init(numgood);
       for (int i=0; i<numgood; i++)
           bestl(i)=i;
     }
  else
    {
      // Try a large number of size numltas random subsets
      // and keep the one that is most spread out.
      for (i = 0; i < 50000; ++i)
	{
	  double min = 0;
	  double min1 = 1000000;
	  lgrid l;
	  
	  random_select(l, numgood, numltas);
	  
	  for (int j = 0; j < numltas; ++j)
	    {
	      double currmin = 1000000;
	      for (int k = 0; k < numltas; ++k)
		{
		  if (k == j)
		    continue;
		  double currdist = 
		    linfdist(good[l(j)]->locationx, good[l(j)]->locationy,
			     good[l(k)]->locationx, good[l(k)]->locationy);
		  if (currdist < currmin)
		    currmin = currdist;
		}
	      if (currmin < min1)
		min1 = currmin;
	      min += currmin;
	    }
	  if ((min1 == bestmin1 && min > bestmin) || (min1 > bestmin1))
	    {
	      bestmin1 = min1;
	      bestmin = min;
	      bestl = l;
	    }
	}
    }
  //
  // Write out the chosen trees to file `ltatrees'
  cout << "NUMLTAS " << numltas << endl; 
  ofstream ofs("ltatrees");
  varwrite(ofs,numltas);
  // write out the best lta trees
  for (i = 0; i < numltas; ++i)
  {
    cout << "tree" << goodx[bestl(i)] << "." << goody[bestl(i)] << endl;
    //    simpstrg ltaname = simpstrg("tree") + ltos(i);
    (*good[bestl(i)]).write(ofs);
  }

} // find_good_trees

// __________________________________________________________________________
// Choose ltatrees from precomputed frequency maps obtained
// for a predetermined collection of edge arrangements such as
// all two edge arrangements.
void choose_ltas(const igrid* temp, int globalminx, int globalminy, int globalmaxx, int globalmaxy, int incrx, int incry)
{
  // Determine the data which will be used to build the lta trees.
  // For lta's, typically all the data is used.
  lgrid datalist;

  // Determine landmark extents

  int numtr=0;
  ivector list(numltas);
  for (int u=0; u<numltas; u++) list[u]=0;

  int lastregx = (globalmaxx-globalminx-1)/regionsize - firstreg ;
  int lastregy = (globalmaxy-globalminy-1)/regionsize - firstreg ;
  cout << "lastregx " << lastregx << endl;

  int nn;
  ofstream ofs("ltatrees");
  varwrite(ofs,nn);
  for (int i = firstreg; i < lastregx ; ++i)
  {
    regionminx = globalminx + i * regionsize;
    regionmaxx = regionminx + regionsize;
    for (int j = firstreg; j < lastregy ; ++j)
    {
    regionminy = globalminy + j * regionsize;
    regionmaxy = regionminy + regionsize;

      {
        int found=0;

          for (int ll=0; ll<numltas && !found; ll++)
          {
	    // poolnumuse - how many times a particular feature type
	    // can be used in a model.
            if (list[ll]<=poolnumuse)
            {
              int count=-1;
              int newk=-1, newl=-1;
              for (int k=regionminx; k<regionmaxx; k++)
                for (int l=regionminy; l<regionmaxy; l++)
                {
                  if (temp[ll].get(k,l) > count)
                  {
                    count = temp[ll].get(k,l);
                    newk = k;
                    newl = l;
                  }
                }
	      // Check frequency.
              if ( (objprop>0 && count > objprop*numdata) || 
                   (objprop<0 && count < -objprop*numdata)  )
              {
                cout << " found " << ll << " " << count << 
                  " " << i<<"."<<j <<endl;
                list[ll] += 1;
                found=1;
                simpstrg currtree = simpstrg("tree") + 
                  ltos(i) + "." + ltos(j);
                ltrees[ll].locationx = newk + incrx;
                ltrees[ll].locationy = newl + incry;
                cout << ltrees[ll].locationx << " " << 
                  ltrees[ll].locationy << endl;
                ltrees[ll].write(ofs);
		numtr++;
              }
//              list[ll]=1;
            }
	  }
      }
      regionminy += regionsize;
      regionmaxy += regionsize;
    }
    regionminx += regionsize;
    regionmaxx += regionsize;
  }
   ofs.seekp(0);
   varwrite(ofs,numtr);
   ofs.close();
   cout << "NUMEDGES " << numtr << endl;
  
} // choose_ltas

static int compar_c(const void *a, const void* b)
{
  return (*((const int*) a) < *((const int*) b));
}

// Make edge arrangements directly from edge frequency maps
// without verifying joint statistics.
void choose_ltas_from_edges(igrids &temp)
{
  // Determine the data which will be used to build the lta trees.
  // For lta's, typically all the data is used.
  lgrid datalist;



  double anginc = 2 * PI / numang;

  int dimx=temp[0].dimx();
  int dimy=temp[0].dimy();
  int lastregx = dimx/regionsize - firstreg ;
  int lastregy = dimy/regionsize - firstreg ;
  cout << firstreg << " " << lastregx << " " << lastregy <<endl;
  int tn=0;
  for (int i = firstreg; i < lastregx ; ++i)
    {
      regionminx = i * regionsize;
      regionmaxx = regionminx + regionsize;
      for (int j = firstreg; j < lastregy ; ++j)
	{
	  regionminy = j * regionsize;
	  regionmaxy = regionminy + regionsize;
	  // Find most common edge in region:
	  int count=0;
	  int olda=-1,oldb=-1,oldc=-1;
	  for (int c=0; c<8; c++)
	    for (int a=regionminx; a<regionmaxx; a++)
	      for (int b=regionminy; b<regionmaxy; b++)
		{
		  if(temp[c].get(a,b) > count)
		    {
		      count=temp[c].get(a,b);
		      olda=a;
		      oldb=b;
		      oldc=c;
		    }
		}
	  cout << i << " " << j << " count " << count << endl;
	  // If sufficiently frequent
	  if (count > objprop*numdata)
	    {
		ltatree ltree;
		ltree.locationx=olda;
		ltree.locationy=oldb;
		ltree.tnumang = numang;
		ltree.tnear = near;
		ltree.maketop();
		int tlen=4*near*near*8;
		igrid TabEdges(4, tlen);
		
		TabEdges.clear();
		int p=0;
		for (int c=0; c<8; c++)
		  {
		  for (int a=olda-near+1; a<olda+near; a++)
		    for (int b=oldb-near+1; b<oldb+near; b++)
		      {
			if ( fabs(a-olda)>=2 || fabs(b-oldb)>=2)
			  {
			    TabEdges(0,p)=temp[c].get(a,b);
			    TabEdges(1,p)=a;
			    TabEdges(2,p)=b;
			    TabEdges(3,p)=c;
			    p++;
			  }
		      }
		  }
		// Sorted edges in neighborhood according
		// to frequency and add them to arrangement
		// in decreasing order, up to maxdepth.
		// The location of the edge in the neighborhood
		// determines the relative angle for the relational
		// tree.
		TabEdges.sort_columns(tlen,compar_c); 

		relation rel;
		int ind=0;
		cout << "count " << oldc << " " << TabEdges(0,0) << endl;
		double dx=TabEdges(1,ind)-olda;
		double dy=TabEdges(2,ind)-oldb;

		double ang = atan2(-dy,dx);
		if ( ang < 0 )
		  ang = 2*PI + atan2(-dy,dx);
		rel.ang = (uchar)((rint)(ang / anginc)); 
		rel.tol = mintol;
		rel.dis = mindis;
		rel.land1 = oldc;
		rel.land2= TabEdges(3,ind);

		ltree.top()->que.typ=1;
		ltree.top()->que.r1 = rel;
		(ltree.top()).makechildren();
		ltaiter diter=ltree.top();
		int highcount=1;
		for (int d=1; d<maxdepth && highcount; d++)
		  {
		    cout << "count " << TabEdges(0,d) << " depth " << d << endl;
		    if (TabEdges(0,d) > objprop*numdata)
			{
			  diter=diter.yes();
			  dx=TabEdges(1,d)-olda;
			  dy=TabEdges(2,d)-oldb;
			  ang = atan2(-dy,dx);
			  if ( ang < 0 )
			    ang = 2*PI + atan2(-dy,dx);
			  rel.ang = (uchar)((rint)(ang / anginc)); 
			  rel.tol = mintol;
			  rel.dis = mindis;
			  rel.land1 = TabEdges(3,d);
			  rel.land2 = -1;
			  rel.v1=0;
			  diter->que.typ=2;
			  diter->que.r1 = rel;
			  diter.makechildren();
			}
		    else
		      highcount=0;
		    }
		if (highcount)
		  {
		    simpstrg currtree = simpstrg("tree")+ltos(tn);
		    tn++;
		    cout << "Tree Locations " << ltree.locationx << " " << 
		      ltree.locationy << endl;
		    ltree.write(currtree);
		  }
		
	      }
	    
	}
      
    }
  
} // choose_ltas



// ________________________________________________
