// Copyright YALI AMIT, January 59, 2001.

#ifndef FACES_C
#define FACES_C

// __________________________________________________________________________



// __________________________________________________________________________

#include <iostream>
#include <cstdio>

#include "includes.h"
#include "landmark.h"

#include "simpstrg.h"
#include "pixel.h"
#include "trainstat.h"
//#include "facetest.h"
#include "ltatree.h"
#include "carttree.h"
#include "viewgrid.h"

#include "globals.h"

extern double ratios[];
extern ltatrees ltrees;
extern carttrees ctrees, ftrees;
int statpts[] = {34, 37, 42, 45, 82, 85, 90, 93, 162, 165, 170, 173,
		 210, 213, 218, 221};

extern ucgrids STAGS;
void init_STAGS(int size, int dimx, int dimy);
void fill_STAGS(const landmarks& lm, int spread = 0, int numt = -1);
void empty_STAGS(const landmarks& lm);
void extract_landmarks(image& im, ltatrees& ltrees, int numltas,
			 landmarks& edgelands, landmarks& ltalands);

void affine_trans_lm(landmarks &l, affinetrans &pqtrans, lmarklist &l2, int &numlands);
int read_ltatrees(ltatrees& ltrees, simpstrg direc);
void register_to_dgrid(image& im, dgrid& regim, affinetrans at);
void find_maxmin(lmarkslist &lmarks,int &globalminx,int &globalmaxx,int &globalminy,int &globalmaxy);
void choose_ltas(const igrid* temp, int globalminx, int globalminy, int globalmaxx, int globalmaxy, int incrx, int incry);
void nistprocess(image &im);
// __________________________________________________________________________
// Read in the coordinates for the training images, scaling as needed.

extern images IMV;
extern int CURRDET;
extern int CURRIM;




// __________________________________________________________________________
// Option 42
// Get statistics of the registration points marked on a data set
// The averages can be used as the anchor points on the reference grid
void get_coord_stats()
{
  // Read and scale the registration points of the training images.
  igrid point1, point2, point3;
  read_coords(point1, point2, point3, ratio, 0, numdigits);

  // Loop thru the desired images.  Calculate the average positions of
  // each of the three special points and the average dimensions
  // Also estimate box factors used to draw a sensible box around
  // a triangle
  double avgdimx = 0, avgdimy = 0;
  double avgp1x = 0, avgp1y = 0;
  double avgp2x = 0, avgp2y = 0;
  double avgp3x = 0, avgp3y = 0;
  boxfactorx = 0, boxfactory = 0;
  cout << numdigits << endl;
  cout << dataset << endl;

  for (int i = 0; i < numdigits; ++i)
  {
    cout << "Processing image " << i << endl;

    // Read and process the image
    image im;
    if (!im.read(i + first))
    {
      cerr << "Error in reading image.\n";
      exit(1);
    }
    im.blur_downscale(ratio, sd);

   
    // Update the average location of the dimensions and registration points
    cout << im.dimx() << " " << im.dimy() << " " << point1(i,0)
         << " " << point3(i,1) << endl;
    avgdimx += im.dimx();
    avgdimy += im.dimy();
    avgp1x += point1(i, 0);
    avgp1y += point1(i, 1);
    avgp2x += point2(i, 0);
    avgp2y += point2(i, 1);
    avgp3x += point3(i, 0);
    avgp3y += point3(i, 1);
  }

  avgdimx /= numdigits;
  avgdimy /= numdigits;
  avgp1x /= numdigits;
  avgp1y /= numdigits;
  avgp2x /= numdigits;
  avgp2y /= numdigits;
  avgp3x /= numdigits;
  avgp3y /= numdigits;

  // Output the coordinate statistics to the file "basis"
  ofstream ofs("basis");


  // The vectors joining the second two points to the first
  ofs << "vector1x = " << rint(avgp2x - avgp1x) << endl;
  ofs << "vector1y = " << rint(avgp2y - avgp1y) << endl;
  ofs << "vector2x = " << rint(avgp3x - avgp1x) << endl;
  ofs << "vector2y = " << rint(avgp3y - avgp1y) << endl;

  // ANchor points for smaller reference grid used for classification trees
  // The registration point locations in a 16x16 image.
  point1x = int(avgp1x * cartsizex / avgdimx + 0.5);
  point1y = int(avgp1y * cartsizey / avgdimy + 0.5);
  point2x = point1x + int((avgp2x - avgp1x) * cartsizex / avgdimx + 0.5);
  point2y = point1y + int((avgp2y - avgp1y) * cartsizey / avgdimy + 0.5);
  point3x = point1x + int((avgp3x - avgp1x) * cartsizex / avgdimx + 0.5);
  point3y = point1y + int((avgp3y - avgp1y) * cartsizey / avgdimy + 0.5);
  ofs << "point1x = " << point1x << endl;
  ofs << "point1y = " << point1y << endl;
  ofs << "point2x = " << point2x << endl;
  ofs << "point2y = " << point2y << endl;
  ofs << "point3x = " << point3x << endl;
  ofs << "point3y = " << point3y << endl;
  ofs << "cartsizex = " << cartsizex << endl;
  ofs << "cartsizey = " << cartsizey << endl;
  ofs << "fpoint1x = " << point1x << endl;
  ofs << "fpoint1y = " << point1y << endl;
  ofs << "fpoint2x = " << point2x << endl;
  ofs << "fpoint2y = " << point2y << endl;
  ofs << "fpoint3x = " << point3x << endl;
  ofs << "fpoint3y = " << point3y << endl;
  ofs << "fcartsizex = " << cartsizex << endl;
  ofs << "fcartsizey = " << cartsizey << endl;


} // get_coord_stats

// __________________________________________________________________________
// Option 43
// Gray level statistics, register image and find means and STD's.
void register_images()
{
  int i;
  ofstream tofs;
  // Declare and initialize the array of registered images.
  dgrid meanim(cartsizex,cartsizey);
  meanim.clear();
  dgrids regims(numdigits);
  if (dum1==0){
  for (i = 0; i < numdigits; ++i)
    regims[i].init(cartsizex, cartsizey);
  }
  if (dum1>0)
    tofs.open("regims");
  // Read and scale the registration coordinates of the 
  // original training images.
  igrid point1, point2, point3;
  read_coords(point1, point2, point3, ratio, 0, numdigits);

  // Calculate the fixed part of the registration transform.
  // The point[xy][12] variables are read from the parameter file.
  partialtrans pt(Pixel(point1x, point1y), Pixel(point2x, point2y),
		  Pixel(point3x, point3y));

  double minstd=10000;
  ucgrid pmask;
  make_procmask(pmask, masktype, cartsizex, cartsizey);
  for (int k = 0; k < numdigits; k++)
  {
    cout << "Registering image " << k << endl;

    // Read and process the image
    image im;
    if (!im.read(k))
    {
      cerr << "Error in reading image.\n";
      exit(1);
    }
    im.blur_downscale(ratio, sd);

    // Finish calculating the registration transform
   if (dum1==0)
     {
    affinetrans at(pt, Pixel(point1(k, 0), point1(k, 1)),
		   Pixel(point2(k, 0), point2(k, 1)), 
		   Pixel(point3(k, 0), point3(k, 1)));

    // Calculate the registered image and normalize it
    register_to_dgrid(im, regims[k], at);
    meanim += regims[k];
     
    double deviation = l2_normalize(regims[k], pmask);
    minstd = MIN(minstd, deviation);
     }
    else
      {
     ucgrid uct(im.dimx(),im.dimy());
     gridfun(im(i,j),uct);
         uct.write(tofs);
      }
  }
  if (dum1==0)
    {
      gridfun(meanim(i,j)/numdigits,meanim);
      view_dnormalized(meanim);
      cout << meanim.dimx() << " " << meanim.dimy() << endl;
    }

  if (dum1>0)
    return;
  
  // Write out the binary registered images to the file "MEAN"
  ofstream rofs("MEAN");
  meanim.write(rofs);
  rofs.close();
  
  // Write out the minimum standard deviation of the registered
  // images to the file "MINSTD"
  ofstream ofs("MINSTD");
  ofs << "MINSTD = " << (minstd * .8) << endl;
  ofs.close();

  // Determine the maximum L1 nearest neighbor distance between
  // the 16x16 training images, enlarge it, and write it out to 
  // the file "MAXL1"
  double maxdist = 0;
  for (i = 0; i < numdigits; i++)
  {
    double mindist = 1000000;
    for (int j = 0; j < numdigits; j++)
    {
      if (j != i)
      {
	double dist = 0;
	for (int x = 0; x < cartsizex; x++)
	  for (int y = 0; y < cartsizey; y++)
	    dist += fabs(regims[i](x, y) - regims[j](x, y));
	mindist = MIN(dist, mindist);
      }
    }
    maxdist = MAX(mindist, maxdist);
  }
  ofs.Open("MAXL1");
  ofs << "maxl1 = " << maxdist * 25 << endl;

} // register_images

// __________________________________________________________________________


void update_temp(const landmarks& lm, igrid *temp, int spread, int numt, int globalminx, int globalminy)
{
  for (int currtype = 0; currtype < numt; currtype++)
  {
    ucgrid ttemp(temp[0].dimx(),temp[0].dimy());
    ttemp.clear();
    const int numins = lm.type_count(currtype);
    for (int currinst = 0; currinst < numins; ++currinst)
    {
      const int x = lm.location_x(currtype, currinst);
      const int y = lm.location_y(currtype, currinst);
     if (ttemp(x,y)==0)
       {
      ttemp(x,y)=1;
      temp[currtype]( (x-globalminx)/spread, (y-globalminy)/spread) += 1;
       }

    }
  }

} // update_temp





// __________________________________________________________________________
// Option 44 - Extract edges or LTA's or hardwired edgepairs from a data base
//             andwrite to a directory `landdir'.
// onlyedges=1 - just edges according to settings of parameters
//               EdgeSizeBlock, EdgeEpsilon, EdgeGoodTests, edgeinc,
//               EdgeNumberOfPixels.
// onlyedges=0 + ltatreedir - computes edges and then LTA's defined
//               by file `ltatrees' in the directory given by ltatreedir
//
// onlyedges=0 + ltausedepth=0 computes all 256 hardwired edgepairs.
//                     `spreadforpairs' sets the ammount of spreading
//                      of each edge.
//                     `EdgeRadius' - Distance of second edge from first.  
// merge=1   merge the original edge features with the LTA's.
void make_edge_maps()
{
  // Read and scale the registration coordinates of the 
  // original training images.  This gets put in the landmarks
  // so that they don't need to be read in again later.
  igrid point1, point2, point3;
  read_coords(point1, point2, point3, ratio, 0, numdigits);

  lmarkslist lands(numdigits);
  lmarkslist ltalands(numdigits);
  // Loop throught the relevant digits.
  for (int ii = 0; ii < numdigits; ++ii)
  {
    // Read in the digit and process it.
    image im;
    //    cout << "Processing digit " << ii << endl;
    if (!im.read(ii + first))
      {
	cerr << "Error in reading image.\n";
	exit(1);
      }
    if (nist)
      nistprocess(im);
    else
      {
	int x;
	im.blur_downscale(ratio, sd);
	if (DIMX>0)
	  {
	    im.imtrim(-DIMX,x,x,x,x,x,x);
	    im.imslant(1);
	    im.imtrim(-DIMX,x,x,x,x,x,x);
	  }
	else if (DIMX<0)
	  im.imtrim(DIMX,x,x,x,x,x,x);
	//view_normalized(im);
      }
      if (alpha2>0)
	{
	  //cout << point1(ii,1) << " " << endl;
	  int x1=point1(ii,0);
	  int y1=point1(ii,1);
	  int x2=point2(ii,0);
	  int y2=point2(ii,1);
	  int x3=point3(ii,0);
	  int y3=point3(ii,1);
	  im.std_process(0, alpha2, 0, 0, x1, y1, x2, y2, x3, y3);
	  //	  view_normalized(im);
	  point1(ii,0)=x1; point1(ii,1)=y1;
	  point2(ii,0)=x2; point2(ii,1)=y2;
	  point3(ii,0)=x3; point3(ii,1)=y3;
	  //cout << point1(ii,1) << " " << endl;
	}

    if (reverseimage)
      gridfun(255-im(i,j),im);
    ltatrees ltrees;
    //onlyedges=1;
    extract_landmarks(im, ltrees, numltas, lands[ii], ltalands[ii]);
    if (onlyedges)
      {
	lands[ii].coords[0] = point1(ii, 0);
	lands[ii].coords[1] = point1(ii, 1);
	lands[ii].coords[2] = point2(ii, 0);
	lands[ii].coords[3] = point2(ii, 1);
	lands[ii].coords[4] = point3(ii, 0);
	lands[ii].coords[5] = point3(ii, 1);
      }
    else
      {
	ltalands[ii].coords[0] = point1(ii, 0);
	ltalands[ii].coords[1] = point1(ii, 1);
	ltalands[ii].coords[2] = point2(ii, 0);
	ltalands[ii].coords[3] = point2(ii, 1);
	ltalands[ii].coords[4] = point3(ii, 0);
	ltalands[ii].coords[5] = point3(ii, 1);
	if (merge)
	  ltalands[ii].merge(lands[ii]);
      }
    
  }

   if (onlyedges)
    write_lands(landdir, numdigits, lands);
   else  
     write_lands(landdir, numdigits, ltalands);


} // make_edge_maps


// __________________________________________________________________________
// option 45
// Register edges from a data base according to affine map taking registration
// points to 3 anchor points. A uniform shift is applied to make sure all coordinates are positive.
void transform_edge_maps()
{
  Pixel q1(0, 0), q2(vector1x, vector1y), q3(vector2x, vector2y);

  int i;
  int count = 0;
  landmarks l;
  lmarkslist lands(numdigits);
  for (i = first; i < first + numdigits; ++i)
  {
    cout << i << endl;
    l.read(i);

    lands[count].set_truth(l.truth());
    lands[count].init_counts_offsets(l.total_types());

    Pixel p1(l.coords[0], l.coords[1]), p2(l.coords[2], l.coords[3]),
             p3(l.coords[4], l.coords[5]);


    affinetrans pqtrans(p1, p2, p3, q1, q2, q3);

    int numlands = 0;
    lmarklist l2(maxlands);
    // Register to anchor points and shift (by `shift') so that all coordinates are positive.
    shift=0;
    affine_trans_lm(l, pqtrans, l2, numlands);
  
   
    lands[count].setup_from_lmarklist(l2, numlands);
    count++;
  }
  int globalmaxx, globalminx, globalmaxy, globalminy;
  // return;

   // Find the required shift to make all cooridnates positive
  find_maxmin(lands, globalminx, globalmaxx, globalminy, globalmaxy);
  if (shiftx<0)
    {
      shiftx=MAX( MAX(0,-globalminx), MAX(0,-globalminy))+5;
      shifty=shiftx;
    }
  cout << " SHIFT " << shiftx << " " << shifty << endl;
  // Shift all edges by that ammount
  for (int i=0; i<count; i++)
    {
      void shift_lm(landmarks &l, int sftx, int sfty);
      shift_lm(lands[i],shiftx,shifty);
      lands[i].coords[0] = q1[0] + shiftx; 
      lands[i].coords[1] = q1[1] + shifty;
      lands[i].coords[2] = q2[0] + shiftx;
      lands[i].coords[3] = q2[1] + shifty;
      lands[i].coords[4] = q3[0] + shiftx;
      lands[i].coords[5] = q3[1] + shifty;
    }
  ofstream ofs("basis2");

  // These will be the anchor points for the the model reference grid.
  ofs << "ltatree1x = " << lands[0].coords[0] << endl;
  ofs << "ltatree1y = " << lands[0].coords[1] << endl;
  ofs << "ltatree2x = " << lands[0].coords[2] << endl;
  ofs << "ltatree2y = " << lands[0].coords[3] << endl;
  ofs << "ltatree3x = " << lands[0].coords[4] << endl;
  ofs << "ltatree3y = " << lands[0].coords[5] << endl;


 // Write out landmarks
    write_lands(newlanddir, count, lands);
} // transform_edge_maps

// opt=28, train frequency maps for deformable image Bernoulli model
// Set reference points point1x ... point3y and dimensions of
// reference grid DIMX, DIMY
void get_edgetemps()
{
 image im;
  landmarks edgelands, ltalands;
  void register_image(const image& im, partialtrans& pt, 
		      dgrid& regim, ucgrids& reglands, int *tripts, int j,
		      int cstyle, int csizex, int csizey, ucgrid& pmask,
		      int lsq, int unions = 0);
  ucgrid pmask;
  partialtrans fpt(Pixel(point1x, point1y), Pixel(point2x, point2y),
		   Pixel(point3x, point3y));
  igrid point1, point2, point3;
  read_coords(point1, point2, point3, ratio, 0, numdigits);
  ucgrids lands;
  dgrid cartims;
  ucgrids totlands;
  int  style=3;
  if (onlyedges)
    style=2;
  dgrid totim;
  for (int ii=0; ii<numdigits; ii++)
    {
      im.read(first+ii);
      if (nist)
	nistprocess(im);
      else
	im.blur_downscale(ratio, sd);
      
      if (ii>0)
	{
	  if(style==2)
	    empty_STAGS(edgelands);
	  if(style==3)
	    empty_STAGS(ltalands);
	}	
      // Extract landmarks for this image.
      extract_landmarks(im, ltrees, numltas, 
			edgelands, ltalands);
      if (onlyedges)
	fill_STAGS(edgelands, spread);
      else 
	fill_STAGS(ltalands, spread);
      
      int pertcoor[6];
      pertcoor[0]=point1(ii,0); pertcoor[1]=point1(ii,1);
      pertcoor[2]=point2(ii,0); pertcoor[3]=point2(ii,1);
      pertcoor[4]=point3(ii,0); pertcoor[5]=point3(ii,1);
      // Register the landmarks.
      register_image(im, fpt, cartims, lands,
		     pertcoor, 0,  style, 
		     DIMX, DIMY, 
		     pmask, leastsquares);
      
      
      if (style>=2)
	{
	  if (ii==0)
	    totlands.init(lands.dim());
	  for (int k=0; k<lands.dim(); k++)
	    {
	      if (ii==0)
		{
		  totlands[k].init(DIMX,DIMY);
		  totlands[k].clear();
		}
	      // Accumulate counts at each location for each type.
	      gridfun(totlands[k](i,j)+lands[k](i,j),totlands[k]);
	    }
	}
    }
  if (style>=2)
    {
      //Write it out.
      ofstream ofs(tempfile);
      for (int k=0; k<totlands.dim(); k++)
	{
	  ucgrid uc(DIMX, DIMY);
	  dgrid duc(DIMX, DIMY);
	  
	  gridfun( (double)totlands[k](i,j)/numdigits,duc);
	  duc.write(ofs);
	  
	  if (showtriangles==1)
	    view_dnormalized(duc);
	  
	}
    }
  
}



// __________________________________________________________________________
// opt=88
// Edge statistics obtained from registered edge maps.
// showedges=0 - Use edge frequencies to determine edge based model on reference grid
// showedges=2 - Show edge histograms as images.
// showedges=5 - Use edge based frequencies to determine ltas without checking joint stats.
void get_mean_edgemap()
{
  void update_temp(const landmarks& lm, igrid *temp, 
			   int spread, int numt, int globalminx, int globalminy);
  void choose_ltas_from_edges(igrids &temp);
  igrids temp;
  ucgrids uctemp;

  numdata=last-first;
  lmarkslist lmarks(numdata);
  for (int ii = 0; ii < numdata; ++ii)
        lmarks(ii).read(ii);

  // Determine landmark extents
  int globalmaxx, globalminx, globalmaxy, globalminy;
  
  find_maxmin(lmarks, globalminx, globalmaxx, globalminy, globalmaxy);
  int globdifx = globalmaxx-globalminx+2;
  int globdify = globalmaxy-globalminy+2;

  
  numltas=lmarks(0).total_types();
  temp.init(numltas);
  uctemp.init(numltas);
  for (int ii=0; ii<numltas; ii++) 
    {
      // Estimate edge probabilities on the original grid
      if (showedges<5)
	{
	  temp[ii].init(globdifx, globdify);
	  uctemp[ii].init(globdifx, globdify);
	}
      // Estimate edge probabilities with locations down sam[pled to a caorser grid by spread
      if (showedges==5)
	temp[ii].init( globdifx/spread, globdify/spread);
      temp[ii].clear();
    }

  

  ofstream ofs(tempfile);

 
  
  // Read in the lists of landmarks
  for (int ii = first; ii < last; ++ii)
    {
      if (showedges<5)
	{
	  fill_STAGS(lmarks(ii),spread);
	  for (int jj = 0; jj < numltas; jj++)
	    {
	      gridfun((STAGS[jj](i+globalminx,j+globalminy)>0)+temp[jj](i,j),temp[jj]);		
	    }
	  empty_STAGS(lmarks(ii));
	}
       else if (showedges==5)
	 update_temp(lmarks(ii), temp, spread, 8, globalminx, globalminy);
      
    }

  if (showedges==2)
    {
      dgrid dtemp(temp[0].dimx(),temp[0].dimy());
      for (int jj=0 ; jj<numltas; jj++)
	{
	  if ( objprop >= 0)
	    {
	      gridfun(temp[jj](i,j)*(double(temp[jj](i,j))/numdata>objprop),dtemp);
	    }
	  else
	    {
	      gridfun((double(temp[jj](i,j))/numdata < -objprop),dtemp);
	    }
	  cout << dtemp.dimx() << " " << dtemp.dimy() << endl;
	  view_dnormalized(dtemp);
	}
    }
 
  // Choose high probability edge locations for edge based model.
  if (showedges==0)
    {
      ltrees.init(numltas);
      for (int jj=0 ; jj<numltas; jj++)
	{
	ltrees[jj].tnear=jj;
	// Flag that this is an edge tree with no angle information
        ltrees[jj].tnumang=-1;
        ltrees[jj].maketop();
	}
      choose_ltas(temp, 0, 0, temp[0].dimx(), temp[0].dimy(), globalminx, globalminy);
    }

  // Choose lta's directly from edge probabilities
  if (showedges==5)
      choose_ltas_from_edges(temp);

  
} // get_mean_edgemap


// Option 89
// lta statistics, also used to get lats from a fixed pool
// Assuming edges are read from NON-registered edge landmark file.
// Reads in edges, computes lta locations, registers them and gets frequencies
// spread > 0 compute frequencies.
//                 showltas>0 - just show frequencies.
//                 showltas=0 - get lta model based on these frequencies.
//		   Is used when a predefined pool of ltas exists.
void get_lta_maps()
{
  int numltas=read_ltatrees(ltrees,ltatreedir);

  igrids temp(numltas);
  igrid stag;
  void view_ltafreqs(igrids &temp, ltatrees &ltrees);
  
  int globalminx=0, globalminy=0, globalmaxx=200, globalmaxy=200;
  Pixel q1(ltatree1x, ltatree1y), q2(ltatree2x, ltatree2y), q3(ltatree3x, ltatree3y);
  int ii;
  for (ii=0; ii<numltas; ii++)
    {
      temp[ii].init(globalmaxx-globalminx+5,globalmaxy-globalminx+5);
      temp[ii].clear();
    }
  


  numang = ltrees[0].tnumang;
  if (near >=0)
    near = ltrees[0].tnear;
  else 
    near = -near;
  make_rel_masks();

  lmarkslist ltalands(last-first);
  int count=0;
  for (int in = first; in < last; ++in)
    {
      //cout << "image " << in << endl;
      
      // Find the triangles
      landmarks edgelands;
      landmarks lands;
      lmarklist l2(maxlands);
      edgelands.read(in);
      void spread_and_aggreg(landmarks &lands, int numt, igrids &temp);
      
      Pixel p1(edgelands.coords[0], edgelands.coords[1]),
	p2(edgelands.coords[2], edgelands.coords[3]),
	p3(edgelands.coords[4], edgelands.coords[5]);
      affinetrans pqtrans(p1, p2, p3, q1, q2, q3);

      edges_to_ltas(edgelands, lands, ltrees, numltas);
      if (spread >= 0)
	{
	  int numl;
	  // Registers the selected points on image to anchor points given by vetcor1-3
	  // Then shift everything by shift to have positive coordinates. This is
	  // Consistent with the way the edges are originally registered to train
	  // the ltas.
	  affine_trans_lm(lands, pqtrans, l2, numl);
	  ltalands[count].init_counts_offsets(lands.total_types());
	  ltalands[count].setup_from_lmarklist(l2, numl);
	  spread_and_aggreg(ltalands[count], numltas, temp);
	}
      count++;

    }
  
  // Write out landmarks
  if (spread < 0)    
    write_lands(newlanddir, count, ltalands);
  else
    {
      if (showltas>=1) // View lta frequencies
	view_ltafreqs(temp, ltrees);
      else       // Choose ltas from high probability locations
	choose_ltas(temp, globalminx, globalminy, globalmaxx, globalmaxy, 0, 0);
    }

} // get_lta_maps
     
// __________________________________________________________________________
// faces.C

#endif // FACES_C


