// Copyright YALI AMIT, January 15, 2001.

#ifndef VIEWPORT_C
#define VIEWPORT_C

// __________________________________________________________________________



// __________________________________________________________________________

#include "includes.h"


#include <iostream>
//#include <sys/types.h>

#ifndef NOX
//#include <unistd.h>
#define String XString
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif
#undef String

#include "viewport.h"
#include "grid.h"
#include "globals.h"

extern double cs[32], sn[32];
int pressnum;
// __________________________________________________________________________

int viewport::refresh(int needkey)
{
#ifdef NOX


  cout << "No graphics " << endl;
  return 0;
#else
 
  pressnum=0;
  // First time thru initialize the window
  if (!exposed) 
  {
    winmapped = 1;
    vpixsize = pixsize;
    vrevvideo = revvideo;
    vshowpixels = showpixels;
    if (gr && gr->dimx() && gr->dimy()) {
      width = gr->dimx() * vpixsize + 2 * imborder;
      height = gr->dimy() * vpixsize + 2 * imborder;
    }
    else {
      height += 2 * imborder;
      width += 2 * imborder;
    }
    border = 4;

    window =
      XCreateWindow(display, RootWindow(display, screen), 0, 0,
                    width, height, border,  DefaultDepth(display, screen), 
                    InputOutput, DefaultVisual(display, screen), CWBackPixel |
                    CWBorderPixel | CWColormap | CWEventMask, &xswa);

    // Initialize the graphics context.
    init_gc();

    // Make sure we don't crash if quit is chosen from the wm.
    XSetWMProtocols(display, window, &wmdw, 1);

    // Initialize the window manager properties.
    char *name = "window";
    XTextProperty window_name;
    XStringListToTextProperty(&name, 1, &window_name);
    xch.res_class = "Face";
    xch.res_name = "face";
    //xsh.flags = PPosition | PSize;
    xwmh.flags = InputHint | StateHint;
    xwmh.input = True;
    xwmh.initial_state = NormalState;
    XSetWMProperties(display, window, &window_name, 0, 0, 0, 
                     &xsh, &xwmh, &xch);
    XFree(window_name.value);

    // Display the window.
    XMapRaised(display, window);
  }

  // Check to see if we need to handle any drawing not specifically
  // handled by exposure events
  if (drawflag) {
    if (drawflag >= 2)
      redraw();
    else
      draw();
    drawflag = 0;
  }

  // Enter event loop.
  XEvent event;
  int gotevent;
  int keycall=1;
  while ((gotevent = XCheckWindowEvent(display, window, ~0, &event)) || 
         !exposed || needkey)
  {
    if (keycall)
      {
	/*
	if (needkey){
	cout << " Type `c' inside window to continue " << endl;
	cout << " Type a number inside window to magnify " << endl;
	cout << " Type `q' inside window to quit " << endl;}
	*/
	keycall=0;
      }
    if (!gotevent)
      XWindowEvent(display, window, ~0, &event);
    switch (event.type)
    {
     case ClientMessage :
      // Handle case WM kills the window.
      if (event.xclient.data.l[0] == int(wmdw))
      {
	XDestroyWindow(display, window);
        winmapped = 0;
	XFlush(display);
        exit(1);
      }
      break;

     case Expose:
      exposed = 1;
      proc_expose(event);
      break;

     case ConfigureNotify:
      proc_confignotify(event);
      break;

     case ButtonPress:
      proc_buttonpress(event);
      break;

     case ButtonRelease:
      proc_buttonrelease(event);
      break;

     case MotionNotify:
      proc_buttonmotion(event);
      break;

     case KeyRelease:
     {
       char thechar = proc_keyrelease(event);
       if (thechar == 'q')
       {
         XDestroyWindow(display, window);
         winmapped = 0;
         XFlush(display);
         exit(1);
       }
       if (thechar == 'c')
         return 0;
       if (thechar == 'z')
         return(-1);
     }
     break;

     default:
      break;
    }
  }

  return 1;
#endif
} // viewport::refresh

// __________________________________________________________________________

int viewport::view()
{
  return refresh(1);

} // viewport::view

// __________________________________________________________________________

int viewport::wait()
{
#ifdef NOX
return 0;
#else

  XEvent event; 
  XPeekEvent(display, &event); 
  return 1; 
#endif
} // viewport::wait

// __________________________________________________________________________
// Draw an image in a window.  Pay close attention to X events to avoid 
// unwanted redraws.

void viewport::draw()
{

#ifdef NOX
return;
#else
  if (!exposed)
    return;

  if (gr && gr->dimx() && gr->dimy())
  {
    if (xlowx < imborder) xlowx = imborder;
    if (xlowy < imborder) xlowy = imborder;
    if (xhighx > imborder + gr->dimx() * vpixsize - 1)
      xhighx = imborder + gr->dimx() * vpixsize - 1;
    if (xhighy > imborder + gr->dimy() * vpixsize - 1)
      xhighy = imborder + gr->dimy() * vpixsize - 1;
    if (xhighx < xlowx || xhighy < xlowy)
      goto GEOM;

    int imdepth = DefaultDepth(display, screen);
    int iwidth = xhighx - xlowx + 1;
    int iheight = xhighy - xlowy + 1;

    Pixmap impixmap;
    impixmap = XCreatePixmap(display, window, iwidth, iheight, imdepth);

    // create an XImage
    int data_len = iwidth * imdepth / 8;
   

    int i = 0;
    for (i = 0; i < iheight; ++i)
    {
      // Stop drawing if an expose or button release event occurs.
      XEvent newevent;
      if (XCheckTypedWindowEvent(display, window, Expose, &newevent) ||
          (XCheckTypedWindowEvent(display, window, ButtonRelease, &newevent) &&
           (newevent.xbutton.button != Button2)))
      {
        XPutBackEvent(display, &newevent);
        XFreePixmap(display, impixmap);
	return;
      }
      const int ival = (i + xlowy - imborder) / vpixsize;
      for (int j = 0; j < iwidth; ++j)
        {
          int numval = (*gr)((j + xlowx - imborder) / vpixsize, ival);
	  unsigned int pval;	  
	  if (br < 1)
            pval = int(br * (vrevvideo ? numcolors - 1 - numval :
			     numval));
          else
            pval = numcolors - int(1 / br * (numcolors -
             (vrevvideo ? numcolors - 1 - numval : numval))); 
	  XSetForeground(display, gc,
            grays[pval].pixel);
          XDrawPoint(display, impixmap, gc, j, i);         
        }
      

    }

   

    if (vshowpixels && vpixsize > 1)
    {
      XSetForeground(display, gc, (int)basecol);
      for (i = 0; i < gr->dimx() * vpixsize; i += vpixsize)
      {
        for (int j = 0; j < gr->dimy() * vpixsize; j += vpixsize)
          if ((*gr)(i / vpixsize, j / vpixsize) != 0)
            XDrawRectangle(display, impixmap, gc, i, j, vpixsize, vpixsize);
      }
    }

    XCopyArea(display, impixmap, window, gc, 0, 0,
              iwidth, iheight, xlowx, xlowy);
    XFreePixmap(display, impixmap);
  }
  else
    XClearWindow(display, window);

  GEOM:

  // Stop drawing if an expose or button release event occurs.
  XEvent newevent;
  if (XCheckTypedWindowEvent(display, window, Expose, &newevent) ||
      (XCheckTypedWindowEvent(display, window, ButtonRelease, &newevent) &&
       (newevent.xbutton.button != Button2)))
  {
    XPutBackEvent(display, &newevent);
    return;
  }

  for (vpiter vpi = points.first(); vpi; ++vpi)
    draw_point(*vpi);
  for (vliter vli = lines.first(); vli; ++vli)
    draw_line(*vli);
  for (vbiter vbi = boxes.first(); vbi; ++vbi)
    draw_box(*vbi);
  for (vtiter vti = texts.first(); vti; ++vti)
    draw_string(*vti);
#endif
} // viewport::draw

// __________________________________________________________________________
// Redraw an image in a window.

void viewport::redraw()
{
  xlowx = xlowy = 0;
  xhighx = xhighy = 1000000;
  draw();

} // viewport::redraw

// __________________________________________________________________________
// Process an Expose event.

#ifndef NOX
void viewport::proc_expose(XEvent &event)
{


  // Update the exposed region bounds.
  xlowx = MIN(xlowx, event.xexpose.x);
  xlowy = MIN(xlowy, event.xexpose.y);
  xhighx = MAX(xhighx, event.xexpose.x + event.xexpose.width);
  xhighy = MAX(xhighy, event.xexpose.y + event.xexpose.height);

  // Leave now if there are more recent expose events.
  if (event.xexpose.count != 0)
    return;

  // Draw the object.
  draw();

  // Reset the exposed region bounds if no pending expose events.
  XEvent newevent;
  if (XCheckTypedWindowEvent(display, window, Expose, &newevent))
  {
    XPutBackEvent(display, &newevent);
    return;
  }
  else
  {
    xlowx = xlowy = 1000000;
    xhighx = xhighy = 0;
  }

} // viewport::proc_expose

// __________________________________________________________________________
// Process a ConfigureNotify event.

void viewport::proc_confignotify(XEvent&)
{


  XWindowAttributes xwa;
  if (XGetWindowAttributes(display, window, &xwa) != 0)
  {
    if (!gr || gr->dimx() == 0 || gr->dimy() == 0) {
      width = xwa.width;
      height = xwa.height;
    }
    else
      vpixsize = MAX(1, MIN((xwa.width - 2 * imborder) / gr->dimx(), 
                           (xwa.height - 2 * imborder) / gr->dimy()));
  }

} // viewport::proc_confignotify

// __________________________________________________________________________
// Process a ButtonRelease event.

void viewport::proc_buttonrelease(XEvent &event)
{

  switch (event.xbutton.button)
  {
   case Button1 :
   {
     int bx = (event.xbutton.x - imborder) / vpixsize;
     int by = (event.xbutton.y - imborder) / vpixsize;
     if (event.xbutton.x >= 0 && event.xbutton.x < width &&
         event.xbutton.y >= 0 && event.xbutton.y < height)
     {
       cout << bx << " " << by << " ";
       add_point(bx, by, basecol,5);
       draw();
       //if (gr) cout << int(gr->get(bx, by));
       cout << endl;
     }
   }

    break;

  case Button3 :
   {
     int bx = (event.xbutton.x - imborder) / vpixsize;
     int by = (event.xbutton.y - imborder) / vpixsize;
     if (event.xbutton.x >= 0 && event.xbutton.x < width &&
         event.xbutton.y >= 0 && event.xbutton.y < height)
     {
       cout << bx << " " << by << " ";
       add_point(bx, by, basecol,3);
       add_string(bx,by,ltos(pressnum),basecol);
       pressnum++;
       draw();
       if (gr) cout << int(gr->get(bx, by));
       cout << endl;
     }
   }

    break;


  

   case Button2 :
   {
     // Increase or decrease the size of the window.
     int oldvpixsize = vpixsize;
     if (event.xbutton.button == Button1)
       vpixsize -= (1 * (vpixsize > 1));
     else
       vpixsize += 1;
     if (gr && gr->dimx() && gr->dimy()) {
       height = gr->dimy() * vpixsize + 2 * imborder;
       width = gr->dimx() * vpixsize + 2 * imborder;
     }
     else {
       height *= vpixsize;
       width *= vpixsize;
       height /= oldvpixsize;
       width /= oldvpixsize;
     }
     XResizeWindow(display, window, width, height);
   }
    break;

   default:
    break;
  }

} // viewport::proc_buttonrelease

// __________________________________________________________________________
// Process a KeyRelease event.

char viewport::proc_keyrelease(XEvent &event)
{

  char thechar = XKeycodeToKeysym(display, event.xkey.keycode, 0);
  if (thechar == 'd')
    redraw();
  else if (thechar == 'b') {
    br /= 0.9;
    redraw();
  }
  else if (thechar == 'w') {
    br *= 0.9;
    redraw();
  }
  else if (thechar == 'r') {
    vrevvideo = 1 - vrevvideo;
    redraw();
  }
  else if (thechar == 's') {
    vshowpixels = 1 - vshowpixels;
    redraw();
  }
  else if (thechar >= '1' && thechar <= '9') {
    vpixsize = thechar - '0';
    if (gr && gr->dimx() && gr->dimy()) {
      height = gr->dimy() * vpixsize + 2 * imborder;
      width = gr->dimx() * vpixsize + 2 * imborder;
      XResizeWindow(display, window, width, height);
    }
    else
      draw();
  }
  else if (vpixsize > 1 && (thechar == '<' || thechar == ',')) {
    --vpixsize;
    if (gr && gr->dimx() && gr->dimy()) {
      height = gr->dimy() * vpixsize + 2 * imborder;
      width = gr->dimx() * vpixsize + 2 * imborder;
    }
    else {
      height -= 10;
      width -= 10;
    }
    XResizeWindow(display, window, width, height);
  }
  else if (thechar == '>' || thechar == '.') {
    ++vpixsize;
    if (gr && gr->dimx() && gr->dimy()) {
      height = gr->dimy() * vpixsize + 2 * imborder;
      width = gr->dimx() * vpixsize + 2 * imborder;
    }
    else
    {
      height += 10;
      width += 10;
    }
    XResizeWindow(display, window, width, height);
  }

  return thechar;

} // viewport::proc_keyrelease

// __________________________________________________________________________
// Create a gc and initialize it's main properties.

void viewport::init_gc()
{
  gc = XCreateGC(display, window, 0, 0);
  XSetFont(display, gc, xfs->fid);
  XSetLineAttributes(display, gc, 2, LineSolid, CapProjecting, JoinMiter);

} // viewport::init_gc
#endif
// __________________________________________________________________________

void viewport::add_grid(ucgrid& g)
{
#ifdef NOX
return;
#else
  gr = &g;
  if (gr->dimx() && gr->dimy())
  {
    height = gr->dimy() * vpixsize + 2 * imborder;
    width = gr->dimx() * vpixsize + 2 * imborder;
    if (exposed)
    {
      XResizeWindow(display, window, width, height);
      redraw();
    }
  }
#endif
} // viewport::add_grid

// __________________________________________________________________________

void viewport::add_ellipse(int x, int y, int xl, int yl, int ecol)
{
  //for (int n = 0; n < 32; n++)
    //add_point(int(x + xl * cs[n]), int(y + yl * sn[n]), ecol);
  for (int n = 0; n < 31; n++)
    add_line(int(x + xl * cs[n]), int(y + yl * sn[n]),
             int(x + xl * cs[n + 1]), int(y + yl * sn[n + 1]), ecol);
    add_line(int(x + xl * cs[31]), int(y + yl * sn[31]),
             int(x + xl * cs[0]), int(y + yl * sn[0]), ecol);

} // add_ellipse

// __________________________________________________________________________

void viewport::add_edge(int x, int y, int edgetype)
{
  int green, blue;
  if (basecol>0)
    {
      green= (int)(alpha*basecol);
      blue= (int)((1-alpha)*basecol);
    }
  else
    {
      green=greencol;
      blue=bluecol;
    }

  switch (edgetype)
  {
   case 0:
    add_point(x - 1, y - 1, blue);
    add_point(x, y - 1, blue);
    add_point(x - 1, y, green);
    add_point(x, y, green);
    break;

   case 1:
    add_point(x - 1, y - 1, green);
    add_point(x, y - 1, green);
    add_point(x - 1, y, blue);
    add_point(x, y, blue);
    break;

  case 2:
    add_point(x - 1, y - 1, blue);
    add_point(x, y , blue);
    add_point(x - 1, y, green);
    add_point(x, y+1, green);
    break;

   case 3:
    add_point(x - 1, y - 1, green);
    add_point(x, y , green);
    add_point(x - 1, y, blue);
    add_point(x, y + 1, blue);
    break;
 
   case 4:
    add_point(x - 1, y - 1, green);
    add_point(x - 1, y, green);
    add_point(x, y - 1, blue);
    add_point(x, y, blue);
    break;

   case 5:
    add_point(x - 1, y - 1, blue);
    add_point(x - 1, y, blue);
    add_point(x, y - 1, green);
    add_point(x, y, green);
    break;

  case 7:
    add_point(x, y , green);
    add_point(x - 1 , y+1, green);
    add_point(x, y-1 , blue);
    add_point(x - 1, y, blue);
    break;

   case 6:
    add_point(x , y , blue);
    add_point(x - 1, y+1, blue);
    add_point(x, y-1 , green);
    add_point(x-1, y, green);
    break;

  }

} // add_edge

// __________________________________________________________________________

void viewport::upd_grid()
{
#ifdef NOX
return;
#else
  if (gr->dimx() && gr->dimy())
  {
    height = gr->dimy() * vpixsize + 2 * imborder;
    width = gr->dimx() * vpixsize + 2 * imborder;
    if (exposed)
    {
      XResizeWindow(display, window, width, height);
      redraw();
    }
  }

#endif
} // viewport::upd_grid

// __________________________________________________________________________

void viewport::del_grid()
{
  gr = 0;

} // viewport::del_grid

// __________________________________________________________________________

void viewport::del_point(int x, int y)
{
  int done = 0;
  for (vpiter vpi = points.first(); !done && vpi; ++vpi) {
    if (vpi->x == x && vpi->y == y) {
      done = 1;
      points.remove(vpi);
    }
  }

} // viewport::del_point

// __________________________________________________________________________

void viewport::del_line(int x1, int y1, int x2, int y2)
{
  int done = 0;
  for (vliter vli = lines.first(); !done && vli; ++vli) {
    if (vli->x1 == x1 && vli->y1 == y1 && vli->x2 == x2 && vli->y2 == y2) {
      done = 1;
      lines.remove(vli);
    }
  }

} // viewport::del_line

// __________________________________________________________________________

void viewport::del_box(int x1, int y1, int x2, int y2)
{
  int done = 0;
  for (vbiter vbi = boxes.first(); !done && vbi; ++vbi) {
    if (vbi->x1 == x1 && vbi->y1 == y1 && vbi->x2 == x2 && vbi->y2 == y2) {
      done = 1;
      boxes.remove(vbi);
    }
  }

} // viewport::del_box

// __________________________________________________________________________

void viewport::del_string(int x, int y)
{
  int done = 0;
  for (vtiter vti = texts.first(); !done && vti; ++vti) {
    if (vti->x == x && vti->y == y) {
      done = 1;
      texts.remove(vti);
    }
  }

} // viewport::del_string

// __________________________________________________________________________

void viewport::clear_geom()
{
  points.empty();
  lines.empty();
  boxes.empty();
  texts.empty();

  drawflag |= 2;

} // viewport::clear_geom

// __________________________________________________________________________
// Draw a point.

void viewport::draw_point(const vpoint& vp) const
{

#ifdef NOX
  cout << "No graphics " << endl;
  return;
#else
  int pointsize = ((vp.sz > 0) ? vp.sz : vpixsize);
  //int pointsize=vp.sz*pixsize;
  XSetForeground(display, gc, defs[vp.col].pixel);

  XFillRectangle(display, window, gc, imborder + vp.x * vpixsize,
		 imborder + vp.y * vpixsize, pointsize, pointsize);

#endif
} // viewport::draw_point

// __________________________________________________________________________
// Draw a line.

void viewport::draw_line(const vline& vl) const
{

#ifdef NOX
  cout << "No graphics " << endl;
  return;
#else
  if (vl.x1 == vl.x2 && vl.y1 == vl.y2)
    return;

  XSetForeground(display, gc, defs[vl.col].pixel);
  if (vl.pos == 0)
  {
    XDrawLine(display, window, gc, 
              imborder + vl.x1 * vpixsize, imborder + vl.y1 * vpixsize, 
              imborder + vl.x2 * vpixsize, imborder + vl.y2 * vpixsize);
  }
  else
  {
    XDrawLine(display, window, gc, 
	      int(imborder + (vl.x1 + 0.5) * vpixsize), 
              int(imborder + (vl.y1 + 0.5) * vpixsize),
	      int(imborder + (vl.x2 + 0.5) * vpixsize), 
              int(imborder + (vl.y2 + 0.5) * vpixsize));
  }
#endif
} // viewport::draw_line

// __________________________________________________________________________

void viewport::draw_box(const vbox& vb) const
{
#ifdef NOX
  cout << "No graphics " << endl;
  return;
#else
  draw_line(vline(vb.x1, vb.y1, vb.x1, vb.y2, vb.col, 1));
  draw_line(vline(vb.x1, vb.y2, vb.x2, vb.y2, vb.col, 1));
  draw_line(vline(vb.x2, vb.y2, vb.x2, vb.y1, vb.col, 1));
  draw_line(vline(vb.x2, vb.y1, vb.x1, vb.y1, vb.col, 1));
#endif
} // viewport::draw_box

// __________________________________________________________________________
// Draw a string beginning on top of pixel (i, j).  If spos != 0, 
// draw it in the middle of the pixel.

void viewport::draw_string(const vtext& vt) const
{
#ifdef NOX
  cout << "No graphics " << endl;
  return;
#else
  XSetForeground(display, gc, defs[vt.col].pixel);
  if (vt.pos == 0)
    XDrawString(display, window, gc, imborder + vt.x * vpixsize, 
                imborder + vt.y * vpixsize, vt.txt, strlen(vt.txt));
  else
    XDrawString(display, window, gc, 
                int(imborder + (vt.x + 0.5) * vpixsize - fontwidth / 2.0),
                int(imborder + (vt.y + 0.5) * vpixsize + fontheight / 2.0),
                vt.txt, strlen(vt.txt));
#endif
} // viewport::draw_string

// __________________________________________________________________________

void viewport::get_marked_points(igrid &out)
{
  out.init(points.length(),2);

  int p =0;
  for (vpiter vpi = points.first(); vpi; ++vpi) {
      out(p, 0)=(*vpi).x; 
      out(p, 1)=(*vpi).y;
      p++;
  }

} // viewport::get_marked_points

// __________________________________________________________________________
// viewport.C



#endif // VIEWPORT_C

