// Copyright YALI AMIT, January 15, 2001.

#ifndef VIEWPORT_H
#define VIEWPORT_H


#ifndef NOX

#define String XString
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif
#undef String

#include "includes.h"
#include "simplist.h"
#include "simpstrg.h"
#ifndef NOX
#include "xobject.h"
#endif
template<class T> class vector;
template<class T> class grid;

typedef grid<unsigned char> ucgrid;
typedef grid<int> igrid;

class vpoint {
public:
  int x, y, col, sz;
  vpoint(int a, int b, int c, int d) : x(a), y(b), col(c), sz(d) { }
};
class vline {
public:
  int x1, y1, x2, y2, col, pos;
  vline(int a, int b, int c, int d, int e, int f) 
      : x1(a), y1(b), x2(c), y2(d), col(e), pos(f) { }
};
class vbox {
public:
  int x1, y1, x2, y2, col;
  vbox(int a, int b, int c, int d, int e) 
      : x1(a), y1(b), x2(c), y2(d), col(e) { }
};
class vtext {
public:
  int x, y, col, pos; simpstrg txt;
  vtext(int a, int b, int c, int d, const char* e) 
      : x(a), y(b), col(c), pos(d), txt(e) { }
};

typedef simplist<vpoint> vpoints;
typedef simplist<vline> vlines;
typedef simplist<vbox> vboxes;
typedef simplist<vtext> vtexts;
typedef simplistiterator<vpoint> vpiter;
typedef simplistiterator<vline> vliter;
typedef simplistiterator<vbox> vbiter;
typedef simplistiterator<vtext> vtiter;

// __________________________________________________________________________

class viewport   
#ifndef NOX 
: public xobject
#endif
{
private:

  // Window data
#ifndef NOX
  Window window;
  GC gc;
#endif

  int border;
  int height;
  int width;
  int exposed;
  int drawflag;
  int vpixsize, vrevvideo, vshowpixels;
  int winmapped;
  double br;

  // Object data
  const ucgrid* gr;
  vpoints points;
  vlines lines;
  vboxes boxes;
  vtexts texts;

  // Variables used to keep track of exposed regions.
  int xlowx, xlowy, xhighx, xhighy;

 public:
  // Setup members functions
  viewport(int h = 100, int w = 100) : 
    border(0), height(h), width(w), exposed(0), drawflag(0), winmapped(0),
    br(1.0), gr(0), xlowx(1000000), xlowy(1000000), xhighx(0), xhighy(0) { }
  viewport(const ucgrid& g) : 
    border(0), height(0), width(0), exposed(0), drawflag(0), vpixsize(0), 
    winmapped(0), br(1.0), gr(&g), xlowx(1000000), xlowy(1000000), xhighx(0), 
    xhighy(0) { }
#ifndef NOX
  ~viewport() { if (winmapped) XDestroyWindow(display, window); }
#endif
  void init_gc();

 private:
  // Event handling member functions

#ifndef NOX
  void proc_expose(XEvent& event);
  void proc_confignotify(XEvent&);
  void proc_buttonpress(XEvent&) { }
  void proc_buttonrelease(XEvent&);
  void proc_buttonmotion(XEvent&){ }
  char proc_keyrelease(XEvent& event);
#endif
 public:
  // Object drawing member functions
  void draw();
  void redraw();
  int refresh(int needkey = 0);
  int view();
  static int wait();
  void add_grid(ucgrid& g);
  void add_edge(int x, int y, int edgetype);
  void upd_grid();
  void add_point(int x, int y, int pcol = 0, int psize = 0) 
    { drawflag |= 1; points.insert(vpoint(x, y, pcol, psize)); }
  void add_line(int x1, int y1, int x2, int y2, int lcol = 0, int lpos = 0)
    { drawflag |= 1; lines.insert(vline(x1, y1, x2, y2, lcol, lpos)); }
  void add_box(int x1, int y1, int x2, int y2, int lcol = 0)
    { drawflag |= 1; boxes.insert(vbox(x1, y1, x2, y2, lcol)); }
  void add_string(int x, int y, const char* txt, int scol = 0, int spos = 0)
    { drawflag |= 1; texts.insert(vtext(x, y, scol, spos, txt)); }
  void add_ellipse(int x, int y, int xl, int yl, int ecol = 0);
  void add_ellipse(double x, double y, double xl, double yl, int ecol = 0)
  { add_ellipse(int(x), int(y), int(xl), int(yl), ecol); }
  void del_grid();
  void del_point(int x, int y);
  void del_line(int x1, int y1, int x2, int y2);
  void del_box(int x1, int y1, int x2, int y2);
  void del_string(int x, int y);
  void clear_geom();
  void draw_point(const vpoint& vp) const;
  void draw_line(const vline& vl) const;
  void draw_box(const vbox& vb) const;
  void draw_string(const vtext& vt) const;
void get_marked_points(igrid &out);
}; // class viewport

typedef vector<viewport> viewports;

// __________________________________________________________________________
// viewport.h



#endif
