// Copyright YALI AMIT, January 15, 2001.
#define CARTTREE_C

// __________________________________________________________________________
// Include files

#include "includes.h"

#include <iostream>
#include <iomanip>
#include <fstream>

#include "includes.h"
#include "simpstrg.h"
#include "carttree.h"
#include "trainstat.h"
#include "viewgrid.h"
#include "globals.h"
#include "ltatree.h"
#include "pixel.h"
#include "image.h"
#include "grid.h"
// __________________________________________________________________________
// Typedefs, etc.

//static impfunction impfn = &histogram::entropy;
int cartprestop(const cartiter& t, int);
void cart_best_question(const cartiter& diter, dgrids& cartims,
			ucgridss& lands, int& bestpix, double& bestthresh);
void random_select(lgrid& randgrid, int max, int num);
void square_image(const image& im, dgrid& imsquared);
void cluster_landmarks(lmarklist& lands, int clustsize, 
		       int numtypes, int& total);
void nistprocess(image &im);
void  showtags(const cartiter& diter, dgrids& cartims);
int check_bgd(ucgrid &in);
extern ucgrids STAGS;
extern double ratios[];
extern ltatrees ltrees;
extern carttrees ctrees, ftrees;
extern igrid label_list;
// __________________________________________________________________________
int  codesize2;


void make_tag_tree()
{

  
  dgrids cartims;
  ucgridss lands;

  codesize2=codesize/2;
  cartsizex=codesize;
  cartsizey=codesize;
  cartims.init(numdigits * numperts);

  int count=0;
  for (int i=0; i<numdigits; i++)
    {
      image im;
      im.read(i);
      if (nist)
	nistprocess(im);

      gridfun( (im(i,j)>0),im);      

      //view_normalized(im);
      for (int p=0; p<numperts; p++)
	{
	  int k= (lrand48() % (im.dimx()-codesize));
	  int l= (lrand48() % (im.dimy()-codesize));



	  ucgrid uc(codesize,codesize);
	  uc.subcopy(k, k+codesize, l, l+codesize, im, 0, 0);  


          if (check_bgd(uc))
	    {

	      cartims[count].init(codesize, codesize);
              gridfun(uc(i,j), cartims[count]);
              //cartims[count].dump();
	      //cout << endl;
             	      count++;
	    }
	}
    }
	  	
  cout << "count " << count << endl;
  ofstream ofs("cartraw");
  if (!ofs)
    {
      cerr << "Unable to open cartraw\n";
      exit(1);
    }
  write_dgrids(cartims, 0, count, "cart");

  numtreedata=count;

  impmeasure="unsup";
  set_impurity_func("unsup",0);

  label_list.init(numtreedata);
  label_list.clear(0);
  lgrid dlist(numtreedata);
  for (int i = 0; i < numtreedata; ++i)
    dlist(i) = i;

  carttree ftree;
  ftree.grow(dlist, cartims, lands);
  ftree.write(simpstrg("codes.tree"));
}


void make_tag_files()
{
  void detecttags(image &im, landmarks &ltalands, carttree &tagtree);

  lmarkslist ltalands(numdigits);

  carttree tagtree;
  tagtree.read("codes.tree");

  cartsizey=cartsizex=codesize=tagtree.cartsx;
  codesize2=codesize/2;

  for (int i=0; i<numdigits; i++)
    {

      image im;
      im.read(first+i);
      nistprocess(im);
      detecttags(im, ltalands[i], tagtree);
      //ltalands[i].dump();
    }

  write_lands(newlanddir, numdigits, ltalands);


}




void  showtags(const cartiter& diter, dgrids& cartims)
{
  if (showedges)
    {
      image im(cartsizex, cartsizex);
      im.clear(numcolors/2-1);
      viewport vptag(im);

      cartiter tt=diter;
      while (tt.parent())
	{       
	  int currpix = tt.parent()->pixindex;
	  if (tt.yn()==1)
	    vptag.add_point(currpix % cartsizex,currpix/ cartsizex, numcolors/8, pixsize-1);
	  else
	    vptag.add_point(currpix % cartsizex,currpix / cartsizex, 7*numcolors/8, pixsize-1);
	  tt=tt.parent();
	}
       vptag.view();
    }
  if (diter.depth()==showedges)
    {


      simpstrg ff = simpstrg("NODE")+diter.history();
      ofstream ofs(ff);
      for (int k=0; k < diter->datalist.dimx(); k++)
	{
	  for (int l=0; l<codesize*codesize; l++)
	    {
	      ofs << (int)(cartims[diter->datalist(k)][l])<< " ";
	    }
	  ofs << endl;
	}
      ofs.close();
      simpstrg comm = simpstrg("sort ")+ff+simpstrg(" | uniq -c | sort | tail -1 > us")+ff;
      system(comm);
      comm = simpstrg("us")+ff;
      ifstream ifs(comm);
      image ucim(codesize,codesize);
      int j;
      ifs >> j;
      for (int l=0; l<codesize*codesize; l++)
	{    
	  ifs >> ucim[l];
	}
      view_normalized(ucim); 
      
    }
}


void carttree::grow(const lgrid& datalist, dgrids& cartims, ucgridss& lands)
{
  // Initialize the tree
  maketop();


  // Fill in the histogram information for the top node
  carthist& dhist = top()->hist;
  dhist.clear();
  int i;
  for (i = 0; i < numtreedata; ++i)
    {
   if (label_list.dimx()==0)
    ++dhist[MIN((datalist(i) / numtrain), numclass)];
   else
     {
       ++dhist[label_list(i)];
     }
    }


  // Relic
  top()->datalist.init(numtreedata);
  for (i = 0; i < numtreedata; ++i)
    top()->datalist(i) = datalist(i);

  // Deepen the tree
 
  deepen(top(), cartims, lands);

} // carttree::grow

// __________________________________________________________________________

int carttree::read(const simpstrg& treename)
{

  Ifstream(ifs, treename);
  if (!ifs)
  {
    cerr << "Unable to read cart tree [" << treename << "]\n";
    exit(1);
  }

  char buf[7];
  ifs.read(buf, 6);
  buf[6] = 0;
  if (strcmp(buf, "CART10") != 0 && strcmp(buf, "CART11") != 0)
  {
    cerr << treename << " is not a cart tree\n";
    exit(1);
  }

  
  if (memcmp(buf, "CART11", 6) == 0)
    {
      varread(ifs, carttype);	 
      if (cartstyle<4)
	{
	  varread(ifs, cartsprd);
	  varread(ifs, cartsx);
	  varread(ifs, cartsy);
	  points.read(ifs);
	}
      else
	{
	  cartsx=cartsizex;
	  cartsy=cartsizey;
	}
    }
  if (!ifs.eof())
    {
      maketop();
      readnodes(top(), ifs);
      updatehists(top());
    }
  
  return 1;

} // carttree::read

// __________________________________________________________________________

void carttree::readnodes(const cartiter& curr, ifstream& ifs)
{

  varread(ifs, curr->pixindex);
  varread(ifs, curr->pixthresh);

  if (curr->pixindex == -1)
  {
    curr->hist.read(ifs);
  }
  else
  {
    curr.makechildren();
    readnodes(curr.no(), ifs);
    readnodes(curr.yes(), ifs);
  }

} // carttree::readnodes

// __________________________________________________________________________

void carttree::updatehists(const cartiter& curr)
{
  if (curr.branch())
  {
    updatehists(curr.no());
    updatehists(curr.yes());
    for (int i = 0; i < curr->hist.dim(); ++i)
      curr->hist[i] = curr.no()->hist[i] + curr.yes()->hist[i];
  }

} // carttree::updatehists

// __________________________________________________________________________

void carttree::write(const simpstrg& treename)
{

  Ofstream(ofs, treename);
  if (!ofs)
  {
    cerr << "Unable to write to treefile [" << treename << "]\n";
    exit(1);
  }

  ofs << "CART11";
  varwrite(ofs, carttype);
  varwrite(ofs, cartspread);
  varwrite(ofs, cartsizex);
  varwrite(ofs, cartsizey);
  points.write(ofs);
  writenodes(top(), ofs);

} // carttree::write

// __________________________________________________________________________

void carttree::writenodes(const cartiter& curr, ofstream& ofs)
{
  if (!curr)
    return;

  varwrite(ofs, curr->pixindex);
  varwrite(ofs, curr->pixthresh);
  if (curr.leaf())
  {
    curr->hist.write(ofs);
  }
  else
  {
    writenodes(curr.no(), ofs);
    writenodes(curr.yes(), ofs);
  }

} // carttree::writenodes
// __________________________________________________________________________

void carttree::dump(ostream& os)
{
  switch (carttype)
  {
   case 0:
    cout << "CART on pixel values\n";
    break;
   case 1:
    cout << "CART on pixel values and differences\n";
    break;
   case 2:
    cout << "CART on edge locations\n";
    break;
   case 3:
    cout << "CART on lta locations\n";
    break;
  }
  cout << " cartspread " << cartsprd << endl;
  cout << " dims " << cartsx << " " << cartsy << endl;
  points.dump();
  os.setf(ios::fixed, ios::floatfield);

  stats = new carttreestats;
  dumpnodes(top(), os);

  calc_stats(top());
  stats->dump();

} // carttree::dump

// __________________________________________________________________________

void carttree::calc_stats(const cartiter& diter)
{
  if (diter.leaf())
  {
    stats->total += (int)(diter->hist.total());
    stats->correct += (int)(diter->hist[diter->hist.mode()]);
    stats->wtdepth += (int)(diter->hist.total() * diter.depth());
    stats->wtent += diter->hist.entropy() * diter->hist.total();
  }
  else
  {
    calc_stats(diter.no());
    calc_stats(diter.yes());
  }

} // carttree::calc_stats
// __________________________________________________________________________

void carttree::dumpnodes(const cartiter& d, ostream& os)
{
  int i;
  for (i = 0; i < 50; ++i) 
    os << "=";
  os << "\nDepth: [" << setw(2) << d.depth()
    << "], History: [" << d.history() << "]\n";

  if (d.branch())
  {
    os << "Pixel [" << d->pixindex % cartsizex << " " << d->pixindex / cartsizex;
    if (carttype <= 1)
      os << "], Threshold [" << d->pixthresh << "]\n";
    else
      os << "], Landmarktype [" << int(d->pixthresh + FUZZ) << "]\n";
  }

  os << "Data at node: " << d->hist.total() << endl;
  os << "Curr:  ";
  if (d.leaf())
    os << " Terminal ";
  for (i = 0; i < d->hist.dim(); ++i)
    os << i << "[" << d->hist[i] << "] ";
  os << endl;

  if (d.branch())
  {
    os << "No < thr.:  ";
    for (i = 0; i < d.no()->hist.dim(); ++i)
      os << i << "[" << d.no()->hist[i] << "] ";
    os << endl;
    os << "Yes > thr.: ";
    for (i = 0; i < d.yes()->hist.dim(); ++i)
      os << i << "[" << d.yes()->hist[i] << "] ";
    os << endl;

  }

  if (d.branch()) 
  {
    dumpnodes(d.no(), os);
    dumpnodes(d.yes(), os);
  }

} // carttree::dumpnodes

// __________________________________________________________________________

void carttree::drop(dgrids& cartims, ucgridss& reglandss, dgrid& agg) const
{
  

  for (int i = 0; i < agg.dimx(); ++i)
  {

    cartiter terminal;
    dropdata(cartims[i], reglandss[i], terminal);
    double total = terminal->hist.total();
    //cout << terminal.history() << endl;
    if (total > 0)
    {
      for (int j = 0; j < terminal->hist.dim(); ++j)
      {
	agg(i, j) += terminal->hist[j] / total;
      }
    }
  }
}

// __________________________________________________________________________

void carttree::drop(dgrid& cartimage, ucgrids& reglands, dgrid& agg) const
{
  cartiter terminal;

  dropdata(cartimage, reglands, terminal);
  double total = terminal->hist.total();
  if (total > 0)
  {
    for (int j = 0; j < MIN(agg.dimy(),terminal->hist.dim()); ++j)
    {
      agg(0,j) += terminal->hist[j] / total;
    }
  }
}

// __________________________________________________________________________

void carttree::dropdata(dgrid& cartim, ucgrids& reglands,
			cartiter& terminal) const
{
  terminal = top();
  while (terminal.branch())
  {
    int currpix = terminal->pixindex;
    double currthresh = terminal->pixthresh;
    int regtype = int(currthresh + FUZZ);
    if ((carttype <= 1 && cartim[currpix] > currthresh) ||
	(carttype > 1 && reglands[regtype][currpix] > 0))
      terminal = terminal.yes();
    else
      terminal = terminal.no();
      
  }

} // carttree::dropdata

// __________________________________________________________________________

void carttree::deepen(const cartiter& diter, dgrids& cartims, ucgridss& lands)
{
  // Set up an alias for the node data and find out how much data there is
  int numnodedata = diter->datalist.dimx();
  cout << "numdat " << numnodedata << endl;
  if (showedges)
    showtags(diter, cartims);

  // cout << "history " << diter.history() << endl;
  int bestpix = -1;
  double bestthresh = 0;
  // What we do depens on whether or not we're at a leaf node or not
  if (diter.leaf())
  {
    // Test the pre-stopping criterion
    if (cartprestop(diter, mindata))
      {
	cout << "stopping " << endl;
      return;
      }

    // Get a new question for this node
    cart_best_question(diter, cartims, lands, bestpix, bestthresh);

    // Prepare to ask the new question
    diter.makechildren();

    diter->pixindex = bestpix;
    diter->pixthresh = bestthresh;
  }
  else
  {
    // Clear the histograms and facelists of the children
    diter.yes()->datalist.init();
    diter.yes()->hist.clear();
    diter.no()->datalist.init();
    diter.no()->hist.clear();
  }

  // Ask the node question of all the data at the current node
  int numyes = 0;
  cgrid answers(numnodedata);
  int i;
  for (i = 0; i < numnodedata; ++i)
  {
    if (carttype <= 1)
      answers(i) = 
	(cartims[diter->datalist(i)][bestpix] > bestthresh);
    else
    {
      int regtype = int(bestthresh + FUZZ);
      answers(i) =
	(lands[diter->datalist(i)][regtype][bestpix] > 0);
    }
    numyes += (answers(i) > 0);
  }

  // Set up the children
  diter.no()->datalist.init(numnodedata - numyes);
  diter.yes()->datalist.init(numyes);

  int noind = 0, yesind = 0;
  for (i = 0; i < numnodedata; ++i)
  {
    if (answers(i))
    {
      diter.yes()->datalist(yesind) = diter->datalist(i);
     if (label_list.dimx()==0)
      ++diter.yes()->hist[MIN((diter->datalist(i) / numtrain), 
				   numclass)];
     else
      ++diter.yes()->hist[label_list(diter->datalist(i))];

      ++yesind;
    }
    else
    {

      diter.no()->datalist(noind) = diter->datalist(i);
     if (label_list.dimx()==0)
      ++diter.no()->hist[MIN((diter->datalist(i) / numtrain),
				  numclass)];
     else
       ++diter.no()->hist[label_list(diter->datalist(i))];
      ++noind;
    }
  }

  assert((yesind == numyes) && (yesind + noind == numnodedata));

  deepen(diter.no(), cartims, lands);
  deepen(diter.yes(), cartims, lands);

}  // carttree::deepen

// __________________________________________________________________________

void cart_best_question(const cartiter& diter, dgrids& cartims,
			ucgridss& lands, int& bestpix, double& bestthresh)
{
  int i, j;
  static ucgrid processmask;
  static int inited = 0;
  if (!inited)
  {
    inited = 1;
    make_procmask(processmask, masktype, cartsizex, cartsizey);
  }

  int ndata = diter->datalist.dimx();

  int numpixels;
  if (cartstyle <= 1)
    numpixels = cartims[0].dimx() * cartims[0].dimy();
  else
    numpixels = lands[0][0].dimx() * lands[0][0].dimy();

 
  // Set up the cart thresholds
  dgrids values;
  if (cartstyle <= 1) {
    values.init(numpixels);
    if(numlevels>1)
      {
	for (i = 0; i < numpixels; ++i)
	  {
	    values[i].init(ndata);
	    for (j = 0; j < ndata; ++j)
	      values[i](j) = cartims[diter->datalist(j)][i];
	    
	    values[i].sort();
	  }
      }
  }

  // Choose the random questions to ask
  int maxquestions;
  if (cartstyle <= 1)
    maxquestions = numpixels * numlevels;
  else
    maxquestions = numpixels * lands[0].dim();
 


  lgrid pixind;
  random_select(pixind, maxquestions, numquesask);

  // Find the best question
  double bestent = 1000000;
  for (i = 0; i < numquesask; ++i)
  {
    int currpix;
  
    if (cartstyle <= 1)
      currpix = pixind(i) / numlevels;
    else
      currpix = pixind(i) / lands[0].dim();
  
    int jj = currpix % cartsizex;
    int kk = currpix / cartsizex;
    if (!processmask(jj, kk))
	continue;

    carthist no, yes;
    double currthresh;
    if (cartstyle <= 1)
    {
      if (numlevels > 1)
      currthresh = values[currpix][(ndata * (pixind(i) % numlevels + 1)) / (numlevels+1)];
      else
	currthresh=0;
      for (j = 0; j < ndata; ++j)
      {
	if (cartims[diter->datalist(j)][currpix] > currthresh)
	  {
          if(label_list.dimx()==0)
	    ++yes[MIN((diter->datalist(j) / numtrain), numclass)];
          else
            ++yes[label_list(diter->datalist(j))];
	  }
	else
	  {
	    if(label_list.dimx()==0)
	      ++no[MIN((diter->datalist(j) / numtrain), numclass)];
            else
              ++no[label_list(diter->datalist(j))];
	  }
      }
    }
    else
    {
      currthresh = pixind(i) % lands[0].dim();
	

      for (j = 0; j < ndata; ++j)
      {

	if (lands[diter->datalist(j)][(int)(currthresh)][currpix] > 0)
	  {
	    if(label_list.dimx()==0)
	      ++yes[MIN((diter->datalist(j) / numtrain), numclass)];
	    else
	      ++yes[label_list(diter->datalist(j))];
	  }
	else
	  {
	    if(label_list.dimx()==0)
	      ++no[MIN((diter->datalist(j) / numtrain), numclass)];
            else
              ++no[label_list(diter->datalist(j))];
	  }
      }
    }
    double newent;

      newent = ave_impurity(yes, no);

    // double newent = no[0];
    if (newent < bestent)
    {
      bestent = newent;
      bestpix = currpix;
      bestthresh = currthresh;
    }

  }

} // cart_best_question

// __________________________________________________________________________

void carttree::clearhists(const cartiter& curr)
{
  if (curr.branch())
  {
    clearhists(curr.yes());
    clearhists(curr.no());
  }
  curr->hist.clear();

} // carttree::clearhists

// __________________________________________________________________________

void carttreestats::dump() const
{
  cout.setf(ios::fixed, ios::floatfield);

  cout << "\nAverage classification rate: " << setprecision(2) 
    << 100.0 * correct / total << endl;
  cout << "Average depth: " << 1.0 * wtdepth / total << endl;
  cout << "Average entropy: " << wtent / total << endl;

} // carttreestats::dump

// __________________________________________________________________________
// Return 1 if we can not continue growing the tree from node 't'.

int cartprestop(const cartiter& t, int)
{

  // Stop if tree depth is too large
  if (t.depth() >= maxdepth)
    {
      //      cout << " depth " << endl;
    return 1;
    }

  // Stop if the second mode count is too low
  if (t->hist[t->hist.mode2()] < modetwo)
    {
      //  cout << " mode2 " << endl;
    return 1;
    }
  // Stop if node is relatively pure
   if (t->hist.pure(purepar))
     {     
       //cout << " pure " << endl;
       return 1;
     }
  // Stop if not enough data to continue
  if (t->hist[t->hist.mode()] < mindata)
    {
      //      cout << " mode " << endl;
      return 1;
    }
  return 0;

} // cartprestop





// __________________________________________________________________________
// carttree.C













