// Copyright YALI AMIT, January 15, 2001.


// __________________________________________________________________________
// Include files

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>
#include "includes.h"
#include "simpstrg.h"
#include "viewport.h"
#include "carttree.h"
#include "image.h"
#include "viewgrid.h"
#include "trainstat.h"
#include "ltatree.h"
#include "globals.h"
static int compar(const void *a, const void* b);
void boost_process(lmarkslist &lmarks, ltatree &ltree, dgrids &agg);
void show_bad_image(int i, dgrid &tempagg);
void nistprocess(image &im);
void detecttags(image &im, landmarks &ltalands, carttree &tagtree);
void extract_landmarks(image& im, ltatrees& ltrees, int numltas,
			 landmarks& edgelands, landmarks& ltalands);
void comput_meanvar(dgrid &mu, dgrid &mu2, igrid &classcount, 
		    double &totvar, double &meanth, int numtrees);
void run_trees(landmarks &lmark, ltatrees &ltrees, dgrid &agg, dgrid &betas, int ft, int lt);
int process_agg(dgrid &agg, dgrid &tempagg, int tr, int rnum, dgrid &CONF);
void fill_STAGS(const landmarks& lm, int spread = 0, int numt = -1);
void empty_STAGS(const landmarks& lm);
dgrid weits;
extern int codesize2;
ucgridss EMATSS;
extern ucgrids STAGS;
int NBG;
igrid trth;

void read_weits()
{
  dgrid ww;
  for (int k=0; k<10; k++)
    {
      simpstrg fn=simpstrg("weits")+ltos(k);
      ifstream ifs(fn);
      ww.read(ifs);

      //cout << ww.dimx() << endl;
      ww.dump();
    }
}

void nearest_neighbors()
{
  
  NBG=STAGSdim*STAGSdim;
  EMATSS.init(numtreedata);
  trth.init(numtreedata);
  
  for (int i = 0; i < numtreedata; ++i)
    {
      
      landmarks lm;
      lm.read(i);
      trth[i]=lm.truth();
      
      EMATSS[i].init(lm.total_types());
      fill_STAGS(lm,spread,-1);
      for (int a=0; a<EMATSS[i].dim(); a++)
	{
	  EMATSS[i](a)=STAGS(a);
	}
      
      empty_STAGS(lm);
    }
  int check_nn();
  int cor=0;
  landdir=dataset; 
  for (int j=first; j<first+numdata;j++)
    {
      landmarks lm;
      lm.read(j);
      fill_STAGS(lm,spread);
      int cl=check_nn();
      empty_STAGS(lm);
      cor += (cl==lm.truth());
      //      cout << j << " Truth " << lm.truth() << " NN " << cl << endl;
    }
  cout << " CORR " << cor << " out of " << numdata << endl;
  
}   

int check_nn()
{
  int olddis=NBG*8;
  int oldcl=-1;
  igrid disvec(numdigits);
  igrid clvec(numdigits);
  disvec.clear(olddis);
  clvec.clear(-1);
  for (int i=0; i<numtreedata; i++)
    {
      int dis=0;
      for (int e=0;e<8; e++)
	for (int x=0; x<NBG; x++)
	  {
	    dis+=(STAGS[e](x)!=EMATSS[i][e](x));
	  }
      int done=0;
      for (int k=0; k<numdigits && !done; k++)
	{
	  if (dis<disvec(k))
	    {
	      for (int l=numdigits-1; l>k; l--)
		{
		  disvec(l)=disvec(l-1);
		  clvec(l)=clvec(l-1);
		}
	      disvec(k)=dis;
	      clvec(k)=trth[i];
	      done=1;
	    }
	}
      /*
      if (dis<olddis)
	{
	  olddis=dis;
	  oldcl=trth[i];
	}
      */
    }
  igrid cls(numclass);
  cls.clear();
  for (int t=0; t<numdigits; t++)
      cls[clvec(t)]++;

  int vt=0, clt=0;
  for (int c=0; c<numclass; c++)
    {
      if (cls(c)>vt)
	{
	  vt=cls(c);
	  clt=c;
	}
    }
  return(clt);

}


// Training classification trees
void grow_class(int numtrees)
{
  cout << useques[1] << " " << useques[2] << " " << useques[3] << " " << useques[4] << endl;
  igrid rlist;
  
 

  int fls=0;
  // Read training data indices from a listfile
  ifstream ifs(listfile);
  if (ifs)
    {
      int dd;
      numtreedata=0;
      fls=1;
      while (ifs>>dd)
	numtreedata++;
      ifs.close();
      ifs.open(listfile);
    }
	
    image im;
    cout << numtreedata << endl;    

    lgrid datalist(numtreedata);
    datalist.clear(-1);


    anginc = 2 * PI / numang;
    reltolerance = anginc;
    int numloops=1;
    if (firstclass>0)
      {
	// Each loop is making trees of one class against all the rest.
	numloops=numclass;
	// An extra `everything else class.
	numclass++;
	ltahistsize=numclass;
      }
    weits.init(numtreedata);

    lmarkslist lmarks;
      lmarks.init(numtreedata);
      // if (spread!=0)
	EMATSS.init(numtreedata);
    dgrids agg(numtreedata);
   



    igrid trth(numtreedata);
    for (int i = 0; i < numtreedata; ++i)
      {
	int ii;
	if (!fls)
	  ii=i;
	else
	  ifs >> ii;

	datalist(i)=i;
	//if (spread==0)
	// lmarks[i].read(ii+first);
	//else if (spread!=0)
	  {
	    landmarks lm;
	    lm.read(ii+first);
	    trth[i]=lm.truth();
	    lmarks[i].set_truth(trth[i]);
	    EMATSS[i].init(lm.total_types());
	    fill_STAGS(lm,spread,-1);
	    for (int a=0; a<EMATSS[i].dim(); a++)
	      {
		EMATSS[i](a)=STAGS(a);
	      }
	    
	    empty_STAGS(lm);
	  }
      }
    
    
    //	seed = time(0);    
    srand48(seed);

    for (int l=0; l<numloops; l++)
      {
	// Read in weit files if such exist.
	// Allows you to run boosting with intermediate stops
	// weights are preserved. 
	simpstrg wt=simpstrg("weits")+ltos(l);
	ifstream wfs(wt);
	if (!wfs)
	  weits.clear(1.);
	else
	  {
	    weits.read(wfs);
	    if (weits.dimx()!=numtreedata)
	      {
		cout << " weit dim not equal number of data points "<< endl;
		weits.init(numtreedata);
		weits.clear(1);
	      }
	  }

	// Relabel if one class against the rest
	if (firstclass>0)
	  {
	    for (int i=0; i<numtreedata; i++)
	      {
		if (trth[i]==l)
		  lmarks[i].set_truth(l);
		else
		  lmarks[i].set_truth(numclass-1);
		
	      }
	  }

	for (int t=firsttree; t<firsttree+numtrees; t++)
	  {

	    ltatree ltree;
	    ltree.tnumang=numang;
	    ltree.tnear=near;
	    ltree.locationx=-1;
	    ltree.locationy=-1;
	    if (althistsize>0)
	      ltree.locationy=althistsize;     

	    // Grow the tree with current weits
	    ltree.grow(datalist,lmarks);
	    // Update weits in boosting	    
	    if (boost)
		boost_process(lmarks, ltree, agg);

	    simpstrg currtree = carttreedir+simpstrg("/tree") + ltos(t);     
	    ltree.write(currtree);
	    
	  }
	firsttree+=numtrees;
	ofstream wofs(wt);
	weits.write(wofs);
      }

    
}

 
void boost_process(lmarkslist &lmarks, ltatree &ltree, dgrids &agg)
{
  double correct=0;
  ucgrid corrlist(numtreedata);
  corrlist.clear();
  for (int i=0; i<numtreedata; ++i)
    { 

      ltaiter terminal;
      igrid insts;
      if (weits(i)>0)
	{
	  ltree.dropdata(lmarks[i],terminal,insts, i);
	  
          int md=-1; 
	  if (boost==1 || boost==3 || boost==4 )
	    {
	      md = terminal->hist.mode();
	    }
	  else if (boost==2)
	    {
	      double tot=terminal->hist.total();
	      md=-1; 
	      double oldval=0;
	      for (int j=0; j<terminal->hist.dim(); j++)
		{
		  
		  agg[i](0,j)+=(terminal->hist[j]/tot);
		  if (agg[i](0,j) >= oldval)
		    {
		      oldval=agg[i](0,j);
		      md=j;
		    }
		}
	    }
	  if (lmarks[i].truth(nist)==md)
	    {
	      correct+=weits(i);
	      corrlist(i)=1;
	    }
	}
    }

   double error = 1.-correct/numtreedata;
   double beta=alpha*.5*log( (1-error)/error);
   if (boost==3)
    cout << "beta = " << beta << endl;
   //   cout<<" Weighted Training Error " << error << " " << alpha/error << endl;
   double weitsum=0;
   for (int i=0; i<numtreedata; ++i)
     {
     if (error > 0)
       {
	 if (!corrlist(i))
	   {
	     if (boost==1)
	       weits(i)*= 1/error;
	     if (boost==3)
	       weits(i)*=(1-error)/error;
	     if (boost==2)
	       weits(i) = 1 + alpha/error;
	   }
	 if (boost==4)
	   {
	   if (corrlist(i))
	     weits(i) = MAX(0, weits(i)-error);
	   else
	     weits(i) = MAX(0, weits(i)+1-error);
	   }
	
       }
     
     //     if ( weits(i)<alpha1)
     //  weits(i)=0;

	 weitsum+=weits(i);
     }
   //   cout << "weitsum " << weitsum/(numtreedata) << endl;

   for (int i=0; i<numtreedata; ++i)
     {
       weits(i)=numtreedata*weits(i)/weitsum;
     }

   
}





static int compar(const void *a, const void* b)
{
  return (*((const double*) a) < *((const double*) b));
}

void drop_class(int numtrees)
{



  ltahistsize=numclass;

  lmarkslist lmarks(numtreedata);
  dgrids agg(numtreedata);
  for (int i = 0; i < numtreedata; ++i)
    {

      lmarks[i].read(i);
      agg[i].init(2,numclass);
      agg[i].clear();
    }
  
  int showi=first;


  // showall=2 : show all images in the terminal node reached by showi in
  // tree showt
  int showall=dum1;

  

  ltaiter terminal_bak=NULL;
  for (int t=0; t<numtrees; t++)
    {
      ltatree ltrees;
      simpstrg currtree = carttreedir + simpstrg("/tree") + ltos(t+firsttree);
      simpstrg com = simpstrg("gunzip ")+currtree;
      system(com);
      if (showall)
	opt=2;
      if(!ltrees.read(currtree))
	{
	  cerr << " Cant read tree " << currtree << endl;
          exit(1);
	}
      com = simpstrg("gzip ")+currtree;
      system(com);
      int treecorrect=0;
      
      for (int i=0; i<numtreedata; ++i)
	{ 
	  if (showall==1 && i != first)
	    continue;
          if (showall==2 && i < first )
            continue;
	  ltaiter terminal;
	  igrid insts;
	  anginc = 2 * PI / ltrees.tnumang;
	  reltolerance = anginc;
	  near = ltrees.tnear;
                
	  ltrees.dropdata(lmarks[i],terminal,insts);
	  
	  
	  // For graphic purposes
	  if ( showall && 
	       (i==showi || (showall==2 && terminal==terminal_bak))
	       && t<numcarts)
	    {
	      void show_graph(int i, igrid &insts, ltaiter& terminal, landmarks &lmarks);
	  
	      show_graph(i,insts,terminal, lmarks[i]);
	      if (i==showi)
		{
		  terminal_bak=terminal;                         
                  for (int k=0; k<terminal->hist.dim(); k++)
		    {
                      if (terminal->hist[k]>0)
                        cout << k << " " << terminal->hist[k] << endl;
		    }
		}
	    }
	  double tot=terminal->hist.total();
	  if (lmarks[i].truth(nist)==terminal->hist.mode())
	    treecorrect++;
	  for (int j=0; j<terminal->hist.dim(); j++)
	    {
		  if (tot)
		    agg[i](0,j)+=(terminal->hist[j]/tot);
		      if (t==0)
			agg[i](1,j)=j;
	    }

	}
      double correct=0;
      for (int i=0; i<agg.dim(); i++)
	{
	  dgrid temp=agg[i];
	  temp.sort_columns(numclass,compar);
	  if (temp(1,0)==lmarks[i].truth(nist))
	    correct++;

	}
      
      cout << t << " Tree class rate " << treecorrect << " " << double(treecorrect)/double(numtreedata) << endl;
      cout << t << " Correct " << correct <<  " " << double(correct)/double(numtreedata) << endl;       
           
    }
  
  
}
   
void comput_meanvar(dgrid &mu, dgrid &var, igrid &classcount, 
		    double &totvar, double &meanth);
void read_pert(image &im,int rnum, landmarks &lmark, carttree &tagtree, double alp, int rnd);

void drop_class_quick(int numtrees)
{

  if (numloc>0)
    {
      numperts=1;
      if (alpha>0)
	numperts*=numloc;
      if (skew>0)
	numperts*=numloc;
      if (rotate>0)
	numperts*=numloc;
      if (shift>0)
	numperts*=(numloc*numloc);
      cout << numperts << endl;
    }

  cout << " NUMCLASS " << numclass << endl;
  int trnumcl=numclass;
  // `everything else class
  if (firstclass>0)
    numclass++;

  if (althistsize==0)
    ltahistsize=numclass;
  dgrid betas(numtrees);
  betas.clear(1);
  ifstream ifs;
  double totbeta=0;
  // Read in TREE weits from classical boosting
  if (boost==3)
   {
    ifs.open("betas");
    for (int k=0; k<numtrees; k++)
      {
	ifs >> betas(k);
	totbeta+=betas(k);
      }
   }

  igrid classcount(numclass);
  classcount.clear();
  dgrid mu(numclass,numclass);
  dgrid mu2(numclass,numclass);
  mu.clear(); mu2.clear();

  carttree tagtree;
  if (landdir=="")
    {
      if (codesize>0)
	{
	  cout << " Reading codes tree " << endl;
	  tagtree.read("codes.tree");
	  codesize2=codesize/2;
	  cartsizex=codesize;
	  cartsizey=codesize;
	}
   }


  dgrid CONF, CONF2;
  if (numclass<20)
    {
      CONF.init(numclass, numclass);
      CONF2.init(numclass, numclass);
      CONF.clear();
      CONF2.clear();
    }

 ltatrees ltrees(numtrees+fnumcarts);

 for (int t=firsttree; t<firsttree+numtrees+fnumcarts; t++)
   {
     simpstrg currtree = carttreedir + simpstrg("/tree") + ltos(t);
     if(!ltrees[t-firsttree].read(currtree))
       {
	 cerr << " Cant read tree " << currtree << " Check if it is gzipped " << endl;
	 exit(1);
       }
   }
 dgrid fy;

 if (showtriangles>=0)
   cout << numtreedata << endl;
 double correct=0, correct2=0;
 double correctpl=0;
 double correctconf=0, correctconf2=0;
 double cornonaccept=0, nonaccept=0;
 double noconf=0;
 double crpl=0;
 int co=0;

 // Test data in a listfile 
  ifstream ifsb;
  if (listfile!="")
    ifsb.open(listfile);
  for (int i=0; i<numtreedata; ++i)
    { 
      

      landmarks lmark;
      co++;
      int rnum=i+first;
      if (listfile!="")
	ifsb >> rnum;
      if (landdir!="")
	lmark.read(rnum);
      else
	{
	  //cout << rnum << endl;
	  image im;
	  im.read(rnum);
	  read_pert(im, rnum, lmark, tagtree, 0, 0);
	}
      
      classcount(lmark.truth(nist))++;     
      //if (spread !=0)
	fill_STAGS(lmark,spread,-1);
      dgrid agg(2,numclass);
      dgrid tagg(2,numclass);
      agg.clear(); tagg.clear();
      for (int j=0; j<numclass; j++)
	{
	  tagg(1,j)=j;	agg(1,j)=j;
	}
      
      
      // Regular dumping down all trees
      run_trees(lmark, ltrees, agg, betas, 0, numtrees);

      // Update statistics variance of aggregate
      for (int l=0; l<numclass; l++)
	{
	  mu2(lmark.truth(nist),l) += agg(0,l)*agg(0,l);
	  mu(lmark.truth(nist),l) += agg(0,l);
	}
      dgrid sortagg;
      int cor;
      // Process the aggegate
      cor=process_agg(agg, sortagg, lmark.truth(nist), rnum, CONF);

      correct+=(cor==1);
      correct2+=(cor!=0);
      
      // Checking for confidence level of classification
      // and Reclassifying perturbed versions.
      if (dum2>0 || numperts>0)
	{
	  int cl1=sortagg(1,0);
	  int cl2=sortagg(1,1);
	  int accept=(sortagg(0,0)>alpha2*sortagg(0,1));
	  noconf+=accept;
	  cor= (accept && (sortagg(1,0)==lmark.truth(nist)));
	  correctconf+=cor;
	  correctconf2+=cor;
	  if (showtriangles==3)
	    {
	      if(!accept)
		cout << " Reject " << lmark.truth() << ":" << rnum << endl;
	      else
		{
		  if (CONF2.dimx()>0)
		    CONF2(lmark.truth(nist),sortagg(1,0))++;
		}
	    }
	  // Here we run an additional group of trees on the rejected points
	  if (!accept && numperts==0 && fnumcarts>0)
	    {
	      
	      run_trees(lmark, ltrees, agg, betas, numtrees, numtrees+fnumcarts);
	      //if (spread!=0)
		empty_STAGS(lmark);
	      cor=process_agg(agg, sortagg, lmark.truth(nist), rnum, CONF);
	      correctconf2+=(cor==1);
	      cornonaccept+=(cor==1);
	      if (showtriangles==1)
		{
		  image im;
		  im.read(rnum);
		  view_normalized(im);
		}
	      nonaccept++;
	    }
	  // Here we perturb and rerun the same trees numperts times.
	  else if (!accept && numperts>0)
	    {
	      //if (spread !=0)
		empty_STAGS(lmark);
	      int oldcor=cor;
	      image im;
	      im.read(rnum);
	      double oldmarg=0;
	      if (showtriangles==1)
		{
		  cout << rnum << " TRUTH " << lmark.truth(nist) << endl;
		agg.dump();
		}
	      for (int j=0; j<agg.dimy(); j++)
		    agg(0,j)=0;
	      for (int p=0; p<numperts; p++)
		{
		  for (int j=0; j<tagg.dimy(); j++)
		    tagg(0,j)=0;
		  
		  image tim=im;
		  landmarks plmark;
		  read_pert(tim, rnum, plmark, tagtree, /*alpha*/0, -p-1);
		  fill_STAGS(plmark,spread,-1);
		  for (int c=0; c<trnumcl; c++)
		    {
		      for (int r=0; r<10; r++)
			{
			  int rb=r*1000;
			  
			  run_trees(plmark, ltrees, tagg, betas, rb+c*numperclass, rb+(c+1)*numperclass);
			}
		      if (tagg(0,c)>agg(0,c))
			agg(0,c)=tagg(0,c);
		    }
		  empty_STAGS(plmark);

		  
		  if (showtriangles==1)
		    {
		      cout << rnum << " " << p << endl;
		      view_normalized(tim);
		      agg.dump();
		    }
		}
	      cor=process_agg(agg, sortagg, lmark.truth(nist), rnum, CONF);
	      correctconf2+=(cor==1);
	      if (cor!=1)
		cout << " BAD " << rnum << endl;
	    }
	  else
	    {
	      //if (spread!=0)
		empty_STAGS(lmark);
	    }
	  
	  for (int p=0; p<dum2; p++)
	    {
	      if (sortagg(1,p)==lmark.truth(nist))
		{
		  correctpl++;
		  break;
		}
	    }
	}
      else
	{
	  //if (spread != 0)
	    empty_STAGS(lmark);
	}
      
      if (CONF.dimx()>0)
	CONF(lmark.truth(nist),sortagg(1,0))++;
      
      if (!cor && showtriangles>=1)
	{
	  cout << rnum << " Truth " << lmark.truth(nist) << " " << sortagg(1,0) << endl;
	  if (showtriangles==2)
	    show_bad_image(rnum,agg);
	}
      
    }
  
  double totvar=0, meanth=0;
  comput_meanvar(mu, mu2, classcount, totvar, meanth, numtrees);
  
  
  
  // Dump out results.  
  if (showtriangles>=0)  
    {
      cout << "CORRECT " << double(correct)/co << ":"<< correct << ":OUT OF " << co << " (correct2 " << correct2 << " )" << " (top 2 " << crpl << " )" << endl;  
      if (dum2>0)
	{
	  cout << "CORRECT-PL " << dum2 << " " << double(correctpl)/co << " ";  
	  cout << "CORRECT-CONF " << alpha2 << " " << correctconf/noconf << " " << ":" << correctconf << ":" << " Out of " << noconf << " " << endl;
	  cout << " CORRECT-CONF2 iterize "<< fnumcarts << " numperts " << numperts << " alpha " << alpha << " shift " << shift << " skew " << skew << " rotate " << rotate << " " << correctconf2/co << " " << correctconf2 << " out of "<< co << endl;
	  cout << " CORRECT-NONACP " << cornonaccept/nonaccept << " " << cornonaccept << " out of "<< nonaccept << endl;
	  
	}
      cout << meanth;
      cout << " " << totvar << endl;
      if (CONF.dimx()>0)
	{
	  if (showtriangles==3)
	    CONF2.dump();
	  else 
	    CONF.dump();
	}
    }
  
}


int process_agg(dgrid &agg, dgrid &tempagg, int tr, int rnum, dgrid &CONF)
{
  tempagg=agg;

  if ( firstclass<=0)
      tempagg.sort_columns(numclass,compar);
  else if (firstclass>0)
      tempagg.sort_columns(numclass-1,compar);

  int cor=0;
  if (tempagg(1,0)==tr)
    {
      if (tempagg(0,0) > tempagg(0,1) )
	cor=1;
      else
	cor=-1;
    }

  
 
  return(cor);
}



void comput_meanvar(dgrid &mu, dgrid &mu2, igrid &classcount, 
		    double &totvar, double &meanth, int numtrees)
{
  int numcl=numclass;
  if (firstclass>0)
    numcl-=1;
  dgrid var(numcl,numcl);
  var.clear();
  for (int k=0; k< numcl; k++)
    {
      for (int l=0; l< numcl; l++)
	{
	  mu(k,l) = mu(k,l)/classcount(k);
	  var(k,l) = mu2(k,l)/classcount(k) - mu(k,l)*mu(k,l);
	}	  
    }
  
  
  for (int k=0; k<numcl; k++)
    {
      double oldmu=0;
      for (int l=0; l<numcl; l++)
	{
	  totvar += var(k,l);
	  if (mu(k,l) > oldmu && l != k)
	    oldmu=mu(k,l);
	}
      meanth += mu(k,k)-oldmu;
   }
  meanth/=(numcl*numtrees);
  totvar/=(numcl*numtrees*numtrees);
}




void run_trees(landmarks &lmark, ltatrees &ltrees,  dgrid &agg, dgrid &betas, int ft, int lt)
{
  
  int lft=MAX(0,ft);
  for (int t=lft; t<lt; t++)
    {

      ltaiter terminal;
      igrid insts;
      anginc = 2 * PI / ltrees[t].tnumang;
      reltolerance = anginc;
      near = ltrees[t].tnear;
      ltrees[t].dropdata(lmark,terminal,insts, -1);
      
      double tot=terminal->hist.total();
      int mod=terminal->hist.mode();
      //      cout << " N0DE NUM " << terminal->datalist(0) << endl;
      if (vote)
	{
	  if (firstclass>0)
	    {
	      if (mod<numclass-1)
		agg(0,mod)++;
	    }
	  else
	    agg(0,mod)+=alpha*betas(t);
	}
      else
	{
	  if (tot)
	    {
		  for (int j=0; j<terminal->hist.dim(); j++)
		    {
		      //agg(0,j)+=log((1+terminal->hist[j])/(tot+numclass));
		      agg(0,j)+=terminal->hist[j]/tot;
		    }
	    }
	}
    }
}


void read_pert(image &im, int rnum, landmarks &lmark, carttree &tagtree, double alp, int rnd)
{
  ltatrees ptrees;
  landmarks edges;
  int imw=im.dimx();
  if (DIMX>0)
    imw=DIMX;
  if (threshold>0)
      nistprocess(im);
  else
    im.blur_downscale(ratio, sd);
  int x;
  //  view_normalized(im);
  if (threshold==0)
    {
      im.imtrim(-imw, x,x,x,x,x,x);
      im.imslant(1);
      im.imtrim(-imw, x,x,x,x,x,x);
    }


  if (rnd>0)
    {
      im.std_process(skew, rotate, slant, scale, x,x,x,x,x,x);
      im.imtrim(-imw, x,x,x,x,x,x);
      if (shift)
	{
	  int sx=int(rint(drand48()*2*shift-shift));
	  int sy=int(rint(drand48()*2*shift-shift));
	  image nim=im;
	  for (int  x=0; x<im.dimx(); x++)
	    for (int y=0; y<im.dimy(); y++)
	      {
		im.set(x+sx,y+sy,nim(x,y));
	      }
	}
     
    }
  else if (rnd<0)
    {
      double numlc=numloc/2;
      int ind=fabs(rnd+1);
      int al, skw, rt, sfx, sfy;
      double fal=0, fskw=0, frt=0, fsfx=0, fsfy=0;
      if (alpha>0)
	{
	  al= ind % numloc; fal=double(numlc-al)*alpha/numlc;
	  ind=ind/numloc;
	}
      if (skew>0)
	{
	  skw = ind % numloc; fskw=double(numlc-skw)*skew/numlc;
	  ind=ind/numloc;
	}
      if (rotate>0)
	{
	  rt =  (ind % numloc); frt=double(numlc-rt)*rotate/numlc;
	  ind=ind/numloc;
	}
      if (shift>0)
	{
	   sfx = (ind % numloc); fsfx=double(numlc-sfx)*shift/numlc;
	   ind=ind/numloc;
	   int sfy = ind; double fsfy=double(numlc-sfy)*shift/numlc;
	}
      if (fal>0)
	{
	  double dalp=exp(fal);
	  dgrid dim(im.dimx(), im.dimy());
	  gridfun(im(i,j),dim);
	  dim.normalize(0.,1.);
	  gridfun(pow(dim(i,j),dalp),dim);
	  dim.normalize(0,255);
	  gridfun(dim(i,j),im);
	}
      if (fskw!=0 || frt != 0)
	im.std_process( fskw, frt, 0, 0, x,x,x,x,x,x);
      im.imtrim(-imw, x,x,x,x,x,x);
      image nim=im;
      for (int  x=0; x<im.dimx(); x++)
	for (int y=0; y<im.dimy(); y++)
	  {
	    im.set(x+fsfx,y+fsfy,nim(x,y));
	  }
    }
  if (codesize>0)
    detecttags(im, lmark, tagtree);
  else
    {
      extract_landmarks(im, ptrees, numltas, edges, lmark);
      lmark.merge(edges);
    }
  /*
  if (showtriangles==1)
    {
      showedges=2;
      void show_landmarks(image &im, landmarks &edgelands, landmarks & ltalands);
      show_landmarks(im, lmark, lmark);
    }
  */

  lmark.set_truth(im.truth());
}

void show_bad_image(int i, dgrid &tempagg)		    			{
  int x;
  filter_size=4;
  image im;
  im.read(i);
  if (nist)
    nistprocess(im);
  else 
    {
      im.blur_downscale(1., sd);
      im.imtrim(-im.dimx(),x,x,x,x,x,x);
      view_normalized(im);
      im.imslant(1);
      
    }  
 cout << " Image " << i << endl; 
 tempagg.dump();  
  view_normalized(im);
 

}




   
void show_graph(int i, igrid &insts, ltaiter& terminal, landmarks &lmarks)
{
  image im;
  im.read(i);
  if (nist)
    nistprocess(im);
  else
  im.blur_downscale(ratio,sd);

  cout << "Image " << i << endl;
  
  if (basecol==0)
    im.normalize(numcolors-1);
  else
    im.normalize(imrange*numcolors,(1-imrange)*numcolors-1);
  viewport vp(im);

  cout << terminal.history() << endl;

  if (class_tree==2)
    {
      ltaiter term=terminal;
      while (term.parent())
	{

	  ltaiter pterm=term.parent();

                  int x=pterm->que.r1.land2 % cartsizex;
                  int y=pterm->que.r1.land2 / cartsizex;
	      if (term.yn())
		vp.add_point(x,y,basecol,5);
              else
                vp.add_point(x,y,12,5);
	      term=term.parent();
	}
      vp.view();
    }    
      else if (class_tree==1)
	{


	  for (int k=0; k<insts.dimy(); k++)
	    {
	      vp.clear_geom();
	      ltaiter term=terminal;
	      int vt=insts.dimx()/2-1;

	      int p=terminal->vertlist.dimx();
	      while (term.parent())
		{
		  

		  
		  ltaiter pterm=term.parent();
		  if (term.yn())
		    {
		     
		      if (pterm->que.typ==6)
			{
			  p=p-2;
			  vp.add_line(insts(pterm->que.r1.v1*2,k),
				      insts(pterm->que.r1.v1*2+1,k), 
				      insts(vt*2,k),insts(vt*2+1,k), basecol);
			  //vp.add_string(insts(vt*2,k),
			  //		insts(vt*2+1,k),
			  //    	ltos(terminal->vertlist(p)));
			  vp.add_point(insts(vt*2,k),
					insts(vt*2+1,k),
					basecol,pixsize);


			  vt-=1;
			}
		      else if (pterm->que.typ==5)
			{
			  p=p-3;
			  vp.add_line(insts(0,k), insts(1,k), insts(2,k), insts(3,k), basecol);
			  vp.add_point(insts(0,k),insts(1,k),basecol,pixsize);
			  vp.add_point(insts(2,k),insts(3,k),basecol,pixsize);
			  //  vp.add_string(insts(0,k),insts(1,k),ltos(terminal->vertlist(0)));
			  // vp.add_string(insts(2,k),insts(3,k),ltos(terminal->vertlist(1)));
			}
		    }
		  term=pterm;
		}
	      vp.view();
	    }
	}
   
}



void naive_train()
{

dgrids probs(numclass);
int cxs=cartsizex*cartsizex;

for (int c=0; c<numclass; c++)
  {
    probs(c).init(numland*cxs);
    probs(c).clear();
  }

igrid clcount(numclass);
clcount.clear();
ucgridss temp(numclass);
for (int c=0; c<numclass;c++)
  {
     temp(c).init(numland);
     for (int k=0; k< numland; k++)
       {
	 temp[c][k].init(cartsizex,cartsizex);
	 temp[c][k].clear();
       }
  }

landmarks lmarks;
for (int i=0; i<numtreedata; i++)
  {

     lmarks.read(i);
     int c=lmarks.truth(nist);
     clcount(c)++;
     for (int k=0; k<numland; k++)
       {
	 for (int l=0; l<lmarks.type_count(k); l++)
	   {
              int x=lmarks.location_x(k,l);
              int y=lmarks.location_y(k,l);
	      for (int n=x-near+1; n<x+near; n++)
	      for (int m=y-near+1; m<y+near; m++)
		{
		  if (n>=0 && m >= 0 && n < cartsizex && m < cartsizex)
		    {
		      if (temp[c][k](n,m)==0)
			{
			  probs[c]( k*cxs + n*cartsizex + m)++;
			  temp[c][k](n,m)=1;
			}
		    }
		}
	   }
       }

  }

for (int c=0; c<numclass; c++)
  {
    gridfun( MAX(log(probs[c](i,j)/clcount(c)),-10),probs[c]);

  }

ofstream ofs("probs");

for (int c=0; c<numclass; c++)
  {
    probs[c].write(ofs);
  }

}

void naive_test()
{

ifstream ifs("probs");
dgrids probs(numclass);
int cxs=cartsizex*cartsizex;
for (int c=0; c<numclass; c++)
  {
    probs[c].read(ifs);
  }

ucgridss temp(numclass);
for (int c=0; c<numclass;c++)
  {
     temp(c).init(numland);
     for (int k=0; k< numland; k++)
       {
	 temp[c][k].init(cartsizex,cartsizex);
       }
  }


double correct=0;
for (int i=0; i<numtreedata; i++)
  {

  for (int c=0; c< numclass; c++)
    for (int k=0; k<numland; k++)
      temp[c][k].clear();
     landmarks lmarks;
     lmarks.read(i);

     dgrid like(2,numclass);
     like.clear();
     for (int c=0; c<numclass; c++)
       like(1,c)=c;
     for (int k=0; k<numland; k++)
       {
	 for (int l=0; l<lmarks.type_count(k); l++)
	   {
              int x=lmarks.location_x(k,l);
              int y=lmarks.location_y(k,l);
	      for (int n=x-near+1; n<x+near; n++)
	      for (int m=y-near+1; m<y+near; m++)
		{
		  if (n>=0 && m >= 0 && n < cartsizex && m < cartsizex)
		    {
		      for (int c=0; c<numclass; c++)
			{
			  if (temp[c][k](n,m)==0)
			    {
			      like(0,c) += probs[c](k*cxs + n*cartsizex + m);
			      temp[c][k](n,m)=1;
			    }
			}
		    }
		}
	   }
       }


     like.sort_columns(numclass,compar);
     //     like.dump();

     if (like(1,0)==lmarks.truth(nist))
             correct++;
  }

 cout << "CORECT " << correct/numtreedata << endl;

}

void anal_trees(ltatrees& ltrees, dgrid& betas, int numtrees)
{

 if (maxdepth > 1)
    return;
  dgrid thist(cartsizex*cartsizex,numland);
  for (int t=firsttree; t<firsttree+numtrees; t++)
    {
      ltaiter d, dy;
      
      d=ltrees[t].top();
      dy=d.yes();
      if (dy->hist[0]>dy->hist[1])
	{
	thist( d->que.r1.land2, d->que.r1.land1)+=betas(t);
	//	cout << d->que.r1.land2 << " " << d->que.r1.land1 << " " << betas(t) << endl;
	}
      else
	{
	thist( d->que.r1.land2, d->que.r1.land1)-=betas(t);
	//cout << d->que.r1.land2 << " " << d->que.r1.land1 << " " << -betas(t) << endl;
	}
    }     
  double toth=0;
  for (int i=0; i<cartsizex*cartsizex; i++)
    for (int j=0; j<numland; j++)
      {
	toth+=fabs(thist(i,j));
      }
 for (int i=0; i<cartsizex*cartsizex; i++)
    for (int j=0; j<numland; j++)
      {
	  cout << "thist " << i  << " " << j << " " << thist(i,j)/toth << endl;
      }
  cout << " Total " << toth << endl;


}

