// Copyright YALI AMIT, January 15, 2001.

#include "includes.h"
#include "globals.h"
#include <iostream>
#include <cstdio>
#include <cmath>
#include <cassert>
#include "grid.h"
#include "image.h"
#include "viewport.h"
#include "vector.h"
#include "curve_tree.h"



int TOTNODES=0;
int LIMNODES=1000;

simpstrg HIST;
double thrd=1./3.;
double diffprob;

int NANG;
dgrid cosvec;
dgrid sinvec;
void prepare_angles();
void prepare_probs();
void prepare_binomial();
void itertreedown(const roaditer& tt);
void itertreeup(const roaditer& tt);
void sweep_tree();

void calc_z(const roaditer& tt);
void itertri(const roaditer& tt);
int relation(const roaditer& s, const roaditer& t);
void  startup(const roaditer& tt, igrid &z);
void eval_dat(const roaditer &tt);
void birth(const roaditer &tt);
void  draw_path(viewport &vp, const roaditer& ttl);
void  update_zdiff(const roaditer& tt);
void  update_zold(const roaditer& tt);
void update_zstar(const roaditer& tt);
void track(igrid &z, viewport& vp);
roaditer TOP;
roaditer TM, TMNEW, TMPATH, TMSTAR;

int DES=0;
int STEP=0;
int MAXSTEP=40;


dgrid P1;
dgrid P0;

double PBrat, PBMOLD, PBMNEW;

double ZMAX=.5;
double ZMIN=.000001;
double ZDIFF=0;// Used to store current closest posterior to ZMAX.
double ZOLD=0.;// Used to store current highest posterior.
int ZDEPTH=0; // Depth of current highest posterior.
image im;

// Initialize tree based tracking using two initial points provided by
// user.  
void roadtrack(image &imin)
{



im=imin;

im.normalize(0, numcolors-1);
imin.normalize(imrange*numcolors, (1-imrange)*numcolors-1);

// The distribution on each segment is assumed binomial with certain
// success probabilities provided by parameter `inside'.
// Off curve probabilities are binomial with parameter outside.
// Length of each segment given by
prepare_binomial();
prepare_angles();


viewport vp(imin);

 
 igrid z;
 if (point1x < 0 || point1y < 0 || point2x < 0 || point2y < 0)
   {
     vp.view();
     vp.add_string(0,0,"CHOOSE FIRST TWO POINTS");
     vp.get_marked_points(z);
   }
 else
   {
     z.init(2,2);
     z(0,0)=point1x; z(0,1)=point1y; z(1,0)=point2x; z(1,1)=point2y;
     vp.add_point(z(0,0),z(0,1),basecol,5);
     vp.add_point(z(1,0),z(1,1),basecol,5);
     vp.view();
   }

 // Rerun tracking algorithm numtrack times. Reinitialized
 // at high posterior point.
 for (int g=0; g<numtrack; g++)
   {
     cout << "TRACKNO " << g << endl;
     track(z, vp);
   }
 
}



// Prepare information for NANG angles to make quick computation
// of endpoints of new children nodes. `deg' is global variable
// determining relative angles of left and right segments. LEN
// is the length of each segment. ORTHLEN is the length orthogonal
// to segment to check pixel values.
void prepare_angles()
{
NANG=360/deg;
cosvec.init(NANG,1);
sinvec.init(NANG,1);

double an=PI/(180/deg);
for (int i=0; i< NANG; i++)
  {
cosvec(i,0)=cos(an*i)*LEN;
sinvec(i,0)=sin(an*i)*LEN;
  }
}


// Start tracking algorithm
void track(igrid &z, viewport& vp) 
{
  TOTNODES=0;
  TM=NULL;
  TMSTAR=NULL;
  TMNEW=NULL;
  
  // Class roadtree - ternary tree. see `curve_tree.h'
  roadtree road;
  
  road.maketop();
  //roaditer tt=road.top();
  TOP=road.top();

  diffprob = (1.0 - sameprob) / 2.0;

  // Start things up by randomly picking one of TOP's sons for first question
  ZDIFF=1.;
  startup(TOP,z);
  MAXSTEP=numiters;
  
  
  for (STEP=1; STEP<MAXSTEP; STEP++)
    {
      //     cout << "STEP " << STEP << endl;
      ZDIFF=1.;
      ZOLD=0.;
      TM=TMNEW;     
      
      // Evaluate data at segment chosen to have most information.
      eval_dat(TM);

      // Show that segment.
       if (dumpdetections)
      	vp.add_line(int(rint(TM->p1[0])),int(rint(TM->p1[1])),
      	    int(rint(TM->p2[0])),int(rint(TM->p2[1])),
		    /* numcolors-*/7);
      
      // Computer factor for updating posteriors given new observed data
      // on all segments of current tree.
      PBrat=1./( (TM->zold)*(P1[TM->Y])+(1-(TM->zold))*P0[TM->Y]);
      
      // sweep through existing tree, if appropriate add 3 children
      //to a terminal node, update posteriors everywhere, and keep track
      //of node with highest information content - posterior closest to ZM
      // and node with highest posterior TMSTAR.
      sweep_tree();

      // If chosen node is the same as previous chosen node,
      // algorithm is stuck, return.
      if (TMNEW==TM)
	{
	  cout << " Nothing new " << endl;
         draw_path(vp,TMPATH);
         vp.view();
	 return;
       }

      if (dumpdetections)
	{
	  
	  if (dumpdetections==1)
	    {
	      draw_path(vp,TMPATH);     
	      vp.refresh();	    
	    }
	  else if (dumpdetections==2)
	    {
	      draw_path(vp,TMPATH);
	      vp.view();
	    }
	  else
	    vp.view();
	
	}
      
      // If the node with highest posterior has sufficient depth
      // restart from that node.
      if (TMSTAR.depth()>mindepth)
	{
	  cout << " RESTARTING FROM HIGH POSTERIOR NODE " << endl;
	  draw_path(vp, TMSTAR);
          vp.view();
          z(0,0)=int(rint(TMSTAR->p1[0]));          
          z(0,1)=int(rint(TMSTAR->p1[1]));  
	  z(1,0)=int(rint(TMSTAR->p2[0]));          
          z(1,1)=int(rint(TMSTAR->p2[1]));          
          return;
	}         
      //cout << " TOTNODES " << TOTNODES << endl;
    }
  draw_path(vp,TMPATH);
  vp.view();
}

// Sweep through current tree.
// First go down from the node at which data was computed. It
// May not be a terminal node.
// Then go up along ancestors of this node
// at each time going back down
void sweep_tree()
{

  //cout << " IN SWEEP TREE " << endl;
  DES=0;
  itertreedown(TM);
  itertreeup(TM);
  
}

// Iterate recursivel downwards from TM where data was evaluated.
void itertreedown(const roaditer& tt)
{


  //cout << " IN ITERDOWN " << DES << endl;

  // If not terminal go to three sones.
  if (tt.branch())
    {
      itertreedown(tt.son1());
      itertreedown(tt.son2());
      itertreedown(tt.son3());
      // Posterior at this node is sum of posteriors of children.
      tt->zold = tt.son1()->zold+tt.son2()->zold+tt.son3()->zold;
      // If posterior is less than ZMIN (very small)
      // the sons are not needed anymore. If they are terminal
      // delete them and `tt' inherits the status of the sons.
      if (tt->zold < ZMIN)
	{
	  if ((tt.son1()).leaf() && tt.son2().leaf() && tt.son3().leaf())
	    {
	      if (tt.son1()==TMSTAR || tt.son2()==TMSTAR || tt.son3()==TMSTAR)
		TMSTAR=tt;
	      if (tt.son1()==TMPATH || tt.son2()==TMPATH || tt.son3()==TMPATH)
		TMPATH=tt;
	      if (tt.son1()==TMNEW || tt.son2()==TMNEW || tt.son3()==TMNEW)
		{ 
		  TMNEW=tt;
		  ZDIFF=fabs(tt->zold-ZMAX);
		}
	      // ELiminate these nodes from the tree.	      
	      if (tt.freechildren())
		TOTNODES-=3;
	    }
	}
      // Update search for maximum posterior
      update_zstar(tt);
      // Update search for most informative node `posterior closest to ZMAX'.
      update_zdiff(tt);
    }
  // Terminal node - update posterior using formulas 
  else
    calc_z(tt); 
  
  return;
}


// Iterate upwards in the tree updating posterior and updating
// maximal posterior and most informative node.

void itertreeup(const roaditer& tt)
{

DES=-1;
//cout << " IN ITERUP " << DES << endl;

  if (tt.depth()>0)
    {
      roaditer par=tt.parent();
      // Only go down sibling nodes.
      if (tt.ord()==1)
	{  
	  itertreedown(par.son2());
	  itertreedown(par.son3());
	}
      else if (tt.ord()==2)  
	{
	  itertreedown(par.son1());
	  itertreedown(par.son3());
	}
      else 
	{
	  itertreedown(par.son1());
	  itertreedown(par.son2());
	}
      // Posterior is sum of posteriors on children
      par->zold = par.son1()->zold+par.son2()->zold+par.son3()->zold; 
      // Update highest posterior node - EXPPP
      update_zstar(tt);
      // Update closest difference of posterior to ZMAX.
      update_zdiff(par);
      // Go up one more step.
      itertreeup(par);
    }

return;
}


// Update posterior of a terminal node using factor PBrat
void calc_z(const roaditer& tt)
{
  
  // Extend t_{m-1} to new nodes of T_m
  
  double PYTMgiventt;

  if (DES<0)
    PYTMgiventt=P0[TM->Y];
  else 
    PYTMgiventt=P1[TM->Y];

  tt->zold = tt->zold*PBrat*PYTMgiventt;

  // Compare updated z->old to ZMAX.


 // Update node closest to ZMAX - record in TM.
 update_zdiff(tt);

 // Update node of highest posterior. - record in TMNEW.
 update_zold(tt);

 // Check if highest posterior node is over ZSTAR for restarting.
 // And record node in TMSTAR;
 update_zstar(tt);

 // If the terminal node tt has posterior greater than ZMAX
 // there is a chance for the children to have posteriors even closer.
 // Produce the children and compute their posteriors.
 if (tt->zold >= ZMAX)
          birth(tt);

}


// Evaluate distance of posterior on current node to  ZMAX.
void  update_zdiff(const roaditer& tt)
{

  //cout <<  " In update " << tt.history() << endl;
  // This node has been used, the data on corresponding
  // segment evaluated. Can not use again.
  if (tt->Y > 0)
    return;

  double zdiff=fabs(tt->zold-ZMAX);
  if ( zdiff < ZDIFF)
    {
      TMNEW=tt;
      ZDIFF=zdiff;
    }
  // If the difference is the same as current closest choose
  //between the two at random
  else if (zdiff==ZDIFF)
    {
      int logic=( (rand() % 1001) > 500 );
      if (logic)
	{
	  TMNEW=tt;
	  ZDIFF=zdiff;
	}
    }
  //cout << " NEWZDIFF " << ZDIFF << " TMNEW " << TMNEW.history() <<endl; 

  
}

void update_zstar(const roaditer& tt)
{

  //  cout << " IN ZSTAR " << tt->zold << " " << tt.depth() << endl;
 if ( tt->zold > ZSTAR && tt.depth() > ZDEPTH)
    {

       TMSTAR=tt;
       ZDEPTH=tt.depth();
    }
}


void update_zold(const roaditer& tt)
{

 if ( tt->zold > ZOLD)
    {
       TMPATH=tt;
       ZOLD=tt->zold;
    }
}

// Produce 3 children of tt, compute posterior, if closer
// than ZDIFF to ZMAX, choose one of the children at random.
void birth(const roaditer &tt)
  {
   if (TOTNODES > LIMNODES)
        return;
    TOTNODES+=3;
    tt.makechildren();
    
    if (priordist[0]=='u') { // uniform.
       tt.son1()->zold=thrd*tt->zold;
      tt.son2()->zold=thrd*tt->zold;
      tt.son3()->zold=thrd*tt->zold;
    }
    else if (priordist[0]=='c') { // prefer same angel as before.
       tt.son1()->zold=tt->zold *
      	(tt.parent().son1()==tt ? sameprob : diffprob);
      tt.son2()->zold=tt->zold *
	(tt.parent().son2()==tt ? sameprob : diffprob);
      tt.son3()->zold=tt->zold *
	(tt.parent().son3()==tt ? sameprob : diffprob);
    }
    else { // Prefer continuing straight
      tt.son1()->zold=tt->zold * diffprob;
      tt.son2()->zold=tt->zold * sameprob;
      tt.son3()->zold=tt->zold * diffprob;
    }

    // Get end-points of curve.
    tt.son1()->getpoints(tt,1);
    tt.son2()->getpoints(tt,2);
    tt.son3()->getpoints(tt,3);
    double zdiff = fabs(tt.son1()->zold-ZMAX);
    //cout << " In birth " << tt.son1()->zold << endl;
    if ( zdiff < ZDIFF)
      {
        ZDIFF=zdiff;
	int logic=( rand() % 300 );
	if (logic < 100)
	  TMNEW=tt.son1();
	else if (100 <= logic && logic < 200)
	  TMNEW=tt.son2();
	else
	  TMNEW=tt.son3();
      }
    //cout << " NEWZDIFF " << ZDIFF << " TMNEW " << TMNEW.history() << endl;     
  }

// Compute initial angle of 2 points provided by user.
void  startup(const roaditer& tt, igrid &z)
{

tt->zold=1.;
tt->p1[0]=z(0,0); tt->p1[1]=z(0,1);
tt->p2[0]=z(1,0); tt->p2[1]=z(1,1);

double ang=atan2(z(1,0)-z(0,0),z(1,1)-z(0,1))*(180./PI);
tt->angindex= ( (( 360 + int(ang) ) / deg ) % NANG);
birth(tt);



}

// Evaluate likelihood at a node `tt'.
void eval_dat(const roaditer &tt)
{
  double uvec[2], ovec[2];  
  
  // The vector given by the arc.
  uvec[0]=(tt->p2[0]-tt->p1[0])/LEN;
  uvec[1]=(tt->p2[1]-tt->p1[1])/LEN;
  
  //cout << uvec[0] << " " << uvec[1] << endl;
  // Orthogonal vector.
  ovec[0]=-uvec[1]*ORTHLEN;
  ovec[1]= uvec[0]*ORTHLEN;
  
  //cout << ovec[0] << " " << ovec[1] << endl;
  double p1[2], p2[2], pleft[2], pright[2];
  p1[0]=tt->p1[0]; p1[1]=tt->p1[1];
  
  int logic=0;
  // Count how many times the appropriate inequalities are obtained.
  for (int j=0; j<LEN; j++)
    {
      p2[0]=p1[0]+uvec[0];
      p2[1]=p1[1]+uvec[1];
      pleft[0]=p1[0]+ovec[0];
      pleft[1]=p1[1]+ovec[1];
      pright[0]=p1[0]-ovec[0];
      pright[1]=p1[1]-ovec[1];
      int v1=im.get(int(rint(p1[0])),int(rint(p1[1])));
      int v2=im.get(int(rint(p2[0])),int(rint(p2[1])));
      double dv = fabs(v1-v2);
      int vr=im.get(int(rint(pright[0])),int(rint(pright[1])));
      int vl=im.get(int(rint(pleft[0])),int(rint(pleft[1])));
      
      logic+=
	(dum1==0 ?
	 ( dv < fabs(v1-vl) && dv < fabs(v1-vr) && v1 > vl && v1 > vr):
	 ( dv < fabs(v1-vl) && dv < fabs(v1-vr) && v1 < vl && v1 < vr));

     
      p1[0]=p2[0]; p1[1]=p2[1];
    }
  
  tt->Y=logic;
}
 

// Find the ZMAX determined by the two binomials for on and off curve.
void prepare_binomial()
{

P1.init(LEN+1);
P0.init(LEN+1);
double pon=pow(inside,LEN);
double qon=1;
double poff=pow(outside,LEN);
double qoff=1;
double sum0=0, sum1=0;
     double BB=1;
     double ent0=0, ent1=0;
  for (int k=LEN; k>=0; k--)
    {
      P1[k]=BB*pon*qon;
      P0[k]=BB*poff*qoff;
      pon=pon/inside;
      qon=qon*(1-inside);
      poff=poff/outside;
      qoff=qoff*(1-outside);
      sum0+=P0[k];
      sum1+=P1[k];
      ent0+=P0[k]*log(P0[k]);
      ent1+=P1[k]*log(P1[k]);
      BB=BB*k/(LEN-k+1);
      //      cout << P0[k] << " " << P1[k] << endl;
    }
double entmix=0;
ZMAX=0;
double hmax=0;
  for (double t=0;t<=1;t+=.001)
    {
       entmix=0;
       for (int k=0; k<=LEN; k++)
	 {
            entmix+=(t*P1[k]+(1-t)*P0[k])*log(t*P1[k]+(1-t)*P0[k]);
         }
        double h = t*ent1+(1-t)*ent0-entmix;
	//cout << h << endl;
      if (h>hmax)
	{
           hmax=h;
           ZMAX=t;
	}

    }

  cout << ZMAX << endl;
  cout << " sums " << sum1 << " " << sum0 << endl;
}



// Get points for children nodes of a current node.
void roadata::getpoints(const roaditer& par, int type)
{
 angindex= (NANG + par->angindex+(type-2)) % NANG;

 p1[0]=par->p2[0];
 p1[1]=par->p2[1];
 p2[0]=par->p2[0]+sinvec[angindex];
 p2[1]=par->p2[1]+cosvec[angindex];


}



void  draw_path(viewport &vp, const roaditer& ttl)
{
  roaditer tt=ttl;
  cout << " Child node " << tt.depth() << " " << tt->zold << endl;
  
  //vp.clear_geom();
  while( tt )
    {
      vp.add_line(int(rint(tt->p1[0])),int(rint(tt->p1[1])),int(rint(tt->p2[0])),int(rint(tt->p2[1])),basecol);
      tt = tt.parent();
    }
  //   vp.refresh();
}
