// Copyright YALI AMIT, January 15, 2001.
#include <iostream>
#include <cstdio>

#include "includes.h"
#include "image.h"
#include "trainstat.h"
#include "detection.h"
#include "viewgrid.h"
#include "viewport.h"
#include "globals.h"
#include "wave.h"
#define log2(a) log((double)a)/log(2.)

double dts,DT;
int N; // Number of coefficients currently being updated
igrid dc(2); // Input points determining location of initial curve.
int ddc; 
dgrids ingamma(2); // Coordinates of initial curve
dgrids coef(2), coef_b(2); // Wavelet coeficients of two components of contour.
dgrid ddim;
int length;
dgrids gam(2); // Updated curve.
dgrid prior_cofs; // Coeficients of prior - multiply the wavelet coeficients.
double NEWVAL=0;
double VAL=0;
double OLDVAL=0;

// Function declarations
void init_algo(int user,image &im, dgrid &m);
void defcont(image& im, dgrid& m, int user, detection &det);
void defcontnew(image& im, dgrid& m, int user, detection& det);
void defcontold(image& im, dgrid& m, int user, detection& det);
double get_time_step(dgrids& dgamm,dgrid& prior_cofs);
void draw_curve(dgrids& ingamma, double ddc, igrid dc, viewport& vp, int type);
void aggreg(dgrid &im, dgrid *R, int type, double &T);
// ___________________________________________________________________________

// Run the deformable contour with an input image and
// obtain initialization parameters from user.
void init_defcont(image &im)
{
  igrid m; 
  dgrid dm;

  im.blur_downscale(ratio,sd);
  


  
    // Read in defcont template
    if (dm.read(tempfile))
      {
	cout << dm.dimx() << endl;
      }
    // If doesnt exist mark your own.
     else
       {
	 image tim=im;
	 tim.normalize(imrange*numcolors,numcolors-1);
	 viewport vp;
	 vp.add_grid(tim);
	 vp.add_string(0,0,"CHOOSE A CURVE");
	 vp.view();
	 vp.get_marked_points(m);
	 double dx=im.dimx()/2.;
	 double dy=im.dimy()/2.;
	 dm.init(m.dimx(),m.dimy());
	 for (int i=0; i<m.dimx(); i++)
   	 { 
           dm(i,0)=(double(m(i,0))-dx)/dx;
	   dm(i,1)=(double(m(i,1))-dy)/dy;
  	 }
	 // write it out  
	 dm.write("defcont.temp");
       }


    detection det;
    defcont(im,dm,1, det);
    // Write final curve as a detection.
    if (detfile!="")
      {
	ofstream ofs(detfile);
	det.write(ofs);
      }
    // Dump final curve.
    if (dumpdetections==1)
      det.dump();
}

// ___________________________________________________________________________
// Run deformable contour algorithm with initialization from a detection
// obtained from the detection algorithm.
void defcont_init(image &im, detection &det)
{


if (alpha2!=0)
  im.blur_downscale(alpha2,sd);
else
  alpha2=1;

  int lshiftx = (ltatree1x + ltatree2x + ltatree3x)/3 ;
  int lshifty = (ltatree1y + ltatree2y + ltatree3y)/3;
  partialtrans pt;
  pt.ptr(Pixel(ltatree1x-lshiftx, ltatree1y-lshifty), 
                  Pixel(ltatree2x-lshiftx, ltatree2y-lshifty),
                  Pixel(ltatree3x-lshiftx, ltatree3y-lshifty));


  affinetrans at(pt, Pixel(alpha2*det.tripts[0], alpha2*det.tripts[1]),
                 Pixel(alpha2*det.tripts[2], alpha2*det.tripts[3]),
                 Pixel(alpha2*det.tripts[4], alpha2*det.tripts[5]));


  igrid dc(2);
  dc(0) = im.dimx()/2;
  dc(1) = im.dimy()/2;  
  double ddc = dc(0);
  dgrid dm;
  dm.read(tempfile);

  // Register template according to affine map of the detection.
  // shiftx, shifty, are translation corrections if you dont want
  // the center of the initial curve to be at the center of the triangle
  // Center of image is assumed to be the origin and the curve is coded in
  // double precision assuming the image is [-1,1]x[-1,1]

  double cx=0, cy=0;

  for (int i=0; i<dm.dimx(); i++)
    {
      cx+=dm(i,0);
      cy+=dm(i,1);
    }
  cx = cx/dm.dimx();
  cy = cy/dm.dimx();

  for (int i=0; i<dm.dimx(); i++)
  {
    double d0 = at.coefs[0][0]*(dm(i,0)-cx)*post+
                at.coefs[0][1]*(dm(i,1)-cy)*post+at.coefs[0][2] + shiftx;
    double d1 = at.coefs[1][0]*(dm(i,0)-cx)*post+
                at.coefs[1][1]*(dm(i,1)-cy)*post+at.coefs[1][2] + shifty;
    dm(i,0)=(d0 - dc(0))/ddc;
    dm(i,1)=(d1 - dc(1))/ddc;

  }

  defcont(im,dm, 0, det);
}


// 
void defcont(image &im, dgrid& m, int user, detection &det)
{

  // Check if curve has power of two points, if not exit.
  // Should be a subroutine to interpolate to power of 2  points. Sorry.
  if ( pow(2,int(rint(log2(m.dimx())))) != m.dimx())
    {
      cerr << " curve does not have 2^n points " << endl;
      exit(0);
    }

  im.normalize(0,64);

  // Initialize parameters and the curve, the prior coeficients. 
  // Get wavelet coefficients of initial curve, and interpolate
  // to higher dimensional curve if required.
  init_algo(user, im, m);



  if (inside < 0)
    defcontnew(im, m, user, det);
  else
    defcontold(im, m, user, det);
}


// Run deformable contour with simultaneous parameter estimation.
// The Gaussian model with unkown means and variances.
// Structure det.locs gets OVERWRITTEN with final contour.
void defcontnew(image& im, dgrid& m, int user, detection& det)
{
  double g0in=0, g1in=0, g2in=0, g0old=0;
  image imt=im;
  imt.normalize(imrange*numcolors,(1-imrange)*numcolors-1);
  viewport vp(imt);

  draw_curve(gam, ddc, dc, vp, 0);


  dgrid ddimsq(im.dimx(),im.dimy());
  gridfun(ddim(i,j)*ddim(i,j),ddimsq);  

  // Buffers for storing partial integrals of I^0, I^, I^2
  dgrid R0[2], R1[2], R2[2];
  double T0,T1,T2;

  // Compute the partiaul integrals ofr I^0, I^1, I^2
  aggreg(ddim,R0,0,T0);
  aggreg(ddim,R1,1,T1);
  aggreg(ddim,R2,2,T2);
 
  dgrids gamm(2);
  dgrids dgamm(2);
  dgrids f0(2),f1(2),f2(2);


  gamm(0).init(1,length);
  gamm(1).init(1,length);
  dgamm(0).init(1,length);
  dgamm(1).init(1,length);
  f2(0).init(1,length);
  f2(1).init(1,length);
  f1(0).init(1,length);
  f1(1).init(1,length);
  f0(0).init(1,length);
  f0(1).init(1,length);

  N=FIRSTDIM;
  if (LASTDIM==0)
    LASTDIM=length;
  int count_neg=0;
  int iter; 
  // Iterate the gradient descent algorithm
  for (iter=0; iter < numiters; iter++)
  {
    //    cout << "iter " << iter << endl;
    // Inverse wavelet transform - coeficients to curve
    if (iter > 0){
      ingamma(0)=gamm(0);
      ingamma(1)=gamm(1);
    }

    // Get curve from wavelet coeficient - inverse wavelet transform on each
    // component
    gamm(0)=coef(0);
    unwavep1(gamm(0),1);
    gamm(1)=coef(1);
    unwavep1(gamm(1),1);

    // Gradient of curve
    for (int i=0; i<length; i++)
    {
            int i1 = (i+1)%length;
      dgamm(0)(0,i)= gamm(0)(0,i1)-gamm(0)(0,i);
      dgamm(1)(0,i)= gamm(1)(0,i1)-gamm(1)(0,i);
    }



    // Recalculate timestep using heurisitc for Hessian
    if ( iter % 10 == 0)
      dts = get_time_step(dgamm,prior_cofs);

    // Display current curve every displayinter iterations
    if ( iter % displayinter == 0)
	draw_curve(gamm, ddc, dc, vp, 1);


    g0old=g0in;
    g0in=0; g1in=0; g2in=0;
    // Get function values of R's at points of curve and integrate along the curve
    for (int i=0; i<length; i++)
    {
      int px = int(rint(gamm(0)(0,i)*ddc+dc(0)));
      int py = int(rint(gamm(1)(0,i)*ddc+dc(1)));
      double rr00=R0[0].get(px,py);
      double rr01=R0[1].get(px,py);
      double rr10=R1[0].get(px,py);
      double rr11=R1[1].get(px,py);
      double rr20=R2[0].get(px,py);
      double rr21=R2[1].get(px,py);
      g0in += rr00*dgamm(0)(0,i)-rr01*dgamm(1)(0,i);
      g1in += rr10*dgamm(0)(0,i)-rr11*dgamm(1)(0,i);
      g2in += rr20*dgamm(0)(0,i)-rr21*dgamm(1)(0,i);
    }
   g0in =g0in*.5;
   g1in =g1in*.5;
   g2in =g2in*.5;
   // Use these cirvilinear integrals to get mean and variance
   // inside and outside the curve
   double g0out=T0-g0in;
   double g1out=T1-g1in;
   double g2out=T2-g2in;
   double meanin=g1in/g0in;
   double meanout=g1out/g0out;
   double varin = MAX(.001,g2in/g0in-meanin*meanin);
   double varout = MAX(.001,g2out/g0out-meanout*meanout);
   // Print out information
   if (dumpdetections==1)
     {
       cout << " areas " << g0in << " " << g0out << " " << g0in/g0old << endl;
       cout << " means " << g1in/g0in << " " << g1out/g0out << endl;
       cout << " variances " << varin << " " << varout << endl;
     }
   // If ratio of newarea to old area is to large stop,
   // things are going haywire
   if ( (iter > 10 && fabs(log(g0in/g0old)) > 1))
     break;
   
   // Compute the function which will be sent to the forward wavelt transform
   for (int i=0; i<length; i++)
    {
      int px = int(rint(gamm(0)(0,i)*ddc+dc(0)));
      int py = int(rint(gamm(1)(0,i)*ddc+dc(1)));
      double tt=ddim.get(px,py);
      double ttsq=ddimsq.get(px,py);
      f2(0)(i) = -ttsq*dgamm(1)(0,i);
      f2(1)(i) = ttsq*dgamm(0)(0,i);
      f1(0)(i) = -tt*dgamm(1)(0,i);
      f1(1)(i) = tt*dgamm(0)(0,i);
      f0(0)(i) = -dgamm(1)(0,i);
      f0(1)(i) = dgamm(0)(0,i);

    }
    // Calculate wavelet transform.
    wavep1(f2(0),1);
    wavep1(f2(1),1);
    wavep1(f1(0),1);
    wavep1(f1(1),1);
    wavep1(f0(0),1);
    wavep1(f0(1),1);


    // Compute current value of cost function (NEWVAL)
    double login = log(varin);
    double logout = log(varout);
    NEWVAL=g0in*login+g0out*logout;   
    VAL=0; 

    // Do the gradietn step. Update coeficients
    // Update active coeficients, from (length-N) to length.
    // The last coefficient of the wavelet transform correspond to 
    // the lower resolutions. N determines the number of coeficients
    // being updated in the coarse to fine method.
    for (int i=length-N; i<length; i++)
    {
      double dcoef0 = f0(0)(0,i)*login + 
	( f2(0)(0,i) - f0(0)(0,i)*varin - 2*f1(0)(0,i)*meanin 
	  + f0(0)(0,i)*meanin*meanin )/varin;
      dcoef0 -= f0(0)(0,i)*logout + 
	( f2(0)(0,i) - f0(0)(0,i)*varout - 2*f1(0)(0,i)*meanout 
	  + f0(0)(0,i)*meanout*meanout )/varout;
      // Penalize deviations of the coeficients from the initial coeficients
      // stored in coef_b. This can be done for all resolutions or only higher ones.
      dcoef0 += prior_cofs(i,0)*(coef(0)(0,i)-coef_b(0)(0,i));
      coef(0)(0,i) -= dcoef0*dts;
      
      double dcoef1 = f0(1)(0,i)*login + 
     ( f2(1)(0,i) - f0(1)(0,i)*varin - 
       2*f1(1)(0,i)*meanin + f0(1)(0,i)*meanin*meanin )/varin;
      dcoef1 -= f0(1)(0,i)*logout + 
	( f2(1)(0,i) - f0(1)(0,i)*varout - 2*f1(1)(0,i)*meanout 
	  + f0(1)(0,i)*meanout*meanout )/varout;
      dcoef1 += prior_cofs(i,0)*(coef(1)(0,i)-coef_b(1)(0,i));
      coef(1)(0,i) -= dcoef1*dts;
      VAL += fabs(dcoef0*dts)+fabs(dcoef1*dts);
    }
    // VAL is a measure of change of the coeficients
    // Check if magnitude of gradient is changing any more.
    // If not double number of active coeficients
    
    double VAL1 = (OLDVAL-NEWVAL);
    count_neg += ( (VAL1 < 0) ? 1 : -1);
    VAL = ddc*(VAL)/length;
    //    cout << " VALS " << VAL << " " << VAL1 << " " << N << endl;
    // Stopping criterion for current level of coeficients (N)
    // If VAL is small or the cost function has gone UP more than 7
    // times goto next dimension or if highest dimension stop.
    if (iter > 10 && ( VAL < (minenergychange) || count_neg > 7) )
      {
	if(N < LASTDIM) 
	  {
	    count_neg=0;
	    N = 2*N;
	    cout << iter << " " << VAL << " N " << N << endl;
	    dts = get_time_step(dgamm, prior_cofs);
	    cout << " dts " << dts << endl;
	  }
	else 
	  break;
      }
    OLDVAL=NEWVAL; 
  }

  cout << " Stopped at " << iter << " Dim " << N << endl; 

  draw_curve(gamm, ddc, dc, vp, 0);

  // Put the curve in the locs entry of the detection.
  det.locs.init(length,2);
  for (int i=0; i<length; i++)
    {
      det.locs(i,0) = int(rint(gamm(0)(0,i)*ddc));
      det.locs(i,1) = int(rint(gamm(1)(0,i)*ddc));
    }
}

// Calculate partial integrals for unkown parameter case, I^0,I^1,I^2
// ___________________________________________________________________________
void aggreg(dgrid &im, dgrid *R, int type, double &T)
{

T=0;

int dx=im.dimx();
int dy=im.dimy();
int lowx=0;
int lowy=1;

// Restrict domain to subimage
if (LBX > 0)
  {
     lowx=LBX;
     dx=LBX+eyex;
     lowy=LBY;
     dy=LBY+eyey;
  }
double fdc=2./double(dx);
R[0].init(dx+1,dy+1);
R[0].clear();
R[1].init(dx+1,dy+1);
R[1].clear();
int x, y;
for (x=lowx; x<dx; x++)
for (y=lowy; y<=dy; y++)
  {
    switch (type) 
      {
      case 0:
	{
	  R[0](x,y)=R[0](x,y-1)+fdc;
	  T += 1;
	}
	break;
      case 1:
	{
	  R[0](x,y)=R[0](x,y-1)+im.get(x,y-1)*fdc;
	  T += im.get(x,y-1);
	}
	break;
      case 2:
	{
	  R[0](x,y)=R[0](x,y-1)+im.get(x,y-1)*im.get(x,y-1)*fdc;
	  T += im.get(x,y-1)*im.get(x,y-1);
	}
        break;
      }
  }



for (y=0; y<dy; y++)
for (x=1; x<=dx; x++)
  {
    switch (type) 
      {
      case 0:
	{
	  R[1](x,y)=R[1](x-1,y)+fdc;
	}
      break;
      case 1:
	{
	  R[1](x,y)=R[1](x-1,y)+im.get(x-1,y)*fdc;
	}
      break;
      case 2:
	{
	R[1](x,y)=R[1](x-1,y)+im.get(x-1,y)*im.get(x-1,y)*fdc;
	}
       break;
      }

  }

T = T*fdc*fdc;
}


// The Gaussian model with known means and fixed variance
// Inside and outside means are provided as `inside' and `outside'
void defcontold(image& im, dgrid& m, int user, detection& det)
{
 
  image imt=im;
  imt.normalize(imrange*numcolors,(1-imrange)*numcolors-1);
  viewport vp(imt);

  // Draw initial curve
  draw_curve(ingamma, ddc, dc, vp, 0);

  dgrids gamm(2);
  dgrids dgamm(2);
  dgrids f(2);
  dgrids oldf(2);

  gamm(0).init(1,length);
  gamm(1).init(1,length);
  dgamm(0).init(1,length);
  dgamm(1).init(1,length);
  f(0).init(1,length);
  f(1).init(1,length);
  oldf(0).init(1,length);
  oldf(1).init(1,length);

  N=FIRSTDIM;
  if (LASTDIM==0)
    LASTDIM=length;
  int iter;

  // Iterate gradient descent
  for (iter=0; iter < numiters; iter++)
  {



    // Inverse wavelet transform - coeficients to curve
    if (iter > 0){
      ingamma(0)=gamm(0);
      ingamma(1)=gamm(1);
    }
    gamm(0)=coef(0);
    unwavep1(gamm(0),1);
    gamm(1)=coef(1);
    unwavep1(gamm(1),1);

    // Gradient of curve
    for (int i=0; i<length; i++)
    {
            int i1 = (i+1)%length;
      dgamm(0)(0,i)= gamm(0)(0,i1)-gamm(0)(0,i);
      dgamm(1)(0,i)= gamm(1)(0,i1)-gamm(1)(0,i);
    }

    // Recalculate timestep
    if ( iter % 10 == 0)
      dts = get_time_step(dgamm,prior_cofs);

    // Display curve every displayinter iterations
    if ( iter % displayinter == 0)
      draw_curve(gamm, ddc, dc, vp, 1);

    if (iter > 0)
    {
      oldf(0)=f(0);
      oldf(1)=f(1);
    }

    // Get image values at points of curve
    for (int i=0; i<length; i++)
    {
      double tt=ddim.get(int(rint(gamm(0)(0,i)*ddc+dc(0))),
			 int(rint(gamm(1)(0,i)*ddc+dc(1))));
      tt =  (tt-inside)*(tt-inside)-(tt-outside)*(tt-outside);
      f(0)(i) = -tt*dgamm(1)(0,i);
      f(1)(i) = tt*dgamm(0)(0,i);
    }

    // Calculate wavelet transform.
    wavep1(f(0),1);
    wavep1(f(1),1);

    // Update active coeficients
    VAL=0;
    for (int i=length-N; i<length; i++)
    {
      double dcoef0 = (prior_cofs(i,0)*(coef(0)(0,i)-coef_b(0)(0,i)) + f(0)(0,i))*dts;
      coef(0)(0,i) = coef(0)(0,i)-dcoef0;
      double dcoef1 = (prior_cofs(i,0)*(coef(1)(0,i)-coef_b(1)(0,i)) + f(1)(0,i))*dts;
      coef(1)(0,i) = coef(1)(0,i)-dcoef1;
      // USed to evaluate degree of change
      VAL += fabs(dcoef0)+fabs(dcoef1);
    }

    // Check if magnitude of change `VAL' is large
    // If not double number of active coeficients
    VAL = ddc*(VAL)/length;
    if (iter > 10 && (VAL) < (minenergychange) )
      {
      if (N < LASTDIM) 
	{
	  N = 2*N;
	  cout << " VAL " << VAL << " " << N << endl;	  
	  dts = get_time_step(dgamm, prior_cofs);
	  cout << " dts " << dts << endl;
	}
      else
        break;
      }
  }

  cout << " Stopped at " << iter << " Dim " << N << endl; 
  draw_curve(gamm, ddc, dc, vp, 0);

 // Put the curve in the locs entry of the detection.
  det.locs.init(length,2);
  for (int i=0; i<length; i++)
    {
      det.locs(i,0) = int(rint(gamm(0)(0,i)*ddc));
      det.locs(i,1) = int(rint(gamm(1)(0,i)*ddc));
    }

}

// ___________________________________________________________________________

double get_time_step(dgrids& dgamm,dgrid& prior_cofs)
{
  double L = 0;
  
  // Corase estimate of Hessian to get time step
  for (int i=0; i < dgamm(0).dimy()-1; i++)
    {
     double sq= sqrt(dgamm(0)(0,i)*dgamm(0)(0,i) + dgamm(1)(0,i)*dgamm(1)(0,i));
     L += sq;
    }

  for (int i=dgamm(0).dimy()-N; i < dgamm(0).dimy(); i++)
    L += (prior_cofs(i,0) + prior_cofs(i,1));
  //  cout << " L " << L << endl;
  return(DT/L);
}

// ___________________________________________________________________________

// Initialize things for the algorithm

void init_algo(int user, image &im, dgrid &m)
{

  // Initial dimension for updated coeffcients
 if (FIRSTDIM==0)
    FIRSTDIM=1;



 
  igrid z; 
  // Initial point given by user as opposed to determinig it from detection
  if (user)
  {
    if (point1x >=0 && point1y >= 0)
      {
        z.init(2);
	z(0)=point1x; z(1)=point1y;
      }
    // Get an initial point for the curve for the user.
    // Scale is determined by the parameter `post'
    else
      {
	viewport vp(im);
	vp.view();
	vp.get_marked_points(z);
	if (z.dimx() < 1)
	  {
	    vp.add_string(20,20,"CHOOSE A POINT");
	    vp.view();
	    vp.get_marked_points(z);
	  }
	// Use only a subregion of the image to drive the algorithm
	// relevant only for online paramtere estimation version
       if (eyex > 0 || eyey > 0)
	 {
	LBX=z(0)-eyex/2;
	LBY=z(1)-eyey/2;
	 }
      }
  }

  // Scaling factor for computed time step
  DT = dt_scale_factor;

  // Get paramteres for wavelet basis of index base
  makevec(base);

  // Normalize image to double betwee 0 and 1.
  ddim.init(im.dimx(), im.dimy());
  gridfun(double(im(i,j)),ddim);
  ddim.normalize(1.);

  dc(0) = ddim.dimx()/2;
  dc(1) = ddim.dimy()/2;
  ddc = dc(0);

  gam(0).init(1,m.dimx());
  gam(1).init(1,m.dimx());

  double cx=0, cy=0;

  // Curve initialized by user, move it to the points (z(0), z(1)) provided above
   if (user)
     {
       for (int i=0; i< m.dimx(); i++)
	 {
	   cx += m(i,0);
	   cy += m(i,1);
	 }
       cx = cx/m.dimx();
       cy = cy/m.dimx();
       for (int i=0; i< m.dimx(); i++)
	 {
	   gam(0)(0,i)=post*((m(i,0)-cx)/ddc)+double(z(0)-dc(0))/ddc;
	   gam(1)(0,i)=post*((m(i,1)-cy)/ddc)+double(z(1)-dc(1))/ddc;

	 }
     }
   // Curve already initialized by detection
    else
    {
      for (int i=0; i< m.dimx(); i++)
	{
	  gam(0)(0,i)=m(i,0);
	  gam(1)(0,i)=m(i,1);
	}
    }

   // It is possible to use a larger  curve_length than provided by the model
   // curve using interpolation. Here interpolation is done with the same wavelet basis
  if (curve_length==0)
   length= m.dimx();
  else 
   length=curve_length;

  // Get wavelet coeficients of input curve
  int gl=gam(0).dimy();

  dgrids cf(2);

  cf(0)=gam(0);
  wavep1(cf(0),1);
  cf(1)=gam(1);
  wavep1(cf(1),1);

  // Plop them into intial coeficient vector
  coef(0).init(1,length);
  coef(1).init(1,length);
  coef_b(0).init(1,length);
  coef_b(1).init(1,length);
  coef(0).clear();
  coef(1).clear();

  int bottom=length-gl;
  if (bottom >0 )
    {
      for (int i=bottom; i < length; i++)
	{
	  coef(0)(0,i)=cf(0)(0,i-bottom)*sqrt(length/gl);
	  coef(1)(0,i)=cf(1)(0,i-bottom)*sqrt(length/gl);
	}
    }
  else if (bottom < 0)
    {
      for (int i=0; i < length; i++)
	{
	  coef(0)(0,i)=cf(0)(0,i-bottom)*sqrt(length/gl);
	  coef(1)(0,i)=cf(1)(0,i-bottom)*sqrt(length/gl);
	}
    }
  else
    {
      coef(0)=cf(0);
      coef(1)=cf(1);
    }

  // Store coeficients of initial curve.
  // coef_b will not change in iterations.
  // Updated coeficient will be in coef.
  coef_b(0)=coef(0);
  coef_b(1)=coef(1);

  // Recompute different resolution initial curve from wavelet coeficients
  ingamma(0)=coef(0);
  unwavep1(ingamma(0),1);
  ingamma(1)=coef(1);
  unwavep1(ingamma(1),1);

  // Get prior coefficients
  prior_cofs.init(length,2);
  prior_cofs.clear();
  int k=1;
  while (k < length)
  {
    for (int i=length-2*k; i < length-k; i++)
    {
      prior_cofs(i,0)=fabs(prior_fac)*sqrt(k);
      prior_cofs(i,1)=fabs(prior_fac)*sqrt(k);
    }

    k = k*2;
  }

  // Zero out low resolution coeficients, no constraints on their change
  if (prior_fac<0)
    {
    for (int i=1; i<4; i++)
      {
	prior_cofs(length-i,0)=0;
	prior_cofs(length-i,1)=0;
      }
    }
 
}


void draw_curve(dgrids& ingamma, double ddc, igrid dc, viewport& vp, int type)
{
  int length = ingamma(0).dimy();
  vp.clear_geom();
  for (int i=0; i < length; i++)
    {
      int  i1 = (i+1) % length;
      vp.add_line(int(rint(ingamma(0)(0,i)*ddc+dc(0))),
		  int(rint(ingamma(1)(0,i)*ddc+dc(1))),
		  int(rint(ingamma(0)(0,i1)*ddc+dc(0))),
		  int(rint(ingamma(1)(0,i1)*ddc+dc(1))),basecol);
      
      if ( dum2 > 0 && i % dum2 == 0)
	{
	  vp.add_point(int(rint(ingamma(0)(0,i)*ddc+dc(0))),
		       int(rint(ingamma(1)(0,i)*ddc+dc(1))),basecol,4);
	  vp.add_string(int(rint(ingamma(0)(0,i)*ddc+dc(0))),
			int(rint(ingamma(1)(0,i)*ddc+dc(1))),ltos(i),basecol);
	}
    }
  
  if (type)
    vp.refresh();
  else 
    vp.view();
  
}
// ___________________________________________________________________________
