// Copyright YALI AMIT, January 15, 2001.



#include "includes.h"
#include "simpstrg.h"
#include "pixel.h"
#include "image.h"
#include "grid.h"

extern simpstrg facenames[];


// __________________________________________________________________________

class detection
{
 private:
  // These member functions are not used.  They are declared to make 
  // sure that default versions aren't called accidentally
  detection(const detection &d);
  void operator==(const detection &d);

 public:
  // Flag used to indicate if detection is valid
  int good;
  // Scale of the original image at which the detection was found
  double scale;
  // Coordinates of the found triangle [x1 y1 x2 y2 x3 y3]
  int tripts[10];
  
  dgrid regim;
  dgrid cartprobs, cartprobs2, cartprobs3;
  // Various counts of detected tags at correct locations
  int counts[6];
  int tricount;
  igrid locs;
  ~detection() { }
  detection() { reinit(); }
  void operator=(const detection& d) {
    good = d.good;
    scale = d.scale;
    memcpy(tripts, d.tripts, 10 * Sint);
    regim=d.regim;
    cartprobs=d.cartprobs;
    cartprobs2=d.cartprobs2;
    cartprobs3 = d.cartprobs3;
    memcpy(counts, d.counts, 6 * Sint);
    locs=d.locs;
  }

  void findmax()
    {
      int maxind=-1;
      double maxprob=0;
      for (int i = 0; i < cartprobs2.dimx(); ++i)
	if (cartprobs2[i] > maxprob) {
	  maxprob = cartprobs2[i];
	  maxind = i;
	}
      counts[5]=maxind;
    }



  void dump() {
    cout << "Image scale\n  ";
    cout << scale << endl;
    int i;
    cout << "Triangle coordinates: ";
    for (i = 0; i < 8; i +=2)
      cout << "( " << tripts[i] << ", " << tripts[i + 1] << " ) ";
    cout << endl;

    cout << "Hough count: " << counts[0] << endl;
    cout << "CART probabilities: ";
    for (i = 0; i < cartprobs.dimx() && cartprobs[i] != -1; ++i)
      cout << cartprobs[i] << " ";
    cout << endl;
    cout << "CART2 probabilities: ";
    for (i = 0; i < cartprobs2.dimx() && cartprobs2[i] != -1; ++i)
      cout << cartprobs2[i] << " ";
    cout << endl;

    cout << "Face belongs to " << facenames[counts[5]] << endl;
    cout << "CART3 probability: " << cartprobs3[0] << ", " << cartprobs3[1] << endl;
    cout << "Numltas " << counts[1] << " hits " << counts[2] << endl;
    cout << "numedgetrees " << counts[3] << " acchits " << counts[4] << endl;
    cout << "label " << counts[5] << endl;

    cout << endl;
  }
  void reinit() { int i; good = 0; scale = 0;
    for (i = 0; i < 10; ++i) tripts[i] = 0;
    regim.clear(-1);
    cartprobs.clear(-1); cartprobs2.clear(-1); cartprobs3.clear(-1);
    for (i = 0; i < 6; ++i) counts[i] = -1;
    locs.init(0,0);
  }
  void read(ifstream& is) {
    varread(is, good);
    varread(is, scale);
    arrayread(is, tripts, 10);
    regim.read(is);
    cartprobs.read(is);
    cartprobs2.read(is);
    cartprobs3.read(is);
    locs.read(is);
    arrayread(is, counts, 6);
  }
  void write(ofstream& os) {
    varwrite(os, good);
    varwrite(os, scale);
    arraywrite(os, tripts, 10);
    regim.write(os);
    cartprobs.write(os);
    cartprobs2.write(os);
    cartprobs3.write(os);
    locs.write(os);
    arraywrite(os, counts, 6);
  }

  void initcart(int numcarts, int itersize, int numclass)
  {
      cartprobs.init(numcarts/itersize); cartprobs.clear();
      cartprobs2.init(numclass); cartprobs2.clear();
      cartprobs3.init(2); cartprobs3.clear();
  }


}; // class detection
typedef vector<detection> detections;





// __________________________________________________________________________
// facetest.h


