// Copyright YALI AMIT, January 15, 2001.

#define GLOBALS_C

// __________________________________________________________________________
// Include files

#include <iostream>
#include <fstream>

#include <cstdlib>
#include <cstring>

#include "includes.h"
#include "simpstrg.h"
#include "globals.h"

// __________________________________________________________________________
// Function prototypes and defines

extern void parse_equation(char* eqn, char*& var, char*& val);
extern int set_global(char* var, char* val);

// __________________________________________________________________________
// Types and lists needed to keep track of global variables.

enum { intvar, longvar, doublevar, simpstrgvar };

struct typedesc { char *v; void *p; int t; };

typedesc varlist[] = 
{ 
#define GLOBAL(x, y) { #y, &y, x##var },
#define GLOBALI(x, y, z) { #y, &y, x##var },
#define GLOBALA(x, y, z, w) { #y, y, x##var },
#include "globlist.h"  
#undef GLOBAL
#undef GLOBALI
#undef GLOBALA
  { "", 0, 0 }
};

// __________________________________________________________________________
// If argv[1] is not an assigment, it is the name of a parameter file. 

void get_parfile(int argc, char** argv, simpstrg& parfile)
{
  if (argc > 1 && strchr(argv[1], '=') == 0)
    parfile = argv[1];

} // get_parfile

// __________________________________________________________________________
// argv may contain statements of the form var=value.  Use these statements
// to set global variables.

void parse_argv(int argc, char** argv)
{
  // Skip argv[1] if it was the name of a parameter file
  int count = (argc > 1 && strchr(argv[1], '=') == 0) ? 2 : 1;

  // Parse the remaining vectors of argv
  while (count < argc)
  {
    char *var, *val;
    parse_equation(argv[count++], var, val);
    if (var != 0 && !set_global(var, val))
      cerr << "Invalid variable [" << var << "] on command line." << endl;
  }

} // parse_argv

// __________________________________________________________________________

int parse_parfile(const simpstrg& parfile)
{
  Ifstream(ifs, parfile);
  if (!ifs)
    return 0;

  char buf[255];
  while (ifs.Getline(buf, 255))
  {
    char *var, *val;
    parse_equation(buf, var, val);
    if (var != 0 && !set_global(var, val))
      cerr << "Invalid variable [" << var << "] in parameter file." << endl;
  }
  return 1;

} // parse_parfile

// __________________________________________________________________________
// For an 'eqn' of the form ' A B C = D E F ', make 'var' equal to "C" and
// 'val' equal to "D E F".  'var' and 'val' end up pointing in to the 'eqn'
// buffer, which is mutilated by this routine.

void parse_equation(char* eqn, char*& var, char*& val)
{
  // Initialize var to something reasonable.
  var = 0;

  // Exit if eqn isn't something reasonable, or starts with a #.
  if (eqn == 0 || eqn[0] == '#')
    return;

  // If we have a newline character in 'eqn', make it a NUL character,
  // then set 'val' pointing to the end of the input string.
  if ((val = strchr(eqn, (int) '\n')) != 0)
    *val-- = '\0';
  else
    val = strchr(eqn, (int) '\0') - 1;

  // Get rid of trailing spaces and tabs in the val - Only
  // important for string variables.
  while (val != (eqn - 1) && (*val == ' ' || *val == '\t'))
    *val-- = '\0';

  // Set 'val' pointing to the character past the '=' character
  // and make the '=' character a NUL character.  Give up if there
  // is no '='.
  if ((val = strchr(eqn, (int) '=')) == 0)
    return;
  *val++ = '\0';

  // Get rid of leading spaces in the val  - Again, this
  // is only important for string variables.
  while (*val == ' ' || *val == '\t')
    val++;

  // Get a pointer to the name of the variable.  This has to be
  // the last word before the equal sign.
  char *nexttok;
  if ((var = strtok(eqn, " \t")) == 0)
    return;
  else
    while ((nexttok = strtok(0, " \t")) != 0)
      var = nexttok;

} // parse_equation

// __________________________________________________________________________
// Set the global variable 'var' equal to 'val' after translating the 
// string 'val' to an appropriate type for 'var'.

int set_global(char* var, char* val)
{
  int ind = 0;
  int done = 0;
  char *brptr = strchr(var, '(');
  if (brptr == 0)
  {
    while (!done)
    {
      if (strcmp(var, varlist[ind].v) == 0)
      {
	switch (varlist[ind].t)
	{
	 case (intvar):
	  *((int *)varlist[ind].p) = atoi(val);
	  break;
	 case (longvar):
	  *((long *)varlist[ind].p) = atol(val);
	  break;
	 case (doublevar):
	  *((double *)varlist[ind].p) = atof(val);
	  break;
	 case (simpstrgvar):
	  *((simpstrg*)varlist[ind].p) = val;
	  break;
	 default:
	  cerr << "Unable to determine type of variable [" << var << "].\n";
	  break;
	}
	return 1;
      }
      done = (varlist[++ind].p == 0);
    }
  }
  else
  {
    *brptr = 0;
    int arrind = atoi(brptr + 1);
    while (!done)
    {
      if (strcmp(var, varlist[ind].v) == 0)
      {
	switch (varlist[ind].t)
	{
	 case (intvar):
	  *(arrind + ((int *)varlist[ind].p)) = atoi(val);
	  break;
	 case (longvar):
	  *(arrind + ((long *)varlist[ind].p)) = atol(val);
	  break;
	 case (doublevar):
	  *(arrind + ((double *)varlist[ind].p)) = atof(val);
	  break;
	 case (simpstrgvar):
	  *(arrind + ((simpstrg*)varlist[ind].p)) = val;
	  break;
	 default:
	  cerr << "Unable to determine type of variable [" << var << "].\n";
	  break;
	}
	return 1;
      }
      done = (varlist[++ind].p == 0);
    }
  }
  return 0;

} // set_global

// __________________________________________________________________________
// Print out all the global variables.  They can be dumped to either cout
// or a file.

void dump_globals(const simpstrg& dumpfile)
{
  if (dumpfile == "")
    return;

  ofstream ofs;
  if (dumpfile != "cout")
  {
    ofs.Open(dumpfile);
    //if (!ofs.is_open())
    if (!ofs)
    {
      cerr << "Unable to open dumpfile [" << dumpfile << "]\n";
      return;
    }
  }

  ostream& os = (dumpfile == "cout") ? (ostream&)cout : (ostream&)ofs;

  int i, ind = 0;

#define GLOBAL(x, y) \
  { os << varlist[ind].v << " = " << *((x *)varlist[ind].p) << endl; ind++; }
#define GLOBALI(x, y, z) \
  { os << varlist[ind].v << " = " << *((x *)varlist[ind].p) << endl; ind++; }
#define GLOBALA(x, y, z, w) \
  { os << varlist[ind].v << "[" << #z << "] = { "; \
    for (i = 0; i < z - 1; ++i) \
      os << *(((x *)varlist[ind].p) + i) << ", "; \
    os << *(((x *)varlist[ind].p) + z - 1) << " "; \
      os << "}\n"; ind++; }
#include "globlist.h"  
#undef GLOBAL
#undef GLOBALI
#undef GLOBALA

} // dump_globals

// __________________________________________________________________________
// globals.c
