// Copyright YALI AMIT, January 15, 2001.

#include <cstring>
#include <iostream>
#include <fstream>


class Graph
{
 public:
  double prune; // percent to prune in dynamic programming,
  igrid graph; // A grid describing edges in graph
  dgrid vert; // Coordinates of model in floating values between 0,1
  dgrid anglim; // Hard constraints on the ranges of the angles in the
               // triangles of the graph, and on the ranges of the lengths.
  int yesgraph; // 1 - Graph provided by user, otherwise linear graph
  int yesanglim; // 1 - Hard constraints provided by user. Otherwise default

  Graph() { }
  ~Graph() { }
  
  void read(ifstream &ifs, double ANGRA);
  void dump();
}; // Graph


void Graph::dump()
{

  cout << " The template points with peeling order "  << endl;
  vert.dump();
  cout << " The graph " << endl;
  graph.dump();
  cout << " The hard constraints " << endl;
  anglim.dump();
 
}

void Graph::read(ifstream &ifs, double ANGRA)
{

  int numv;
  ifs >> numv;
  ifs >> prune;
  ifs >> yesgraph;
  ifs >> yesanglim;

  
  vert.init(numv,3);
  
  for (int i=0; i<vert.dimx(); i++)
    for (int j=0; j<vert.dimy(); j++)
      ifs >> vert(i,j);

  graph.init(numv,numv);
  graph.clear();
  
  // Read in the graph itself 
  if (yesgraph)
    {
      int d;
      for (int i=0; i<graph.dimx(); i++)
	for (int j=0; j<graph.dimy(); j++)
	  {
	  ifs >> d;
	  graph(i,j)=d;
	  }      
    }
  else // Make the default linear graph
    {
      for (int i=0;i<graph.dimx()-2;i++)
	{  
	  graph(i+1,i) = 1;
	  graph(i+2,i) = 1;
	  graph(i,i+1) = 1;
	  graph(i,i+2) = 1;
	}
      graph(graph.dimx()-2,graph.dimx()-1) = 1;
    }
  

  // Limits for variation of each of the angles of each of the triangles
  
  anglim.init(numv-2,15);
  anglim.clear();

  if (yesanglim)
    {
      for (int i=0; i<anglim.dimx(); i++)
	for (int j=0; j<anglim.dimy(); j++)
	  ifs >> anglim(i,j);

           for (int i=0;i<anglim.dimx();i++)
	     {
	       for (int j=anglim.dimy()-3;j<anglim.dimy();j++)
		 anglim(i,j) = -log(anglim(i,j));
	       for (int j=0;j<6;j++)
		 anglim(i,j) = anglim(i,j)/ANGRA;

	       // Hard constrainnts on absolute orientation 
	       for (int j=9;j<=11;j++)
		 anglim(i,j) = anglim(i,j)/ANGRA;
	     }
	   
    }
  else
    {
      for (int i=0;i<anglim.dimx();i++)
	{
	  for (int j=0;j<6;j++) anglim(i,j) = 90./ANGRA;
	  for (int j=9;j<12;j++) anglim(i,j) = 90./ANGRA;
	  for (int j=12;j<15;j++) anglim(i,j) = -log(.2);
	}
    }
 

}  
  
