// Copyright YALI AMIT, January 15, 2001.

#define HISTOGRAM_C


// __________________________________________________________________________
// Include files

#include "includes.h"
#include "histgram.h"

// __________________________________________________________________________
// Inline function

inline double LOG(double x) { return ((x > 0.0) ? log(x) : 0.0); }

// __________________________________________________________________________
void histogram::cop(histogram &m)
{
  delete [] counts;
  counts = new float[m.dim()];
  size = m.dim();
  for (int i=0; i<m.dim(); i++)
     counts[i]=m[i];
}

double histogram::entropy()
{
  double t = total();
  if (t == 0.0)
    return 0.0;

  double ent = 0.0;
  for (int i = 0; i < size; ++i)
  {
    double pcnt = counts[i] / t;
    ent -= pcnt * LOG(pcnt);
  }
  return ent;

} // histogram::entropy

// __________________________________________________________________________

double histogram::gini()
{
  double t = total();
  if (t == 0.0)
    return 0.0;

  double gini = 0.0;
  for (int i = 0; i < size; ++i)
  {
    double pcnt1 = counts[i] / t;
    for (int j = i + 1; j < size; ++j)
    {
      double pcnt2 = counts[j] / t;
      gini += 2 * pcnt1 * pcnt2;
    }
  }
  return gini;

} // histogram::gini

// __________________________________________________________________________
double histogram::unsup()
{
  return LOG(total());
}


double histogram::kolsmir()
{
  double t = total();
  if (t == 0.0)
    return 0.0;

  double kolsmir = 0.0;
  for (int i = 0; i < size; ++i)
  {
    double pcnt1 = counts[i] / t;
    for (int j = i + 1; j < size; ++j)
    {
      double pcnt2 = counts[j] / t;
      kolsmir += 2 * pcnt1 * pcnt2;
    }
  }
  return kolsmir;

} // histogram::kolsmir

// __________________________________________________________________________
// histgram.C









