// Copyright YALI AMIT, January 15, 2001.

#ifndef IMAGE_H
#define IMAGE_H


// _________________________________________________________________________
// Includes

#include <iostream>

#include "grid.h"
#include "landmark.h"

class simpstrg;

// __________________________________________________________________________
// data structure used to store and manipulate images

class image : public ucgrid, public landmarks
{
 private:
  int tr;

 public:
  image() : ucgrid(), landmarks(), tr(0) { }
  image(int x) : ucgrid(x), tr(0) { }
  image(int x, int y) : ucgrid(x, y), tr(0) { }
  image(const image &i) : ucgrid(i), tr(i.tr) { }
  image& operator=(const image &i) { 
    ucgrid::operator=(i); tr = i.tr; return *this;
  }
  ~image() { }

  int truth() const { return tr; }
  void set_truth(int x) { tr = x; }
  int read(const char* file) { return ucgrid::read(file); }
  int read(int index, int flag = 0);
  int make_big(int sizex, int sizey, int maxent, int num,
	       int goodent = 0, int numgood = 0);
  int make_occl(int sizex, int sizey, int maxent, int num);
  image& thin(int masksize = 1);
  image& blur_downscale(double ratio = 0, double sd = 1);
  image& subtract_lsq(const ucgrid& mask = ucgrid(), 
                      int xlow = 0, int ylow = 0);
  image& medfilt(int size = 1, int times = 1);
  image& meanfilt(int size = 1, int times = 1);
  image& histeq(int maxval = 255);
  image& make_spots(int num_spots, int spot_size);
  image& imblur(int blur, int& x1, int& y1, int& x2, int& y2,
		int& x3, int& y3);
  image& shift(int sx, int sy, int& x1, int& y1, int& x2, int& y2,
	       int& x3, int& y3);
  image& imslant(int slnt);
    image& linearfit();
  image& equalize();
  image& downscale(int size);
  image& imtrim(int amount, int& x1, int& y1, int& x2, int& y2,
		int& x3, int& y3);
  image& std_process(double skw, double rrr, double slnt, double scl, 
		     int& x1, int& y1, int& x2, int& y2, int& x3, int& y3);

}; // class image

typedef vector<image> images;

// __________________________________________________________________________
// Function prototypes

extern int data_dir_open(ifstream& ifs, const simpstrg& datafile, int abrt = 1);
extern int get_truth(int digit);
extern void random_select(lgrid& randgrid, int max, int num);
void write_images(const vector<image>& im, int numdigits);
void write_images(image* ims, int numim, char* name);
void view_image_list();

// __________________________________________________________________________
// image.h

#endif
