// Copyright YALI AMIT, January 15, 2001.

#ifndef LANDMARK_H
#define LANDMARK_H


#include <cstring>
#include <iostream>
#include <fstream>

#include "grid.h"

// __________________________________________________________________________
// A single landmark (code)

class landmark
{
 public:
  int index;    // Which landmark is it?
  int cx, cy;   // The center of mass of the landmark

  landmark() : index(0), cx(0), cy(0) { }
};

typedef vector<landmark> lmarklist;

// __________________________________________________________________________
// Arrays of landmarks and instances



class landmarks
{
  friend class Edge;
 public:

 private:
  int value;
  igrid typecounts;
  igrid typeoffsets;
  igrid locations;
  ucgrid spillflags;

  landmarks(const landmarks&) { exit(1); }
  void operator=(const landmarks&) const { exit(1); }

 public:
  // This should be private, but needs some member functions first
  int coords[6];

  landmarks() : value(0) { memset(coords, 0, 6 * Sint); }
  void init_counts_offsets(int num) 
  { typecounts.init(num); typeoffsets.init(num); }
  void clear_counts() { typecounts.clear(); }
  void init_locations(int num) 
  { locations.init(num, 2); spillflags.init(num); spillflags.clear(); }

  const int& truth() const { return (value); }
  int truth(int nn) {return (value - nn);}
  void set_truth(int val) { value = val; }

  int total_types() const { return typecounts.dimx(); }
  int& type_count(int t) const { return typecounts(t); }
  int& type_offset(int t) const { return typeoffsets(t); }

  int total_instances() const { return locations.dimx(); }
  int& location_x(int t) const { return locations(t, 0); }
  int& location_y(int t) const { return locations(t, 1); }
  int& location_x(int t) { return locations(t, 0); }
  int& location_y(int t) { return locations(t, 1); }

  int& location_x(int t, int ind) const
  { return location_x(typeoffsets(t) + ind); }
  int& location_y(int t, int ind) const
  { return location_y(typeoffsets(t) + ind); }
  uchar& spill(int t, int ind) const
  { return spillflags(typeoffsets(t) + ind); }

  int read(int index);
  int read(ifstream& ifs);
  int write(ofstream& ofs);
  int size();
  void dump();
  void merge(landmarks &ne);
  void setup_from_lmarklist(lmarklist& land, int numlands);

}; // class landmarks

// __________________________________________________________________________
// Function prototypes
typedef vector<landmarks> lmarkslist;
void write_lands(const char* landfile, int count, lmarkslist& lands);
void read_lands(const char* landfile, int count, lmarkslist& lands);
void cluster_landmarks(lmarklist& lands, int clustsize, 
		       int numtypes, int& total);
void clust_landmarks(lmarklist& lands, int clustsize, 
		       int numtypes, int& total);

// __________________________________________________________________________
// landmark.h



#endif // LANDMARK_H

