// Copyright YALI AMIT, January 15, 2001.

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>
#include "simpstrg.h"
#include "includes.h"
#include "pixel.h"
#include "viewport.h"
#include "landmark.h"
#include "trainstat.h"
#include "ltatree.h"
#include "globals.h"
static int compar(const void *a, const void* b);

static int compar(const void *a, const void* b)
{
  return (*((const double*) a) < *((const double*) b));
}


int ltatree::read(const char* treename)
{

  Ifstream(ifs, treename);
  if (!ifs)
    return 0;

  char buf[5];
  ifs.read(buf, 4);
  buf[4] = 0;
  if (strcmp(buf, "TR12") != 0)
  {
    cerr << treename << " is not an lta tree file\n";
    exit(1);
  }
 
  varread(ifs, locationx);
  varread(ifs, locationy);
  varread(ifs, tnumang);
  varread(ifs, tnear);
 if (locationx==-1 && locationy>0)
   althistsize=locationy;

 
  if (!ifs.eof())
  {
    maketop();
    readnodes(top(), ifs);
   if (opt==2)
    updatehists(top());
  }

  return 1;

} // ltatree::read

void ltatree::readnodes(const ltaiter& curr, ifstream& ifs)
{
  curr->que.read(ifs);
  curr->vertlist.read(ifs);

  if (curr->que.typ == 0)
   {
   if (althistsize < ltahistsize && althistsize>0 )
     {
       histogram temph(althistsize*2);
       temph.read(ifs);
       for (int i=0; i<althistsize; i++)
           curr->hist[(int)(temph[2*i+1])]=temph[2*i];
     }
     else
        curr->hist.read(ifs);
  }
  else
  {
    curr.makechildren();
    readnodes(curr.no(), ifs);
    readnodes(curr.yes(), ifs);
  }

} // ltatree::readnodes

void ltatree::write(ofstream& ofs)
{
 ofs << "TR12";
  varwrite(ofs, locationx);
  varwrite(ofs, locationy);
  varwrite(ofs, tnumang);
  varwrite(ofs, tnear);
  writenodes(top(), ofs);
}

void ltatree::read(ifstream& ifs)
{

 
  char buf[5];
  ifs.read(buf, 4);
  buf[4] = 0;
  if (strcmp(buf, "TR12") != 0)
    {
      exit(1);
    }
  
  varread(ifs, locationx);
  varread(ifs, locationy);
  varread(ifs, tnumang);
  varread(ifs, tnear);
  if (locationx==-1 && locationy>0)
    althistsize=locationy;
  
  
 
  if (!ifs.eof())
    {
      maketop();
      readnodes(top(), ifs);
    } 
}

void ltatree::write(const char* treename)
{
  ofstream ofs(treename);
  if (!ofs) {
    cerr << "Unable to write to treefile [" << treename << "]\n";
    exit(1);
  }

  ofs << "TR12";
  varwrite(ofs, locationx);
  varwrite(ofs, locationy);
  varwrite(ofs, tnumang);
  varwrite(ofs, tnear);
  writenodes(top(), ofs);

} // ltatree::write

// __________________________________________________________________________

void ltatree::writenodes(const ltaiter& curr, ofstream& ofs)
{
  if (!curr)
    return;

  curr->que.write(ofs);
  curr->vertlist.write(ofs);

  if (curr.leaf())
  {
    if (althistsize < ltahistsize/2 && althistsize>0 )
      {
        dgrid temp(2,ltahistsize);
        for (int i=0; i<ltahistsize; i++)
	  {
            temp(0,i)=curr->hist[i];
            temp(1,i)=i;
	  }
        temp.sort_columns(ltahistsize,compar);
	histogram temph(althistsize*2);
	//cout << " node " << endl;
        for (int i=0; i<althistsize; i++)
	  {
	    temph[i+i]=temp(0,i); temph[i+i+1]=temp(1,i);
	  }

        temph.write(ofs);
      }
     else
       curr->hist.write(ofs);
  }
  else
  {
    writenodes(curr.no(), ofs);
    writenodes(curr.yes(), ofs);
  }

} // ltatree::writenodes
// __________________________________________________________________________

void ltatree::dump(ostream& os)
{

  os.setf(ios::fixed, ios::floatfield);

  stats = new ltatreestats;
  calc_stats(top());

  dumpnodes(top(), os);
  stats->dump(1);

  os << "max yes depth = [" << maxyesdepth() << "]\n";

  os << "\n\nLocation = [" << locationx << ", " << locationy << "]\n";

  os << "\n\nRelations:\n";

  int i;
  for (i = 0; i < stats->taguse.dimx(); ++i)
    if (stats->taguse(i) > 0)
      os << "Tag " << i << " : " << stats->taguse(i) << " times\n";

  os << endl;
  for (i = 0; i < stats->anguse.dimx(); ++i)
    if (stats->anguse(i) > 0)
      os << "Angle " << i << " : " << stats->anguse(i) << " times\n";

  os << endl;
  for (i = 0; i < stats->toluse.dimx(); ++i)
    if (stats->toluse(i) > 0)
      os << "Tolerance " << i << " : " << stats->toluse(i) << " times\n";

  os << endl;
  for (i = 0; i < stats->disuse.dimx(); ++i)
    if (stats->disuse(i) > 0)
      os << "Distance " << i << " : " << stats->disuse(i) << " times\n";

  os << " near =" << tnear << endl;
  os << " numang = " << tnumang << endl;
  os << endl;
} // ltatree::dump

// __________________________________________________________________________

void ltatree::calc_stats(const ltaiter& diter)
{
  if (diter.leaf())
  {
    stats->total += diter->hist.total();
    stats->correct += diter->hist[diter->hist.mode()];
    stats->wtdepth +=int( diter->hist.total() * diter.depth());
    stats->wtent += diter->hist.entropy() * diter->hist.total();
  }
  else
  {
   if (diter->que.typ != 10 && diter->que.typ != 7)
     {
       ++stats->queuse(diter->que.typ);
       ++stats->anguse(diter->que.r1.ang);
       ++stats->toluse(diter->que.r1.tol);
       ++stats->disuse(diter->que.r1.dis);
     }
    switch(diter->que.typ)
    {
    case 1:
    case 3:
    case 5:
      ++stats->taguse(diter->que.r1.land1);
      ++stats->taguse(diter->que.r1.land2);
      break;
    case 10:
    case 2:
    case 4:
    case 6:
    case 7:
      ++stats->taguse(diter->que.r1.land1);
      break;
    default:
      cerr << "Unknown question type [" << diter->que.typ 
	   << "] in calc_stats\n";
      exit(1);
    }
    calc_stats(diter.no());
    calc_stats(diter.yes());
  }

} // ltatree::calc_stats
// __________________________________________________________________________

void ltatree::dumpnodes(const ltaiter& d, ostream& os)
{
  int i;
  for (i = 0; i < 78; ++i) 
    os << "-";
  os << "\nDepth: [" << setw(2) << d.depth()
    << "], History: [" << d.history() << "]\n";

  if (d.branch()) {
    if (d->que.typ==1 || d->que.typ==2 || d->que.typ==5 || d->que.typ==6) {
      os << "Question = Type: [" << setw(2) << d->que.typ << "], Ang: [" 
	 << "index " << int(d->que.r1.ang)<<"] [" 
	 << setw(6) << setprecision(2) << d->que.r1.relangle() * 180 / PI
	 << "], Tol: [" << setw(6) << setprecision(2) 
	 << d->que.r1.reltolerance() * 180 / PI << "], Dis: [" << setw(1) 
	 << int(d->que.r1.dis) << "]\n";
      if (d->que.typ == 1 || d->que.typ==5)
	os << "Edge types " << d->que.r1.land1 << " "
	   << d->que.r1.land2 << endl;
      if (d->que.typ == 2 || d->que.typ==6) {
	os << "New edge type " << d->que.r1.land1 << endl;
	if (class_tree)
	  os << " Connected to " << (int)(d->que.r1.v1) <<endl;
      }      
    }
    else if (d->que.typ==10) {
      os << "Question = Type: [" << setw(2) << d->que.typ << " ]" 
	 << " Dimensions " << int(d->que.r1.dis) << " Downsample " 
	 << (int)d->que.r1.tol << " Spread " << (int)d->que.r1.ang << endl;    
      os << "Edge types " << d->que.r1.land1 << " Coordinates " 
	 << d->que.r1.land2/int(d->que.r1.dis) << " " 
	 << d->que.r1.land2 % int(d->que.r1.dis) << endl;
    }
    else if (d->que.typ==7) {
      os << "Question = Type: [" << setw(2) << d->que.typ << " ]" 
	 << " Dimensions " << int(d->que.r1.dis) << " Downsample " 
	 << (int)d->que.r1.tol << " Spread " << (int)d->que.r1.ang << endl;    
      os << "Edge types " << d->que.r1.land1 << " Coordinates " 
	 << d->que.r1.land2/int(d->que.r1.dis) << " " 
	 << d->que.r1.land2 % int(d->que.r1.dis) << endl; 
      os << "Cutoff " << (int)d->que.r1.v1 << endl;
    }
    else if (d->que.typ==3 || d->que.typ==4) { 
      os << "Question = Type: [" << setw(2) << d->que.typ << " ]" 
	 << " coords " << int(d->que.r1.v1) << " " << (int)d->que.r1.v2
	 << endl;
      if (d->que.typ == 3)
	os << "Edge types " << d->que.r1.land1 << " "
	   << d->que.r1.land2 << endl;
      if (d->que.typ == 4)
	os << "New edge type " << d->que.r1.land1 << endl;
    }
  }
  os << "Data at node: " << d->hist.total() << endl;
  if (class_tree)
    {
      if (!d.branch())
	{
	cout << "Terminal ";
        if ( d->hist[d->hist.mode()]==d->hist[d->hist.mode2()])
	  cout << " TIE ";
	}
    for (i = 0; i < d->hist.dim(); ++i)
      os << i << "[" << d->hist[i] << "] ";
    os << endl;
    }
  if (((ltausedepth == -1 && !d.branch()) || d.depth() == ltausedepth) && 
      (atoi(&(d.history())[d.depth() - 1]) == 1))
    cout << "Frequency = " << d->hist.total() << endl;

  if (d.branch()) 
  {
    dumpnodes(d.yes(), os);
    dumpnodes(d.no(), os);
  }
  else
  {
//    if (atoi(&(d.history())[d.depth()-1]) == 1)
//      cout << "Frequency = " << d->hist.total() << endl;
  }
} // ltatree::dumpnodes

void ltatreestats::dump(int n) const
{
  cout.setf(ios::fixed, ios::floatfield);
  cout << "\nAverage classification rate: " << setprecision(2) 
    << 100.0 * correct / total << endl;
  cout << "Average depth: " << 1.0 * wtdepth / total << endl;
  cout << "Average entropy: " << wtent / total << endl;
  cout << "\nQuestion usage:\n";
  int i, quetotal = 0;
  for (i = 0; i < 20; ++i)
    quetotal += queuse(i);
  for (i = 0; i < 20; ++i)
    if (queuse(i) > 0)
    {
      cout << "Question " << i << ": " << setw(7) << 1.0 * queuse(i) / n
	<< " (" << setw(6) << 100.0 * queuse(i) / quetotal << " %)\n";
    }
} // ltatreestats::dump

// Used for lta's only, all these simple yes trees are stored in one file.
int read_ltatrees(ltatrees& ltrees, simpstrg direc)
{

simpstrg file=direc+simpstrg("/ltatrees");
ifstream ifs((const char*)file);

if (!ifs)
  {
    cerr << " Can't open Trees file " << file << endl;
     exit(2);
  }

int numl;
varread(ifs,numl);

 if (numl > 0)
   {
     ltrees.init(numl);
     for (int i = 0; i < numl; ++i)
       {
          ltrees[i].read(ifs);
       }
     // If a real tree and not a just a record of an edge type and a location
     if (ltrees[0].maxyesdepth()>0)
       {
	 if (dum1==100)
	   ltrees[0].dump();
	 numang = ltrees[0].tnumang;
	 if (near>=0)
	 near = ltrees[0].tnear;
         else 
	   near=-near;
	 make_rel_masks();
       }
    }
 else
   onlyedges=1;

 cout << " NUMANG " << numang << " NEAR " << near << endl;

return numl;
}

// Used for lta's only
void write_ltatrees(ltatrees& ltrees)
{
ofstream ofs("ltatrees");

int nn=ltrees.dim();
varwrite(ofs,nn);
for (int i = 0; i < ltrees.dim(); ++i)
	{
          ltrees[i].write(ofs);
	}
}


void old_to_new_ltatrees()
{
void read_ltatrees(ltatrees& ltrees);
ltatrees ltrees;
read_ltatrees(ltrees);
write_ltatrees(ltrees);
}

// Old format not used any more.
void read_ltatrees(ltatrees& ltrees)
{
  if (numltas > 0)
    {
      ltrees.init(numltas);
      for (int i = 0; i < numltas; ++i)
	{
         
	  if (!ltrees[i].read(ltatreedir + simpstrg("/tree") + 
			      ltos(i + firstlta)))
	    {
	      cerr << "Unable to read in the ltrees\n";
	      exit(1);
	    }
	}
      numang = ltrees[0].tnumang;
      near = ltrees[0].tnear;
      make_rel_masks();
    }
  else
    onlyedges=1;
}

