// Copyright YALI AMIT, January 15, 2001.


// Pixel member functions 

// __________________________________________________________________________

#include <cctype>

#include "includes.h"
#include "pixel.h"


Pixel operator- (const Pixel& a)
{ return Pixel(-a.n[0],-a.n[1]); }

Pixel operator-(const Pixel& a, const Pixel& b)
{ return Pixel(a.n[0] - b.n[0], a.n[1] - b.n[1]); }

Pixel operator+(const Pixel& a, const Pixel& b)
{ return Pixel(a.n[0] + b.n[0], a.n[1] + b.n[1]); }

Pixel operator* (const Pixel& a, const int d)
{ return Pixel((int)rint(d*(double)a.n[0]),(int)rint(d*(double)a.n[1])); }

Pixel operator* (const Pixel& a, const double d)
{ return Pixel((int)rint(d*(double)a.n[0]),(int)rint(d*(double)a.n[1])); }

Pixel operator* (const double d, const Pixel& a )
{ return Pixel((int)rint(d*(double)a.n[0]),(int)rint(d*(double)a.n[1])); }

Pixel operator* (const int d, const Pixel& a)
{ return a*d; }

Pixel operator / (const Pixel& a, const int d)
{ return Pixel(a.n[0]/d, a.n[1]/d); }

ostream& operator<< (ostream& s, const Pixel& v)
{ return s << "(" << v.n[0] <<"," << v.n[1] << ")"; }

istream& operator>> (istream& s, Pixel& v) {
    Pixel	v_tmp;
    char	c = ' ';

    while (isspace(c))
	s >> c;
    // The Pixels can be formatted either as x y or | x y |
    if (c == '|') {
	s >> v_tmp[0] >> v_tmp[1];
	while (s >> c && isspace(c)) ;
// KJW - Changed this... What is set?
	if (c != '|') ;
	    //s.set(_bad);
	    //s.set(4);
	}
    else {
	s.putback(c);
	s >> v_tmp[0] >> v_tmp[1];
	}
    if (s)
	v = v_tmp;
    return s;
}

int DistInf(Pixel & P,Pixel & Q)
{
  int x = abs(P[0]-Q[0]);
  int y = abs(P[1]-Q[1]);
  return (x < y ? y : x);
}

Pixel Pixel::Rot90()
{
  return Pixel(-n[1], n[0]);
}

double Pixel::Norm()
{
  return sqrt(n[0]*n[0]+n[1]*n[1]);
}

double Pixel::Cos()
{
  return n[0]/Norm();
}

double Pixel::Cos(Pixel P)
{
  return (double)(P[0]*n[0]+P[1]*n[1])/(P.Norm()*Norm());
}

double Pixel::Sin()
{
  return n[1] / Norm();
}
Pixel operator / (const Pixel& a, const Pixel& b)
{
 double l = b.n[0]*b.n[0] + b.n[1]*b.n[1];
 return Pixel( (a.n[0]*b.n[0]+a.n[1]*b.n[1])/l,
                (a.n[1]*b.n[0]-a.n[0]*b.n[1])/l);
}

Pixel operator * (const Pixel& a, const Pixel& b)
{
return Pixel (a.n[0]*b.n[0]-a.n[1]*b.n[1], a.n[0]*b.n[1]+a.n[1]*b.n[0]);
}

double Pixel::angle()
{
 return atan2(n[0],n[1]);
}

// __________________________________________________________________________
// Calculate the affine transformation which sends p1->q1, p2->q2, and p3->q3.
// Store it in a 2x3 matrix so that for any p, p goes to q, with q given by
//   q[i] = c[i][0] * p[0] + c[i][1] * p[1] + c[i][2];

 affinetrans::affinetrans(const Pixel& p1,const  Pixel& p2,const  Pixel& p3,
			  const Pixel& q1,const  Pixel& q2,const  Pixel& q3)
{
  double det = ((p2[0] * p3[1] - p3[0] * p2[1]) +
		(p3[0] * p1[1] - p1[0] * p3[1]) +
		(p1[0] * p2[1] - p2[0] * p1[1]));

  if (fabs(det) < 0.0001)
  {
    coefs[0][0] = coefs[0][1] = coefs[0][2] = 
      coefs[1][0] = coefs[1][1] = coefs[1][2] = 0;
    return;
  }

  double p00 = (p2[1] - p3[1]) / det;
  double p10 = (p3[1] - p1[1]) / det;
  double p20 = (p1[1] - p2[1]) / det;
  double p01 = (p3[0] - p2[0]) / det;
  double p11 = (p1[0] - p3[0]) / det;
  double p21 = (p2[0] - p1[0]) / det;
  double p02 = (p2[0] * p3[1] - p3[0] * p2[1]) / det;
  double p12 = (p3[0] * p1[1] - p1[0] * p3[1]) / det;
  double p22 = (p1[0] * p2[1] - p2[0] * p1[1]) / det;

  coefs[0][0] = q1[0] * p00 + q2[0] * p10 + q3[0] * p20;
  coefs[0][1] = q1[0] * p01 + q2[0] * p11 + q3[0] * p21;
  coefs[0][2] = q1[0] * p02 + q2[0] * p12 + q3[0] * p22;
  coefs[1][0] = q1[1] * p00 + q2[1] * p10 + q3[1] * p20;
  coefs[1][1] = q1[1] * p01 + q2[1] * p11 + q3[1] * p21;
  coefs[1][2] = q1[1] * p02 + q2[1] * p12 + q3[1] * p22;

} // affinetrans::affinetrans



// __________________________________________________________________________
// Calculate the pixel obtained by applying an affine transformation.

void Pixel::trans(const affinetrans &at, double &x, double &y)
{
    x = at.coefs[0][0]*n[0] + at.coefs[0][1]*n[1] + at.coefs[0][2];
    y = at.coefs[1][0]*n[0] + at.coefs[1][1]*n[1] + at.coefs[1][2];
}

int Pixel::bilinear_interp(const affinetrans &at, image &im)
{
  double x,y;

   trans(at, x, y);
   int ix=int(x); 
   int iy=int(y);
   double a = x-ix;
   double b = y-iy;
    int g=int(rint(a*b*im.get(ix+1,iy+1)+(1.-a)*b*im.get(ix, iy+1)
	   +a*(1.-b)*im.get(ix+1,iy)+(1.-a)*(1.-b)*im.get(ix,iy)));
   return(g);
}

Pixel affinetrans::operator*(const Pixel& p)
{
  return Pixel(rint(coefs[0][0] * p[0] + coefs[0][1] * p[1] + coefs[0][2]),
	       rint(coefs[1][0] * p[0] + coefs[1][1] * p[1] + coefs[1][2]));
}

// __________________________________________________________________________
// Calculate part of the transformation which sends three pixels to
// three other pixels q1, q2, q3.  This part just depends on the qi and
// can be used with different initial points.
partialtrans::partialtrans(const Pixel& p1,const  Pixel& p2,const  Pixel& p3)
   { 
   ptr(p1, p2, p3);
   }

void  partialtrans::ptr(const Pixel& p1,const  Pixel& p2,const  Pixel& p3)
{
  double det = ((p2[0] * p3[1] - p3[0] * p2[1]) +
		(p3[0] * p1[1] - p1[0] * p3[1]) +
		(p1[0] * p2[1] - p2[0] * p1[1]));

  if (fabs(det) < 0.0001)
  {
    coefs[0][0] = coefs[0][1] = coefs[0][2] =
      coefs[1][0] = coefs[1][1] = coefs[1][2] =
	coefs[2][0] = coefs[2][1] = coefs[2][2] = 0;
    return;
  }

  coefs[0][0] = (p2[1] - p3[1]) / det;
  coefs[1][0] = (p3[1] - p1[1]) / det;
  coefs[2][0] = (p1[1] - p2[1]) / det;
  coefs[0][1] = (p3[0] - p2[0]) / det;
  coefs[1][1] = (p1[0] - p3[0]) / det;
  coefs[2][1] = (p2[0] - p1[0]) / det;
  coefs[0][2] = (p2[0] * p3[1] - p3[0] * p2[1]) / det;
  coefs[1][2] = (p3[0] * p1[1] - p1[0] * p3[1]) / det;
  coefs[2][2] = (p1[0] * p2[1] - p2[0] * p1[1]) / det;

}


// __________________________________________________________________________

 affinetrans::affinetrans( partialtrans& qtrans,
			   const Pixel& q1,const  Pixel& q2,const  Pixel& q3)
{
  coefs[0][0] = q1[0] * qtrans.coefs[0][0] + q2[0] * 
    qtrans.coefs[1][0] + q3[0] * qtrans.coefs[2][0];
  coefs[0][1] = q1[0] * qtrans.coefs[0][1] + q2[0] * 
    qtrans.coefs[1][1] + q3[0] * qtrans.coefs[2][1];
  coefs[0][2] = q1[0] * qtrans.coefs[0][2] + q2[0] * 
    qtrans.coefs[1][2] + q3[0] * qtrans.coefs[2][2];
  coefs[1][0] = q1[1] * qtrans.coefs[0][0] + q2[1] * 
    qtrans.coefs[1][0] + q3[1] * qtrans.coefs[2][0];
  coefs[1][1] = q1[1] * qtrans.coefs[0][1] + q2[1] * 
    qtrans.coefs[1][1] + q3[1] * qtrans.coefs[2][1];
  coefs[1][2] = q1[1] * qtrans.coefs[0][2] + q2[1] * 
    qtrans.coefs[1][2] + q3[1] * qtrans.coefs[2][2];
}

void  affinetrans::affine( partialtrans& qtrans,
			   const Pixel& q1,const  Pixel& q2,const  Pixel& q3)
{
coefs[0][0] = q1[0] * qtrans.coefs[0][0] + q2[0] * 
    qtrans.coefs[1][0] + q3[0] * qtrans.coefs[2][0];
  coefs[0][1] = q1[0] * qtrans.coefs[0][1] + q2[0] * 
    qtrans.coefs[1][1] + q3[0] * qtrans.coefs[2][1];
  coefs[0][2] = q1[0] * qtrans.coefs[0][2] + q2[0] * 
    qtrans.coefs[1][2] + q3[0] * qtrans.coefs[2][2];
  coefs[1][0] = q1[1] * qtrans.coefs[0][0] + q2[1] * 
    qtrans.coefs[1][0] + q3[1] * qtrans.coefs[2][0];
  coefs[1][1] = q1[1] * qtrans.coefs[0][1] + q2[1] * 
    qtrans.coefs[1][1] + q3[1] * qtrans.coefs[2][1];
  coefs[1][2] = q1[1] * qtrans.coefs[0][2] + q2[1] * 
    qtrans.coefs[1][2] + q3[1] * qtrans.coefs[2][2];
}
// __________________________________________________________________________

