// Copyright YALI AMIT, January 15, 2001.

#ifndef PIXEL_H
#define PIXEL_H

// __________________________________________________________________________

#include <iostream>
#include "vector.h"
#include "image.h"

class affinetrans;

class Pixel
{
 protected:
  int  n[2];

 public: 
  // Constructors
  Pixel() {}
  Pixel(int x, int y) { n[0] = x; n[1] = y; }
  Pixel(double x, double y) { n[0] = int(x); n[1] = int(y); }
  Pixel(int d) { n[0] = n[1] = d; }
  Pixel(const Pixel& v) { n[0] = v.n[0]; n[1] = v.n[1]; }
  inline Pixel(const char* FileLine);  

  // Assignment operators
  Pixel& operator=(const Pixel& v) 
  { n[0] = v.n[0]; n[1] = v.n[1]; return *this; }
  Pixel& operator=(int* v)
  { n[0] = v[0]; n[1] = v[1]; return *this; }
  Pixel& operator+=(const Pixel& v)
  { n[0] += v.n[0]; n[1] += v.n[1]; return *this; }
  Pixel& operator-=(const Pixel& v)
  { n[0] -= v.n[0]; n[1] -= v.n[1]; return *this; }
  Pixel& operator*=(const int d)
  { n[0] *= d; n[1] *= d; return *this; }
  Pixel& operator/=(const int d)
  { n[0] /= d; n[1] /= d; return *this; }
  //Pixel operator+(const Pixel& v) { return ((*this) += v); }
  //Pixel operator-(const Pixel& v) { return ((*this) -= v); }
  int& operator[](int i) { return n[i]; }
  int operator[](int i) const { return n[i]; }
  void get(const char* FileLine);

  // geometrical trasformations
  Pixel SymFirstDiag() { return Pixel(n[1], n[0]); }
  Pixel SymAxisX() { return Pixel(-n[0], n[1]); }
  Pixel SymAxisY() { return Pixel(n[0], -n[1]); }
  Pixel SymCenter() { return Pixel(-n[0], -n[1]); }
  Pixel TransXY(int x, int y) { return Pixel(n[0] + x, n[1] + y); }
  Pixel TransX(int x) { return TransXY(x, 0); }
  Pixel TransY(int y) { return TransXY(0, y); }
  Pixel Rot90();
  int bilinear_interp(const affinetrans &at, image &im);
  void trans(const affinetrans &at, double &x, double &y);
  double Cos();
  double Cos(Pixel P);
  double Sin();
  double angle();
  double Norm();

  // friends

  friend Pixel operator- (const Pixel& v);
  friend Pixel operator+ (const Pixel& a, const Pixel& b);
  friend Pixel operator- (const Pixel& a, const Pixel& b);
  friend Pixel operator* (const Pixel& a, const int d);
  friend Pixel operator* (const Pixel& a, const double d);
  friend Pixel operator* (const double d, const Pixel& a);
  friend Pixel operator* (const int d, const Pixel& a);
  friend Pixel operator/ (const Pixel& a, const int d);
  friend Pixel operator/ (const Pixel& a, const Pixel& b);
  friend Pixel operator* (const Pixel& a, const Pixel& b);

  friend ostream& operator<< (ostream& s, const Pixel& v);
  friend istream& operator>> (istream& s, Pixel& v);
  friend int DistInf(Pixel & P,Pixel & Q);

}; // class Pixel

// __________________________________________________________________________

class partialtrans;

class affinetrans
{
 private:



 public:
  affinetrans() { }
  double coefs[2][3];
  affinetrans(const Pixel& p1,const  Pixel& p2,const  Pixel& p3,
	      const Pixel& q1,const  Pixel& q2,const  Pixel& q3);
  affinetrans( partialtrans& ptrans,
	       const Pixel& q1,const  Pixel& q2,const  Pixel& q3);
  Pixel operator*(const Pixel& p);
  void affine( partialtrans& ptrans,
	       const Pixel& q1,const  Pixel& q2,const  Pixel& q3);
}; // class affinetrans

class partialtrans
{
  friend class affinetrans;

 
 public:
  double coefs[3][3];
  partialtrans() {for (int i=0; i<3; i++)
           for (int j=0; j<3; j++)
         coefs[i][j]=0;  }
 void ptr(const Pixel& p1,const  Pixel& p2,const  Pixel& p3); 
 partialtrans(const Pixel& p1,const  Pixel& p2,const  Pixel& p3); 
 
 
}; // class partialtrans

typedef vector<Pixel> Pixels;
typedef vector<Pixels> Pixelss;

#endif // PIXEL_H

