// Copyright YALI AMIT, January 15, 2001.

#ifndef TREE_H
#define TREE_H


// __________________________________________________________________________
// Include files

#include <iostream>
#include "simpstrg.h"

// __________________________________________________________________________

template <class T>
class treeiterator
{
  class treenode
  {
   public:
    treenode *parent, *no, *yes;
    T data;

    treenode(treenode* p = 0, treenode* n = 0, treenode* y = 0) : 
        parent(p), no(n), yes(y) { }
  };

 private:
  treenode *p;

 public:
  treeiterator(treenode *x = 0) : p(x) { }
  treeiterator(const treeiterator<T>& t) : p(t.p) { }
  void freenode() const { delete p;  }
  void makenode(treenode* pa = 0, treenode* n = 0, treenode* y = 0)
    { p = new treenode(pa, n, y); }
  void makechildren() const
    { assert(p);
      p->no = new treenode(p, 0, 0);
      p->yes = new treenode(p, 0, 0);
    }
  void operator=(const treeiterator<T>& t) { p = t.p; }
  int operator==(const treeiterator<T>& t) const { return (p == t.p); }
  operator int() const { return (p ? 1 : 0); }
  T* operator->() const { assert(p); return &p->data; }
  treeiterator<T> parent() const { return treeiterator<T>(p ? p->parent : 0); }
  treeiterator<T> no() const { return treeiterator<T>(p ? p->no : 0); }
  treeiterator<T> yes() const { return treeiterator<T>(p ? p->yes : 0); }
  int branch() const { return (p && p->yes != 0 && p->no != 0); }
  int leaf() const { return (p && p->yes == 0 && p->no == 0); }
  int yn() const { return (p && p->parent ? p == p->parent->yes : 0); }
  int depth() const { return (p && p->parent ? 1 + parent().depth() : 0); }
  simpstrg history() const {
    return (p && p->parent ? (parent().history() + (yn() ? "1" : "0")) : 
            simpstrg("")); 
  }

}; // class treeiterator<T>

// __________________________________________________________________________

template <class T>
class tree
{
private:
  treeiterator<T> tt;
  void freenodes(const treeiterator<T>& t);

public:
  tree() { }
  ~tree() { freenodes(tt); }
  void maketop() { tt.makenode(); }
  const treeiterator<T> top() const { return tt; }
};

template <class T>
inline void tree<T>::freenodes(const treeiterator<T>& t)
{
  if (!t)
    return;

  if (t.branch()) {
    freenodes(t.yes());
    freenodes(t.no());
  }

  t.freenode();

} // tree::freenodes

// __________________________________________________________________________
// tree.h
#endif

