// Copyright YALI AMIT, January 15, 2001.

#ifndef XOBJECT_C
#define XOBJECT_C

// __________________________________________________________________________


// __________________________________________________________________________

#include "includes.h"
#ifndef NOX

#include <iostream>
//#include <sys/types.h>
//#include <unistd.h>
#define String XString
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#undef String

#include "xobject.h"
#include "globals.h"

// __________________________________________________________________________
// Define static xobject members

int xobject::objcount = 0;
int xobject::inited = 0;

Display *xobject::display;
int xobject::screen;

Colormap xobject::cmap;
XColor xobject::grays[256];
XColor xobject::defs[8];

XSetWindowAttributes xobject::xswa;
Atom xobject::wmdw;
XClassHint xobject::xch;
XSizeHints xobject::xsh;
XWMHints xobject::xwmh;

XFontStruct *xobject::xfs;
int xobject::fontheight;
int xobject::fontwidth;

// __________________________________________________________________________
// Establish a connection to a display and set up defaults.

int xobject::initX()
{
  // Return now if this routine was already successfully called.
  if (inited)
    return 1;

  // Open the default display.
  if ((display = XOpenDisplay(0)) == 0)
    return 0;

  // Set up the default screen.
  screen = DefaultScreen(display);

  // Allocate a font.
  if ((xfs = XLoadQueryFont(display, fontname)) == 0)
    return 0;
  fontheight = xfs->max_bounds.ascent + xfs->max_bounds.descent;
  fontwidth = fontheight;

  // Set up the default colormap.
  cmap = DefaultColormap(display, screen);

  // Allocate some colorful colors.
  defs[redcol].red = 50000; defs[redcol].green = defs[redcol].blue = 0;
  if (XAllocColor(display, cmap, &defs[redcol]) == 0)
    return 0;
  defs[greencol].green = 50000; defs[greencol].red = defs[greencol].blue = 0;
  if (XAllocColor(display, cmap, &defs[greencol]) == 0)
    return 0;
  defs[bluecol].blue = 50000; defs[bluecol].red = defs[bluecol].green = 0;
  if (XAllocColor(display, cmap, &defs[bluecol]) == 0)
    return 0;


  defs[cyancol].red = 0; defs[cyancol].green = defs[cyancol].blue = 50000;
  if (XAllocColor(display, cmap, &defs[cyancol]) == 0)
    return 0;
  defs[magcol].green = 0; defs[magcol].red = defs[magcol].blue = 50000;
  if (XAllocColor(display, cmap, &defs[magcol]) == 0)
    return 0;
  defs[yelcol].blue = 0; defs[yelcol].red = defs[yelcol].green = 50000;
  if (XAllocColor(display, cmap, &defs[yelcol]) == 0)
    return 0;
  defs[purple].blue = defs[purple].red = defs[purple].green = 65535;
  if (XAllocColor(display, cmap, &defs[purple]) == 0)
    return 0;

  // Allocate a bunch of gray scale colors or more colorful colors.
  if (colors)
  {
    int steps = 6;
    if (numcolors < 256)
      steps = 5;
    if (numcolors < 125)
      steps = 4;
    if (numcolors < 64)
      steps = 3;
    if (numcolors < 27)
      steps = 2;

    for (int i = 0; i < steps; ++i)
      for (int j = 0; j < steps; ++j)
	for (int k = 0; k < steps; ++k)
	{
	  grays[i * steps * steps + j * steps + k].red = 
	    65535 - 65535 * i / (steps - 1);
	  grays[i * steps * steps + j * steps + k].green = 
	    65535 - 65535 * j / (steps - 1);
	  grays[i * steps * steps + j * steps + k].blue = 
	    65535 - 65535 * k / (steps - 1);
	  if (XAllocColor(display, cmap, 
			  &grays[i * steps * steps + j * steps + k]) == 0)
	    return 0;
	}
  }
  else
  {
    for (int i = 0; i < numcolors; ++i)
    {
      grays[i].red = grays[i].green = grays[i].blue = 
	65535 - (65535 / numcolors * i);
      if (XAllocColor(display, cmap, &grays[i]) == 0)
	return 0;
    }
  }

  // Set up some defaults for the windows
  xswa.event_mask = ExposureMask | StructureNotifyMask |
    ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | KeyReleaseMask;
  xswa.border_pixel = BlackPixel(display, screen);
  //xswa.border_pixel=defs[redcol].pixel;
  //  xswa.background_pixel = grays[numcolors / 6].pixel;
  xswa.background_pixel = grays[numcolors / 20].pixel;
  
  xswa.colormap = DefaultColormap(display, screen);

  // Set up an atom for later call to window manager function
  wmdw = XInternAtom(display, "WM_DELETE_WINDOW", False);

  inited = 1;
  return 1;

} // xobject::initX

// __________________________________________________________________________
// Close a connection to a display.

int xobject::quitX()
{
  inited = 0;

  XCloseDisplay(display);

  return 1;

} // xobject::quitX

// __________________________________________________________________________
// xobject.C

#endif // NOX

#endif // XOBJECT_C
